// bteste.C Version 7.42e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-26 15:33:48 tac>
//
// 度数集計表を入力し、出現率一定の検定または二項検定と出現率の信頼区間を
// 計算する。度数集計表は、計数値解析コマンド用の標準入力形式に準じた形式
// になっている必要がある。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W bteste bteste.C
//               -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "bteste  Version 7.42e        Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:bteste  [<OPTION> ...]  [<TAB FILE> ...]",
  "  The test of proportion with the frequency table.",
  "<OPTION>",
  "  -ccs[={1|2}]:specify the type of confidence interval",
  "    1:one-tailed  2:two-tailed(default)",
  "  -f=<FILE>:read options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -p=<PROBABILITY1>[,<PROBABILITY2>,...]:specify the theoretical probability",
  "    default:-p=1/number of class,1/number of class,... (0<PROBABILITY<1)",
  "  -res=[-][{def|all|bt[c]|cc[<C-COEF>]|hwt[{f|o}]|m[{b|f|s}]|tab}][,...]",
  "    :select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    bt[{c|o}]:test of proportion",
  "      c:with Yates's correction  o:no Yates's correction",
  "    cc[<C-COEF>]:confidence coefficient(0<=C-COEF<1 cc:cc0.95)",
  "    hwt[{f|o}]:HWE test",
  "      f:add Fisher's exact test  o:no Fisher's exact test",
  "    m[{b|f|s}]:multiple comparison",
  "      b:Bonferroni type  f:Fisher type  s:Scheffe type",
  "    tab:frequency table",
  "    default:auto judgment according to data"
};

const string Title = "=== Comparison of Proportion for Nominal Data ==="; // 表題

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  testtype vbt;                 // 出現率の検定の種類
  double vcc;                   // 信頼係数
  int vccs;                     // 信頼区間の種類
  testtype vhwt;                // Hardy-Weinberg平衡の検定の種類
  mctype vm;                    // 2分類ごとの多重比較法
  lstr plst;                    // 理論確率リスト
  vecdbl vpvec;                 // 理論確率ベクトル
  tabtype vtab;                 // 度数集計表出力フラッグ
  lstr vreslst;                 // 結果出力リスト
  map<string, bool> vresmap;    // 結果出力の有無マップ
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  testtype bt() const           // 出現率の検定の種類を返す
  { return vbt; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  int ccs() const               // 信頼区間の種類を返す
  { return vccs; }
  testtype hwt() const         // Hardy-Weinberg平衡の検定の種類を返す
  { return vhwt; }
  mctype m() const              // 2分類ごとの多重比較法を返す
  { return vm; }
  double p0(int pno) const      // 理論確率を返す
  { return (pno < (int)(vpvec.size())) ? vpvec[pno] : BLANKf; }
  const vecdbl& pvec() const    // 理論確率ベクトルを返す
  { return vpvec; }
  tabtype tab() const           // 度数集計表の種類を返す
  { return vtab; }
  bool res(const string& code); // 解析結果出力フラッグを返す
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(int nr);            // オプションをチェックする
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vbt(testcor), vcc(CCdef), vccs(2), vhwt(testetc), vm(mcbon), plst(), vpvec(1), vtab(tabfreq), vreslst(), vresmap()
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
  vpvec[0] = 1.0;
}

bool option::res(const string& code) // 結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "ccs")          // 信頼区間の指定
  {
    if (opt[1].empty() || opt[1] == "2") vccs = 2;
    else if (opt[1] == "1") vccs = 1;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "p")       // 理論確率指定
  {
    if (strsplit(plst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "res")     // 解析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(int nr)       // オプションをチェックする
{
  if (nr < 1) return 0;         // 理論確率自動設定
  if (nr <= (int)(plst.size())) return disperr(vpname, "Too many theoretical probabilities !");
  vpvec.resize(nr + 1);
  vpvec[0] = 1.0;               // 0:不等確率,1:等確率
  double p = 0.0, pt = 0.0;
  int i = 1;
  lstr_citr pitr = plst.begin();
  while (pitr != plst.end() && i < nr)
  {
    p = strtodf(*pitr);
    pt += p;
    if (p <= 0.0 || 1.0 <= p || pt > 1.0)
    { return disperr(vpname, "Wrong theoretical probability !"); }
    else
    {
      if (i > 1 && p != vpvec[i - 1]) vpvec[0] = 0.0;
      vpvec[i] = p;
      ++pitr;  ++i;
    }
  }
  p = (1.0 - pt) / (double)(nr + 1 - i); // 残りの理論確率
  if (i > 1 && p != vpvec[i - 1]) vpvec[0] = 0.0;
  while (i <= nr)
  {
    vpvec[i] = p;
    ++i;
  }
  if (set_res() < 0) return -2; // -resオプションのセット
  return 0;                     // エラー無し
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["bt"] = vresmap["cc"] = vresmap["tab"] = true;
  vresmap["hwt"] = vresmap["m"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["bt"] = vresmap["cc"] = vresmap["tab"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        {
          vresmap[res.substr(1)] = false;
          if (res == "-cc") vccs = 0; // 信頼区間を出力しない
        }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["bt"] = vresmap["cc"] = vresmap["tab"] = true;
      }
      else if (res == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["bt"] = vresmap["cc"] = vresmap["hwt"] = vresmap["m"] = vresmap["tab"] = true;
        vhwt = testetc;         // Fisherの正確検定も出力する
      }
      else if (res == "btc") // 連続修正有り
      { vresmap["bt"] = true;  vbt = testcor; }
      else if (res == "bto") // 連続修正無し
      { vresmap["bt"] = true;  vbt = testorg; }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (res == "hwtf") // Fisherの正確検定も出力する
      { vresmap["hwt"] = true;  vhwt = testetc; }
      else if (res == "hwto") // Fisherの正確検定は出力しない
      { vresmap["hwt"] = true;  vhwt = testorg; }
      else if (res[0] == 'm' && res.size() > 1)
      {
        switch (res[1])
        {
        case 'b' : vm = mcbon;
          break;
        case 'f' : vm = mcfis;
          break;
        case 's' : vm = mcsch;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " option !");
        }
        vresmap["m"] = true;
      }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class btestclass                // btestメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, idd& id) const;
  // タイトルとデータ識別情報を出力する
  void put_table(ostream& fout, const option& opt, const ftable& ftab) const;
  // 度数集計表を出力する
  void put_btest(ostream& fout, const option& opt, const ftable& ftab) const;
  // 二項検定の結果を出力する
  void put_ctest(ostream& fout, const option& opt, const ftable& ftab) const;
  // χ2乗検定の結果を出力する
  void put_cmcon(ostream& fout, const option& opt, mctype mc, const ftable& ftab) const;
  // 出現率の多重比較結果を出力する
  void put_hwtest(ostream& fout, const option& opt, const ftable& ftab) const;
  // Hardy-Weinberg平衡の検定の結果を出力する
  ldbl hwtest(double& p1, double& p2, double& vr, double& csqr, const ftable& ftab) const;
  // 遺伝子に関するHardy-Weinberg平衡の検定(Fisherの両側直接確率を返す)
public:
  btestclass() : fname()        // コンストラクタ
  {}
  ~btestclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst);
  // 度数集計表を読み込む
  void put(option& opt, listidd& idlst, listftab& ftablst) const;
  // 結果を出力する
};

void btestclass::put_id(ostream& fout, idd& id) const
// タイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put(fout);
}

void btestclass::put_table(ostream& fout, const option& opt, const ftable& ftab) const
// 度数集計表を出力する
{
  string codei(ftab.ccode(0));
  int npad = codei.size() - strclen(codei, CHARwide);
  fout << "\n" << left << setw(ftab.cwide(0) + npad) << codei << right;
  codei = ftab.ccode(1);
  npad = codei.size() - strclen(codei, CHARwide);
  fout << setw(ftab.cwide(1) + npad) << codei;
  fout << " (  %   ";
  int w = ftab.hwide() + ftab.cwide(1) + 10;
  if (opt.ccs() != 0)
  {
    fout << setw(5) << opt.cc() * 100.0;
    if (opt.ccs() == 1) fout << "% One-tailed CI";
    else fout << "% Two-tailed CI";
    w += 19;
  }
  fout << ")\n";
  putline(fout, w);
  int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
  fout.precision(3);  fout.setf(ios::fixed);
  int nt = ftab.ftotal();
  double cc = opt.cc();
  if (opt.ccs() == 1) cc = (opt.cc() > 0.5) ? opt.cc() * 2.0 - 1.0 : 0.0;
  for (size_t i = 1; i <= ftab.nr(); ++i)
  {
    codei = ftab.rcode(i);
    npad = codei.size() - strclen(codei, CHARwide);
    fout << left << setw(ftab.cwide(0) + npad) << codei << right;
    int ni = ftab.f(i, 1);
    double pr = (double)ni / (double)nt;
    fout << setw(ftab.cwide(1)) << ni;
    fout << " (" << setw(7) << pr * 100.0;
    if (opt.ccs() != 0)
    {
      double plow = 0.0, pup = 1.0;
      calbexci(plow, pup, ni, nt, cc);
      if (opt.ccs() == 1)
      {
        if (pr < 0.5) plow = 0.0;
        else pup = 1.0;
      }
      fout << setw(8) << plow * 100.0 << " --" << setw(8) << pup * 100.0;
    }
    fout << ")\n";
  }
  putline(fout, w);
  codei = "TOTAL";
  npad = codei.size() - strclen(codei, CHARwide);
  fout << left << setw(ftab.hwide() + npad) << codei << right;
  fout << setw(ftab.cwide(1)) << ftab.fctotal(1);
  fout << " (" << setw(7) << 100.0;
  fout << ")\n";
  fout.precision(prec0);  fout.unsetf(ios::fixed);
}

void btestclass::put_btest(ostream& fout, const option& opt, const ftable& ftab) const
// 二項検定の結果を出力する
{
  ldbl pr1 = binom(ftab.f(1, 1), ftab.f(2, 1), opt.p0(1));
  ldbl pr = (pr1 < 0.5L) ? 2.0L * pr1 : 1.0L; // 両側p値
  fout << "\nBinomial Test:One-tailed p=" << pr1 << pind(pr1) << "  Two-tailed p=" << pr << pind(pr);
}

void btestclass::put_ctest(ostream& fout, const option& opt, const ftable& ftab) const
// χ2乗検定の結果を出力する
{
  int df = ftab.nr() - 1;
  double csqr = ftest(opt.pvec(), ftab, (opt.bt() == testcor) ? true : false);
  ldbl pr = ctop(csqr, df);
  fout << "\nChi-squared Test(";
  if (opt.bt() == testcor) fout << "with Yates's correction,";
  fout << ftab.nr() << "x1)\n";
  fout << "  chi-squared=" << csqr << "  DF=" << df << "  p=" << pr << pind(pr) << "\n";
  fout << "  Theoretical Probability:";
  for (size_t i = 1; i <= ftab.nr(); ++i) fout << ftab.rcode(i) << "=" << opt.p0(i) << "  ";
  fout << endl;
}

void btestclass::put_cmcon(ostream& fout, const option& opt, mctype mc, const ftable& ftab) const
// 出現率の多重比較結果を出力する
{
  double f0 = 0.5 * (double)(ftab.nr()) / double(ftab.ftotal());
  int df = 1, dc = ftab.nr() * (ftab.nr() - 1) / 2;
  string method("BONFERRONI");
  switch (mc)
  {
  case mcfis : method = "FISHER";  dc = 1;  df = 1;
    break;
  case mcsch : method = "SCHEFFE";
    df = ftab.nr() - 1;  dc = 1;
    break;
  }
  fout << "\n<" << method << " TYPE MULTIPLE COMPARISON>\n";
  fout << "CODE    :       % - CODE    :       % Chi-squared      DF       p-value\n";
  constexpr int LENline = 75;   // 区切り線の長さ
  putline(fout, LENline);
  for (size_t i = 1; i < ftab.nr(); ++i)
  {
    string codei(ftab.rcode(i));
    int npadi = codei.size() - strclen(codei, CHARwide);
    int fi = ftab.f(i, 1);
    double pri = (double)fi / (double)(ftab.ftotal()) * 100.0;
    for (size_t j = i + 1; j <= ftab.nr(); ++j)
    {
      double prj = (double)(ftab.f(j, 1)) / (double)(ftab.ftotal()) * 100.0;
      double csqr = f0 * (double)(fi - ftab.f(j, 1)) * (fi - ftab.f(j, 1));
      double pr = calmcpr(mc, csqr, ftab.nr(), df, 0, dc);
      string codej(ftab.rcode(j));
      int npad = codej.size() - strclen(codej, CHARwide);
      fout << left << setw(8 + npad) << codej;
      fout << ":" << right << setw(8) << prj;
      fout << left << " - ";
      fout << setw(8 + npadi) << codei;
      fout << ":" << right << setw(8) << pri;
      fout << setw(12) << csqr << setw(8) << df;
      fout << setw(14) << pr << pind(pr) << "\n";
    }
  }
  putline(fout, LENline);
}

void btestclass::put_hwtest(ostream& fout, const option& opt, const ftable& ftab) const
// Hardy-Weinberg平衡の検定の結果を出力する
{
  double p1 = 0.0, p2 = 0.0, vr = 0.0, csqr = 0.0;
  ldbl prfis = hwtest(p1, p2, vr, csqr, ftab);
  double nn = 2 * ftab.ftotal();
  double r2 = csqr / nn;
  double r = sqrt(r2);
  fout << "\n<TEST OF HARDY-WEINBERG EQUILIBRIUM>\n";
  fout << "  Estimated Probability:Gene1=" << p1 << "  Gene2=" << p2 << "  Vr=" << vr << "\n";
  fout << "  Theoretical Probability:" << ftab.rcode(1) << "=" << p1 * p1;
  fout << "  " << ftab.rcode(2) << "=" << 2.0 * p1 * p2;
  fout << "  " << ftab.rcode(3) << "=" << p2 * p2 << "\n";
  fout << "  Coefficient of Contingency:V=" << r << "  V-squared=" << r2 << "\n";
  double csqrl = 0.0, csqru = 0.0;
  calcsqrci(csqrl, csqru, opt.cc(), 1, csqr);
  double rl = sgn(r) * sqrt(csqrl / nn);
  double ru = sgn(r) * sqrt(csqru / nn);
  if (rl < -1.0) rl = -1.0;
  if (rl > 1.0) rl = 0.0;
  if (ru < -1.0) ru = -1.0;
  if (ru > 1.0) ru = 1.0;
  ldbl pr = ctop(csqr, 1);
  fout << "  " << opt.cc() * 100.0 << "% CI of V:Lower=" << rl << "  Upper=" << ru << "\n";
  fout << "  chi-squared=" << csqr << "  DF=1  p=" << pr << pind(pr) << "\n";
  if (opt.hwt() == testetc) fout << "  Fisher's Exact Test:Two-tailed p=" << prfis << pind(prfis) << "\n";
}

ldbl btestclass::hwtest(double& p1, double& p2, double& vr, double& csqr, const ftable& ftab) const
// 遺伝子に関するHardy-Weinberg平衡の検定(Fisherの両側直接確率を返す)
{
  p1 = (double)(2 * ftab.f(1, 1) + ftab.f(2, 1)) / (double)(2 * ftab.ftotal());
  p2 = 1.0 - p1;
  vr = p1 * p2 / (double)(2 * ftab.ftotal());
  vecdbl pvec(4);
  pvec[0] = 0.0;
  pvec[1] = p1 * p1;  pvec[2] = 2.0 * p1 * p2;  pvec[3] = p2 * p2;
  ftable ftab2(ftab);           // 計算用
  ftab2.f(1, 1 ,2 * ftab.f(1, 1));  ftab2.f(2, 1, 2 * ftab.f(2, 1));
  ftab2.f(3, 1, 2 * ftab.f(3, 1));
  csqr = ftest(pvec, ftab2, false);
  ldbl pr = fisher(2 * ftab.f(1, 1), ftab.f(2, 1), ftab.f(2, 1), 2 * ftab.f(3, 1));
  return (pr < 0.5L) ? 2.0L * pr : 1.0L; // 両側確率を返す
}

int btestclass::get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst)
// 度数集計表を読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nftab = fgetftable(fname, idlst, ftablst);
  if (nftab < 0) throw runtime_error("Run-time error !");
  return nftab;                 // 読み込んだ度数集計表の数を返す
}

void btestclass::put(option& opt, listidd& idlst, listftab& ftablst) const
// 結果を出力する
{
  listidd::iterator iitr = idlst.begin();
  for (listftab::const_iterator fitr = ftablst.begin(); fitr != ftablst.end(); ++iitr, ++fitr)
  {
    if (opt.check(fitr->nr()) != 0) throw option_error("Run-time oprion error !");
    if (opt.res("res"))
    {
      put_id(cout, *iitr);
      if (fitr->nr() > 0)
      {
        if (opt.res("tab")) put_table(cout, opt, *fitr);
        if (opt.res("bt"))      // 出現率の検定出力
        {
          if (fitr->nr() == 2) put_btest(cout, opt, *fitr);
          put_ctest(cout, opt, *fitr);
        }
        if (fitr->nr() > 2 && opt.res("m") && opt.p0(0) == 1.0) put_cmcon(cout, opt, opt.m(), *fitr);
        if (fitr->nr() == 3 && opt.res("hwt")) put_hwtest(cout, opt, *fitr);
      }
      cout << "\n" << endl;
    }
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  btestclass btest;             // メインクラス変数
  option opt(argv[0]);          // オプション
  listidd idlst;                // データ識別情報リスト
  listftab ftablst;             // 度数表リスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok)                // 標準入力から読み込む
        {
          btest.get(nullptr, opt, idlst, ftablst);
          btest.put(opt, idlst, ftablst);
        }
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      {
        btest.get(argv[i], opt, idlst, ftablst);  fok = false;
        btest.put(opt, idlst, ftablst);
      }
    }
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
