// crosse.C Version 7.42e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-26 15:52:24 tac>
//
// 縦と横のクロス集計を行い、度数集計表を出力する。
// 度数集計表は計数値解析コマンド用の標準入力形式に従っている。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W crosse crosse.C
//               -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "crosse  Version 7.42e        Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:crosse  [<OPTION> ...]  [<DATA FILE> ...]",
  "  The cross-tabulation of items.",
  "<OPTION>",
  "  -c=[<START NUM>][-[<END NUM>]]:specify the item number of the column items",
  "  -f=<FILE>:read options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=crosse_ngp<2-digit number of row>",
  "    <2-digit number of column>.dat",
  "  -r=[<START NUM>][-[<END NUM>]]:specify the item number of the row items",
  "  -s[={-|+|c|r}]:select the data aggregation method(-s:-s=+)",
  "    -:aggregate as figures(default)  +:aggregate as strings",
  "    c:aggregate only column item as strings",
  "    r:aggregate only row item as strings",
  "  -sep[={-|+}]:specify whether to aggregate by each item(-sep:-sep=+)",
  "  -tab[={-|+|data|dns|fr|pc|pr|pt}]:select the type of output table",
  "    data:standard input format for numerical analysis commands(default)",
  "    fr:only frequency  pc:also output % for column total",
  "    pr:also output % for row total  pt:also output % for all total",
  "  -trm[={-|+}]:specify whether to trimming for cross-table(-trm:-trm=+)",
  "  -wc=<NUMBER>:specify the number of outout digits for column data",
  "    default:-wc=6",
  "  -wr=<NUMBER>:specify the number of outout digits for row data",
  "    default:-wr=6(minimum number)"
};

const string Title = "=== Cross-tabulation ==="; // 表題
const string NGPNAMEdef = "crosse_ngp"; // 省略時Ngraphファイル名
constexpr int Wdef = 6;          // 省略時出力桁数
constexpr int Wmax = LENcrt / 2; // 最大出力桁数

using paircode = pair<string, double>; // 行データの別名定義

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  pairint rno;                  // 行の集計開始・終了項目番号
  pairint cno;                  // 列の集計開始・終了項目番号
  string vngpname;              // Ngraphファイル基本名
  string vngpnamen;             // Ngraphファイル名(番号付き)
  bool vstrc;                   // 行データ文字列集計フラッグ
  bool vstrr;                   // 例データ文字列集計フラッグ
  bool vsep;                    // 項目ごと集計フラッグ
  tabtype vtab;                 // 度数表の種類
  bool vtrm;                    // クロス集計表のトリミングフラッグ
  int vwr;                      // 行データの出力桁数
  int vwc;                      // 列データの出力桁数
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  int rs() const                // 行の集計開始項目番号を返す
  { return rno.first; }
  int re() const                // 行の集計終了項目番号を返す
  { return rno.second; }
  int cs() const                // 列の集計開始項目番号を返す
  { return cno.first; }
  int ce() const                // 列の集計終了項目番号を返す
  { return cno.second; }
  const string& ngpname(int rno, int cno); // Ngraphデータファイル名を返す
  bool sep() const              // 項目ごとの集計フラッグを返す
  { return vsep; }
  bool strc() const             // 列文字列集計フラッグを返す
  { return vstrc; }
  bool strr() const             // 行文字列集計フラッグを返す
  { return vstrr; }
  tabtype tab() const           // 度数表の種類を返す
  { return vtab; }
  bool trm() const           // クロス集計表のトリミングフラッグを返す
  { return vtrm; }
  int wr() const                // 行データの出力桁数を返す
  { return vwr; }
  int wc() const                // 列データの出力桁数を返す
  { return vwc; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オブションのチェックとセット
};

option::option(const char* c_pname) : vpath(), vpname(), rno(1, 1), cno(2, INT_MAX), vngpname(), vngpnamen(), vsep(true), vstrc(false), vstrr(false), vtab(tabdata), vtrm(true), vwr(Wdef), vwc(Wdef)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

const string& option::ngpname(int rno, int cno) // Ngraphデータファイル名を返す
{
  if (vngpname == "") return vngpname; // -ngpname指定なし
  ostringstream ostr;
  if (rno < 10)
  {
    if (cno < 10) ostr << vngpname << "0" << rno << "0" << cno << ".dat";
    else ostr << vngpname << "0" << rno << cno << ".dat";
  }
  else
  {
    if (cno < 10) ostr << vngpname << rno << "0" << cno << ".dat";
    else ostr << vngpname << rno << cno << ".dat";
  }
  vngpnamen = (string)(ostr.str());
  return vngpnamen;             // 番号付きファイル名を返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "c")            // 列項目の指定
  {
    cno.first = 1;  cno.second = INT_MAX;
    if (getrange(cno.first, cno.second, opt[1]) < 1) err = badopt(vpname, optstr, USAGE);
    else if (cno.first < 1 || cno.first > cno.second) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
  }
  else if (opt[0] == "r")       // 行項目の指定
  {
    rno.first = 1;  rno.second = INT_MAX;
    if (getrange(rno.first, rno.second, opt[1]) < 1) err = badopt(vpname, optstr, USAGE);
    else if (rno.first < 1 || rno.first > rno.second) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "s")       // 文字列として集計指定
  {
    if (opt[1].empty() || opt[1] == "+") vstrc = vstrr = true;
    else if (opt[1] == "-") vstrc = vstrr = false;
    else if (opt[1] == "c") vstrc = true;
    else if (opt[1] == "r") vstrr = true;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "sep")     // 項目ごとの集計指定
  {
    if (opt[1].empty() || opt[1] == "+") vsep = true;
    else if (opt[1] == "-") vsep = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "tab")     // 度数表の種類指定
  {
    if (opt[1].empty() || opt[1] == "+" || opt[1] == "data") vtab = tabdata;
    else if (opt[1] == "-") vtab = tabnon;
    else if (opt[1] == "dns") vtab = tabetc;
    else if (opt[1] == "fr") vtab = tabfreq;
    else if (opt[1] == "pr" || opt[1] == "p") vtab = tabfprr;
    else if (opt[1] == "pc") vtab = tabfprc;
    else if (opt[1] == "pt") vtab = tabfprt;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "trm")     // クロス集計表のトリミング指定
  {
    if (opt[1].empty() || opt[1] == "+") vtrm = true;
    else if (opt[1] == "-") vtrm = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "wc")      // 列データの出力桁数指定
  {
    char* ptr = nullptr;
    vwc = strtol(opt[1].c_str(), &ptr, 0);
    if (vwc < 2 || Wmax < vwc) err = badopt(vpname, optstr, USAGE, 2, Wmax);
  }
  else if (opt[0] == "wr")      // 行データの出力桁数指定
  {
    char* ptr = nullptr;
    vwr = strtol(opt[1].c_str(), &ptr, 0);
    if (vwr < Wdef || Wmax < vwr) err = badopt(vpname, optstr, USAGE, Wdef, Wmax);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error !") + optstr);
}

int option::check(size_t ni)    // オプションのチェックとセット
{
  if ((int)ni < rno.first) return disperr(vpname, "Wrong number of -r option !");
  if (rno.second != INT_MAX && (int)ni < rno.second) return disperr(vpname, "Wrong number of -r option !");
  if ((int)ni < cno.first) return disperr(vpname, "Wrong number of -c option !");
  if (cno.second != INT_MAX && (int)ni < cno.second) return disperr(vpname, "Wrong number of -c option !");
  return 0;
}

class varclass                  // 変数クラス
{
private:
  int vni;                      // 項目数
  ftable vftab;                 // 度数表
  set<paircode> rset;           // 行データセット
  set<paircode> cset;           // 列データセット
public:
  varclass(size_t ni) : vni(ni), vftab(), rset(), cset() // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int ni() const                // 項目数を返す
  { return vni; }
  ftable& ftab()                // 度数表を返す
  { return vftab; }
  int fcode(const option& opt, ldata& datalst); // コードセットを作成する
  void fsetcode(const option& opt); // 度数表のコードをセットする
  void ffreq(const option& opt, ldata& datalst, int ridx, int cidx);
  // 度数表の度数をクロス集計する
};

int varclass::fcode(const option& opt, ldata& datalst) // コードセットを作成する
{
  int ncode = 0;
  for (auto& cdata : datalst)
  {
    int ie = min(opt.re(), vni);
    bool ok = false;
    for (int j = opt.rs() - 1; j < ie; ++j)
    {
      double fdata = cdata.fdata(j);
      if (opt.strr()) ok = (cdata.sdata(j) == BLANKs) ? false : true;
      else ok = (fdata == BLANKf) ? false : true;
      if (ok)
      {
        paircode tdata("", fdata);
        if (opt.strr()) tdata.first = cdata.sdata(j); // 文字として集計
        rset.insert(tdata);       // 新規データだけ挿入する
        ++ncode;                  // 集計したコード数を数える
      }
    }
    ie = min(opt.ce(), (int)vni);  ok = false;
    for (int j = opt.cs() - 1; j < ie; ++j)
    {
      double fdata = cdata.fdata(j);
      if (opt.strc()) ok = (cdata.sdata(j) == BLANKs) ? false : true;
      else ok = (fdata == BLANKf) ? false : true;
      if (ok)
      {
        paircode tdata("", fdata);
        if (opt.strc()) tdata.first = cdata.sdata(j);
        cset.insert(tdata);
        ++ncode;
      }
    }
  }
  return ncode;                 // 集計したコード数を返す
}

void varclass::fsetcode(const option& opt) // 度数表のコードをセットする
{
  vftab.resize(rset.size(), cset.size());
  ostringstream ostr;
  string tstr("CODE");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  ostr << left << setw(opt.wr() + npad) << tstr << right;
  vftab.hed(ostr.str());
  int i = 1;
  for (auto& codef : cset)
  {
    ostr.str("");
    if (opt.strc())
    {
      tstr = codef.first;
      npad = tstr.size() - strclen(tstr, CHARwide);
      ostr << setw(opt.wc() + npad) << tstr;
    }
    else
    { ostr << setw(opt.wc()) << codef.second; }
    vftab.ccode(i, ostr.str());
    ++i;
  }
  i = 1;
  for (auto& codef : rset)
  {
    ostr.str("");
    if (opt.strr()) ostr << codef.first;
    else ostr << codef.second;
    vftab.rcode(i, ostr.str());
    ++i;
  }
}

void varclass::ffreq(const option& opt, ldata& datalst, int ridx, int cidx)
// 度数表の度数をクロス集計する
{
  int ridxs = opt.rs() - 1, ridxe = min(opt.re(), vni);
  int cidxs = opt.cs() - 1, cidxe = min(opt.ce(), vni);
  if (opt.sep())                // 項目ごとに集計
  { ridxs = ridx;  ridxe = ridx + 1;  cidxs = cidx;  cidxe = cidx + 1; }
  vecdbl fdatavec(vni);
  for (auto& cdata : datalst)
  {
    if (opt.strc() == false)
    {
      for (int j = cidxs; j < cidxe; ++j) fdatavec[j] = cdata.fdata(j);
    }
    for (int i = ridxs; i < ridxe; ++i)
    {
      int rno = -1;
      if (opt.strr())
      {
        if (cdata.sdata(i) != BLANKs) rno = search_data(rset, cdata.sdata(i));
      }
      else
      {
        double fdata = cdata.fdata(i);
        if (fdata != BLANKf) rno = search_data(rset, fdata);
      }
      if (rno > 0)
      {
        for (int j = cidxs; j < cidxe; ++j)
        {
          int cno = -1;
          if (opt.strc())
          {
            if (cdata.sdata(j) != BLANKs) cno = search_data(cset, cdata.sdata(j));
          }
          else
          {
            if (fdatavec[j] != BLANKf) cno = search_data(cset, fdatavec[j]);
          }
          if (cno > 0) vftab.f(rno, cno, vftab.f(rno, cno) + 1);
        }
      }
    }
  }
  vftab.set();
}

class crossclass                // crossメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, const option& opt, idd& id, int ridx, int cidx, varclass& var) const;
  // 度数表のタイトルとデータ識別情報を出力する
  void put_dtab(ostream& fout, varclass& var) const;
  // 度数表を度数データ形式で出力する
  void put_dns(ostream& fout, varclass& var) const;
  // 度数表をデータファイル形式で出力する
  void plot_ngp(ofstream& fout, const option& opt, idd& id, int ridx, int cidx, varclass& var) const;
  // Ngraph用にバーチャートデータを出力する
  void plot_ngpid(ofstream& fout, const option& opt, idd& id, int ridx , int cidx, varclass& var) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
public:
  crossclass() : fname()        // コンストラクタ
  {}
  ~crossclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込み累積する
  void put(option& opt, idd& id, ldata& datalst) const;
  // 集計結果を出力する
};

void crossclass::put_id(ostream& fout, const option& opt, idd& id, int ridx , int cidx, varclass& var) const
// タイトルとデータ識別情報を出力する
{
  idd tid(id.title(), 1 + (var.ftab()).nc() * 2);
  string titem, tcode;
  int rs = opt.rs(), re = min(opt.re(), (int)(id.ni()));
  int cs = opt.cs(), ce = min(opt.ce(), (int)(id.ni()));
  if (opt.sep())                // 項目ごとに集計
  { rs = ridx + 1;  re = ridx + 1;  cs = cidx + 1;  ce = cidx + 1; }
  switch (opt.tab())
  {
  case tabdata : id.put_title(fout, formtag);
    id.put_stra(fout, formtag);
    for (int i = rs; i <= re; ++i)
    {
      id.assign_item(i - 1, id.item(i - 1), atype, avalrow);
      id.put_item(fout, i - 1, formtag);
    }
    for (int i = cs; i <= ce; ++i)
    {
      id.assign_item(i - 1, id.item(i - 1), atype, avalcol);
      id.put_item(fout, i - 1, formtag);
    }
    break;
  case tabetc : id.put_title(fout, formtag);
    id.put_stra(fout, formtag);
    if (rs == re) { tid.assign_item(0, id.item(rs - 1)); } // 縦項目が1項目
    else { tid.assign_item(0, id.item(rs - 1)+"-"+id.item(re - 1)); }
    if (cs == ce) { titem = id.item(cs - 1); } // 横項目が1項目
    else { titem = id.item(cs - 1)+"-"+id.item(ce - 1); }
    for (size_t i = 1; i <= (var.ftab()).nc(); ++i)
    {
      tcode = (var.ftab()).ccode(i);
      strtrm(tcode);
      size_t iidx = 1 + (i - 1) * 2;
      tid.assign_item(iidx, titem+"="+tcode+":n");
      tid.assign_item(iidx + 1, titem+"="+tcode+":(%)");
    }
    tid.put_item(fout, formtag);
    break;
  default : puttitle(fout, Title);
    fout << "\n";
    id.put_title(fout);
    id.put_stra(fout);
    for (int i = rs; i <= re; ++i)
    {
      id.assign_item(i - 1, id.item(i - 1), atype, avalrow);
      fout << "ROW ITEM   :" << id.item(i - 1) << "\n";
    }
    for (int i = cs; i <= ce; ++i)
    {
      id.assign_item(i - 1, id.item(i - 1), atype, avalcol);
      fout << "COLUMN ITEM:" << id.item(i - 1) << "\n";
    }
  }
}

void crossclass::put_dtab(ostream& fout, varclass& var) const
// 度数表を度数データ形式で出力する
{
  datac tdata((var.ftab()).nc() + 1);
  tdata.assign(0, (var.ftab()).hed());
  for (size_t j = 1; j <= (var.ftab()).nc(); ++j) tdata.assign(j, (var.ftab()).ccode(j));
  tdata.put(fout);
  for (size_t i = 1; i <= (var.ftab()).nr(); ++i)
  {
    tdata.assign(0, (var.ftab()).rcode(i));
    for (size_t j = 1; j <= (var.ftab()).nc(); ++j) tdata.assign(j, (double)((var.ftab()).f(i, j)));
    tdata.put(fout);
  }
}

void crossclass::put_dns(ostream& fout, varclass& var) const
// 度数表をデータファイル形式で出力する
{
  datac tdata(1 + (var.ftab()).nc() * 2);
  tdata.assign(0, "CODE");
  for (size_t j = 1; j <= (var.ftab()).nc(); ++j)
  {
    string tcode((var.ftab()).ccode(j));
    strtrm(tcode);
    size_t didx = 1 + (j - 1) * 2;
    tdata.assign(didx, tcode);
    tdata.assign(didx + 1, tcode+"(%)");
  }
  tdata.put(fout);
  for (size_t i = 1; i <= (var.ftab()).nr(); ++i)
  {
    tdata.assign(0, (var.ftab()).rcode(i));
    double frti = var.ftab().frtotal(i);
    for (size_t j = 1; j <= (var.ftab()).nc(); ++j)
    {
      size_t didx = 1 + (j - 1) * 2;
      double fij = var.ftab().f(i, j);
      tdata.assign(didx, fij);
      double pro = (frti > 0.0) ? fij / frti * 100.0 : BLANKf;
      tdata.assign(didx + 1, pro);
    }
    tdata.put(fout);
  }
  tdata.assign(0, "TOTAL");
  double ft = var.ftab().ftotal();
  for (size_t j = 1; j <= (var.ftab()).nc(); ++j)
  {
    size_t didx = 1 + (j - 1) * 2;
    double fctj = var.ftab().fctotal(j);
    tdata.assign(didx, fctj);
    double pro = (ft > 0.0) ? fctj / ft * 100.0 : BLANKf;
    tdata.assign(didx + 1, pro);
  }
  tdata.put(fout);
}

void crossclass::plot_ngp(ofstream& fout, const option& opt, idd& id, int ridx, int cidx, varclass& var) const
// Ngraph用にバーチャートデータを出力する
{
  plot_ngpid(fout, opt, id, ridx, cidx, var);
  double x = 0.0, y = 0.0;
  int n = 0;
  for (size_t i = 1; i <= (var.ftab()).nr(); ++i)
  {
    if ((var.ftab()).frtotal(i) > 0)
    {
      fout << "%" << (var.ftab()).rcode(i) << "\n";
      x = 0.0;  y = (double)((var.ftab()).nr() + 1 - i) + 0.5;  n = 0;
      for (size_t j = 1; j <= (var.ftab()).nc(); ++j)
      {
        if ((var.ftab()).f(i, j) > 0)
        {
          fout << x << " " << y - 0.3 << " ";
          n += (var.ftab()).f(i, j);
          x = (double)n / (double)((var.ftab()).frtotal(i)) * 100.0;
          fout << x << " " << y + 0.3 << endl;
        }
      }
    }
  }
  if ((var.ftab()).frtotal(0) > 0)
  {
    fout << "%" << (var.ftab()).rcode(0) << "\n";
    x = 0.0;  y = 0.4;  n = 0;
    for (size_t j = 1; j <= (var.ftab()).nc(); ++j)
    {
      if ((var.ftab()).f(0, j) > 0)
      {
        fout << x << " " << y - 0.3 << " ";
        n += (var.ftab()).f(0, j);
        x = (double)n / (double)((var.ftab()).frtotal(0)) * 100.0;
        fout << x << " " << y + 0.3 << endl;
      }
    }
  }
}

void crossclass::plot_ngpid(ofstream& fout, const option& opt, idd& id, int ridx , int cidx, varclass& var) const
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  int rs = opt.rs(), re = min(opt.re(), (int)(id.ni()));
  int cs = opt.cs(), ce = min(opt.ce(), (int)(id.ni()));
  if (opt.sep())                // 項目ごとに集計
  { rs = ridx + 1;  re = ridx + 1;  cs = cidx + 1;  ce = cidx + 1; }
  fout << "-s1 -drectangle_fill -x1 -y3 -zx0,100,20 -zy0,";
  fout << (var.ftab()).nr() + 1 << ",1\n";
  fout << "%";  puttitle(fout, Title);  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  for (int i = rs; i <= re; ++i)
  {
    id.assign_item(i - 1, id.item(i - 1), atype, avalrow);
    id.put_item(fout, i - 1, formngp);
  }
  for (int i = cs; i <= ce; ++i)
  {
    id.assign_item(i - 1, id.item(i - 1), atype, avalcol);
    id.put_item(fout, i - 1, formngp);
  }
}

int crossclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込み累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time opttion error !");
  return nc;                    // 読み込んだケース数を返す
}

void crossclass::put(option& opt, idd& id, ldata& datalst) const
// 集計結果を出力する
{
  varclass var(id.ni());               // 変数オブジェクト
  int ncode = var.fcode(opt, datalst); // コードセットを作成する
  var.fsetcode(opt);                   // 度数表のコードをセットする
  int ridxs = 0, ridxe = 1, cidxs = 0, cidxe = 1;
  bool async = true;
  if (opt.sep())                // 項目ごとに集計
  {
    ridxs = opt.rs() - 1;  ridxe = min(opt.re(), int(id.ni()));
    cidxs = opt.cs() - 1;  cidxe = min(opt.ce(), int(id.ni()));
    if (ridxe - ridxs == cidxe - cidxs) async = false;
  }
  int j = cidxs;
  for (int i = ridxs; i < ridxe; ++i)
  {
    if (async) j = cidxs;
    do
    {
      if (i > ridxs || j > cidxs)
      {
        if (opt.trm()) var.fsetcode(opt);
        (var.ftab()).fclear();
        if (opt.tab() != tabfreq && opt.tab() != tabfprc && opt.tab() != tabfprr && opt.tab() != tabfprt) cout << "\n";
      }
      var.ffreq(opt, datalst, i , j);
      if (opt.trm()) (var.ftab()).trm(); // 行・列の合計が0のものを削除する
      if (opt.tab() != tabnon)
      {
        put_id(cout, opt, id, i, j, var);
        if (id.nc() < 1 || ncode < 1) // データなし
        {
          if (opt.tab() != tabdata) cout << "\n" << endl;
          return;
        }
        switch (opt.tab())
        {
        case tabdata : put_dtab(cout, var); // 度数データ形式で出力
          break;
        case tabetc : put_dns(cout, var); // データファイル形式で出力
          break;
        default : putftable(cout, var.ftab(), opt.tab());
          cout << "\n" << endl;     // 度数表形式で出力
        }
      }
      if (opt.ngpname(i + 1, j + 1) != "") // Ngraph用バーチャートデータ出力
      {
        ofstream foutngp;
        if (ofopen(foutngp, (opt.ngpname(i + 1, j + 1)).c_str()) == false) return;
        plot_ngp(foutngp, opt, id, i, j, var);
        foutngp.close();
      }
      ++j;
    } while (async && j < cidxe);
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  crossclass cross;             // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) cross.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { cross.get(argv[i], opt, id, datalst);  fok = false; }
    }
    cross.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
