// freqe.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-26 16:15:47 tac>
//
// dansデータファイルを入力し、度数集計表を出力する。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W freqe freqe.C
//                -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "freqe  Version 7.41e         Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:freqe  [<OPTION> ...]  [<DATA FILE> ...]",
  "  The tally frequencies of specified items.",
  "<OPTION>",
  "  -b[=<STRINGS>]:specify strings to be excluded from the tally",
  "  -f=<FILE>:read options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -i=[<START NUMBER>][-[<END NUMBER>]]:specify the number of the item",
  "  -ngp={fr|pi}[,...]:select the data for Ngraph",
  "    fr:frequency distribution  pi:pie chart  default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output",
  "    -ngpname:-ngpname=freq_ngp<2-digit number of the item>.dat",
  "  -s[={-|+}]:select the data aggregation method(-s:-s=+)",
  "  -sep[={-|+}]:specify whether to aggregate by each item(-sep:-sep=+)",
  "  -tab[={-|btest|code|fr}]:select the type of output table",
  "    -:no output  btest:data format for the bteste  code:output only codes",
  "    fr:frequency table format(default)",
  "  -w=<NUMBER>:specify the number of output digits for data",
  "    minimum:6  minus number:aligning to the left"
};

const string Title = "=== Frequency Table ==="; // 表題
const string NGPNAMEdef = "freqe_ngp"; // 省略時Ngraphファイル基本名
constexpr int Wdef = -6;                 // 省略時出力桁数
constexpr int Wmax = LENcrt / 2;         // 最大出力桁数
constexpr int WMAXfd = 12;               // 浮動小数点表示の最大桁数

using mapsdi = map<pair<string, double>, int>; // 度数データマップの別名定義
using mapsdi_itr = mapsdi::iterator;           // 反復子の別名定義
using mapsdi_citr = mapsdi::const_iterator;    // const反復子の別名定義

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  string vblank;                // ブランク扱い文字列
  pairint ino;                  // 集計開始・終了項目番号
  string vngpname;              // Ngraphファイル基本名
  string vngpnamen;             // Ngraphファイル名(番号付き)
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  bool vstr;                    // 文字列集計フラッグ
  bool vsep;                    // 項目ごと集計フラッグ
  tabtype vtab;                 // 度数集計表の種類
  int vw;                       // データの出力桁数
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  const string& blank() const   // ブランク扱い文字列を返す
  { return vblank; }
  int is() const                // 集計開始項目番号を返す
  { return ino.first; }
  int ie() const                // 集計終了項目番号を返す
  { return ino.second; }
  const string& ngpname(int ino); // Ngraphデータファイル名を返す
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool str() const              // 文字列集計フラッグを返す
  { return vstr; }
  bool sep() const              // 項目ごと集計フラッグを返す
  { return vsep; }
  tabtype tab() const           // 度数集計表の種類を返す
  { return vtab; }
  int w() const                 // データの出力桁数を返す
  { return vw; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vblank(BLANKs), ino(1, INT_MAX), vngpname(), vngpnamen(), vngplst(), vngpmap(), vstr(false), vsep(true), vtab(tabfreq), vw(Wdef)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

const string& option::ngpname(int ino) // Ngraphデータファイル名を返す
{
  if (vngpname == "") return vngpname; // -ngpname指定なし
  ostringstream ostr;
  if (ino < 10) ostr << vngpname << "0" << ino << ".dat";
  else ostr << vngpname << ino << ".dat";
  vngpnamen = (string)(ostr.str());
  return vngpnamen;             // 番号付きファイル名を返す
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngpitr = vngpmap.find(code);
  if (ngpitr == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "b")            // ブランク扱い文字列指定
  {
    vblank = BLANKs;
    if (opt[1].size() > 0) vblank = strtodstr(opt[1]);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "i")       // 集計項目指定
  {
    if (getrange(ino.first, ino.second, opt[1]) < 1) err = badopt(vpname, optstr, USAGE);
    else if (ino.first < 1 || ino.first > ino.second) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル基本名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
  }
  else if (opt[0] == "s")       // 文字列として集計指定
  {
    if (opt[1].empty() || opt[1] == "+") vstr = true;
    else if (opt[1] == "-") vstr = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "sep")     // 項目ごとの集計指定
  {
    if (opt[1].empty() || opt[1] == "+") vsep = true;
    else if (opt[1] == "-") vsep = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "tab")     // 度数表の種類指定
  {
    if (opt[1].empty() || opt[1] == "+" || opt[1] == "fr") vtab = tabfreq;
    else if (opt[1] == "-") vtab = tabnon;
    else if (opt[1] == "btest") vtab = tabdata;
    else if (opt[1] == "code") vtab = tabcode;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "w")       // 列データの出力桁数指定
  {
    char* ptr = nullptr;
    vw = (int)(strtol(opt[1].c_str(), &ptr, 10));
    if (abs(vw) < 6 || Wmax < abs(vw)) err = badopt(vpname, optstr, USAGE, -Wmax, Wmax);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのチェックとセット
{
  if ((int)ni < ino.first || (ino.second != INT_MAX && (int)ni < ino.second)) return disperr(vpname, "Wrong item number for -i option !");
  if (set_ngp() < 0) return -2; // -ngpオプションのエラー
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["fr"] = vngpmap["pi"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["fr"] = vngpmap["pi"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngpitr = vngpmap.find(ngp);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

class varclass                  // 変数クラス
{
private:
  int vni;                      // 項目数
  int vwmax;                    // データの最大桁数
  mapsdi vfreqmap;              // 度数データマップ
public:
  varclass(size_t ni) : vni(ni), vwmax(0), vfreqmap() // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int ni() const                // 項目数を返す
  { return vni; }
  int wmax() const              // データの最大桁数を返す
  { return vwmax; }
  const mapsdi& freqmap() const // 度数データマップ返す
  { return vfreqmap; }
  int sum(const option& opt, ldata& datalst, int iidx); // データを累積する
};

int varclass::sum(const option& opt, ldata& datalst, int iidx)
// データを累積する
{
  vfreqmap.clear();
  int istart = opt.is() - 1, iend = min(opt.ie(), vni);
  if (opt.sep()) { istart = iidx;  iend = iidx + 1; } // 項目ごとに集計
  vwmax = 0;
  for (auto& cdata : datalst)
  {
    for (int i = istart; i < iend; ++i)
    {
      if (cdata.sdata(i) != opt.blank())
      {
        double fdata = cdata.fdata(i);
        pair<string, double> tdata("", fdata);
        if (opt.str()) tdata.first = cdata.sdata(i); // 文字として集計
        mapsdi_itr fitr = vfreqmap.find(tdata);
        if (fitr == vfreqmap.end()) // 新規データ
        {
          vfreqmap.insert(mapsdi::value_type(tdata, 1));
          if (opt.str())
          { vwmax = max(vwmax, strclen(tdata.first, CHARwide)); }
          else
          {
            double fd = fabs(fdata);
            if (fd > 0.0 && log10(fd) >= 6.0) vwmax = WMAXfd; // 浮動小数点表示の最大桁数
          }
        }
        else                    // 既存データ
        { ++(fitr->second); }
      }
    }
  }
  return (int)(vfreqmap.size()); // 度数の数を返す
}

class freqclass                 // freqメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, const option& opt, idd& id, int iidx) const;
  // 度数表のタイトルとデータ識別情報を出力する
  void put_btest(ostream& fout, const option& opt, const varclass& var) const;
  // btest用度数分布表を出力する
  void put_code(ostream& fout, const option& opt, const varclass& var) const;
  // コードだけ出力する
  void put_freq(ostream& fout, const option& opt, const idd& id, const varclass& var) const;
  // 度数分布表用を出力する
  void plot_ngp(ofstream& fout, option& opt, idd& id, int ino, const varclass& var) const;
  // ngraph用データを出力する
  void plot_ngpid(ofstream& fout, const option& opt, idd& id, int ino) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_freq(ofstream& fout, const varclass& var) const;
  // Ngraph用に度数分布データを出力する
  void plot_pi(ofstream& fout, const varclass& var) const;
  // Ngraph用にパイチャートデータを出力する
public:
  freqclass() : fname()         // コンストラクタ
  {}
  ~freqclass()                  // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込むる
  void put(option& opt, idd& id, ldata& datalst) const;
  // 集計結果を出力する
};

void freqclass::put_id(ostream& fout, const option& opt, idd& id, int iidx) const
// 度数表のタイトルとデータ識別情報を出力する
{
  int istart = opt.is(), iend = min(opt.ie(), (int)(id.ni()));
  if (opt.sep()) { istart = iidx + 1;  iend = iidx + 1; } // 項目ごとに集計
  switch (opt.tab())
  {
  case tabdata : id.put_title(fout, formtag); // btest用
    id.put_stra(fout, formtag);
    for (int i = istart; i <= iend; ++i) id.put_item(fout, i - 1, formtag);
    break;
  default : puttitle(fout, Title); // 度数分布またはコードのみ用
    fout << "\n";
    id.put_title(fout);
    id.put_stra(fout);
    for (int i = istart; i <= iend; ++i) id.put_item(fout, i - 1);
  }
}

void freqclass::put_btest(ostream& fout, const option& opt, const varclass& var) const
// btest用度数分布表を出力する
{
  datac tdata(2);
  ostringstream ostr;
  int w = max(abs(opt.w()), var.wmax());
  string tstr("DATA");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  ostr << left << setw(w + npad) << tstr << right;
  tdata.assign(0, ostr.str());
  tdata.assign(1, "       n");
  tdata.put(fout);
  for (mapsdi_citr fitr = (var.freqmap()).begin(); fitr != (var.freqmap()).end(); ++fitr)
  {
    ostringstream ostr;
    if (opt.str())              // 文字として集計
    { ostr << fitr->first.first; }
    else if (fitr->first.second == BLANKf) // 数字にできない文字データ
    { ostr << "(strings)"; }
    else                        // 数字として集計
    { ostr << fitr->first.second; }
    tdata.assign(0, ostr.str());
    tdata.assign(1, (double)(fitr->second));
    tdata.put(fout);
  }
}

void freqclass::put_code(ostream& fout, const option& opt, const varclass& var) const
// コードだけ出力する
{
  for (mapsdi_citr fitr = (var.freqmap()).begin(); fitr != (var.freqmap()).end(); ++fitr)
  {
    if (opt.str())
    { fout << fitr->first.first; }
    else
    {
      if (fitr->first.second != BLANKf) fout << fitr->first.second;
    }
    fout << endl;
  }
}

void freqclass::put_freq(ostream& fout, const option& opt, const idd& id, const varclass& var) const
// 度数分布表用を出力する
{
  int w = max(abs(opt.w()), var.wmax());
  constexpr int Wn = 8, Wpro = 12, Wadd = 3; // 件数と%の出力桁数と追加桁数
  string tstr("DATA");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << "\n" << left << setw(w + npad) << tstr << right;
  fout << "   CODES (CODES/CASES%)\n";
  putline(fout, w + Wn + Wpro + Wadd);
  int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
  int nt = 0;
  for (mapsdi_citr fitr = (var.freqmap()).begin(); fitr != (var.freqmap()).end(); ++fitr)
  {
    if (opt.w() < 0) fout << left;
    if (opt.str())
    {
      npad = fitr->first.first.size() - strclen(fitr->first.first, CHARwide);
      fout << setw(w + npad) << fitr->first.first;
    }
    else if (fitr->first.second == BLANKf)
    {
      tstr = "(strings)";
      npad = tstr.size() - strclen(tstr, CHARwide);
      fout << setw(w + npad) << tstr;
    }
    else
    { fout << setw(w) << fitr->first.second; }
    if (opt.w() < 0) fout << right;
    fout << setw(Wn) << fitr->second;
    fout.precision(3);  fout.setf(ios::fixed);
    fout << " (" << setw(Wpro) << (double)(fitr->second) / (double)(id.nc()) * 100.0 << ")\n";
    fout.precision(prec0);  fout.unsetf(ios::fixed);
    nt += fitr->second;
  }
  putline(fout, w + Wn + Wpro + Wadd);
  tstr = "TOTAL";
  npad = tstr.size() - strclen(tstr, CHARwide);
  fout << left << setw(w + npad) << tstr;
  fout << right << setw(Wn) << nt << "  cases=" << id.nc() << endl;
}

void freqclass::plot_ngp(ofstream& fout, option& opt, idd& id, int iidx, const varclass& var) const
// Ngraph用データを出力する
{
  bool delml = false;
  if (opt.ngp("fr"))            // 度数分布
  {
    fout << "-s1 -drectangle_fill -x1 -y3 -c255,0,0 -zx0,";
    fout << (var.freqmap()).size() + 1 << ",1\n";
    plot_ngpid(fout, opt, id, iidx);
    plot_freq(fout, var);
    delml = true;
  }
  if (opt.ngp("pi"))            // パイチャート
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dline -zx-1,1,0 -zy-1,1,0\n";
    plot_ngpid(fout, opt, id, iidx);
    plot_pi(fout, var);
  }
}

void freqclass::plot_ngpid(ofstream& fout, const option& opt, idd& id, int iidx) const
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  int istart = opt.is(), iend = min(opt.ie(), (int)(id.ni()));
  if (opt.sep()) { istart = iidx + 1;  iend = iidx + 1; } // 項目ごとに集計
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  for (int i = istart; i <= iend; ++i) id.put_item(fout, i - 1, formngp);
}

void freqclass::plot_freq(ofstream& fout, const varclass& var) const
// Ngraph用に度数分布データを出力する
{
  int i = 0;
  for (mapsdi_citr fitr = (var.freqmap()).begin(); fitr != (var.freqmap()).end(); ++fitr)
  {
    fout << (double)i + 0.7 << " 0 " << (double)i + 1.3 << " " << fitr->second << endl;
    ++i;
  }
}

void freqclass::plot_pi(ofstream& fout, const varclass& var) const
// Ngraph用にパイチャートデータを出力する
{
  constexpr uint RDdef = 48;    // 座標軸の分割数
  double dh = 2.0 * M_PI / (double)RDdef;
  double h = 0.0;
  fout << "%<AXIS OF POLAR COORDINATES>\n";
  for (uint i = 0; i <= RDdef; ++i, h += dh) fout << cos(h) << " " <<  sin(h) << "\n";
  int nf = 0, nt = 0;
  for (mapsdi_citr fitr = (var.freqmap()).begin(); fitr != (var.freqmap()).end(); ++fitr) nt += fitr->second;
  int n = 0;
  h = 0.0;
  fout << NGPdelm << "\n";
  fout << "%<PROPORTION DATA>\n";
  for (mapsdi_citr fitr = (var.freqmap()).begin(); fitr != (var.freqmap()).end(); ++fitr)
  {
    fout << "= =\n";
    fout << "0 0\n";
    n += fitr->second;
    h = M_PI * (0.5 - 2.0 * (double)n / (double)nt);
    fout << cos(h) << " " << sin(h) << endl;
  }
}

int freqclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込むる
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;                    // 読み込んだケース数を返す
}

void freqclass::put(option& opt, idd& id, ldata& datalst) const
// 集計結果を出力する
{
  if (id.nc() < 1) return;      // データなし
  varclass var(id.ni());        // 変数オブジェクト
  int istart = 0, iend = 1;
  if (opt.sep())                // 項目ごとに集計
  { istart = opt.is() - 1, iend = min(opt.ie(), (int)(id.ni())); }
  for (int i = istart; i < iend; ++i)
  {
    if (var.sum(opt, datalst, i) > 0) // データあり
    {
      switch (opt.tab())
      {
      case tabdata :            // btest用
        if (i > istart) cout << "\n";
        put_id(cout, opt, id, i);
        put_btest(cout, opt, var);
        break;
      case tabcode : put_code(cout, opt, var); // コードのみ
        break;
      case tabfreq : put_id(cout, opt, id, i); // 度数表形式
        put_freq(cout, opt, id, var);
        cout << "\n" << endl;
      }
      if (opt.ngpname(i + 1) != "") // Ngraph用データ出力
      {
        ofstream foutngp;
        if (ofopen(foutngp, (opt.ngpname(i + 1)).c_str()) == false) return;
        plot_ngp(foutngp, opt, id, i, var);
        foutngp.close();
      }
    }
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  freqclass freq;               // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) freq.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { freq.get(argv[i], opt, id, datalst);  fok = false; }
    }
    freq.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
