// friede.C Version 7.42e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-26 16:40:55 tac>
//
// 対応のある多項目(時期等)の順序データについてFriedmanの検定を行う。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W friede friede.C
//                -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "friede  Version 7.42e        Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:friede  [<OPTION> ...]  [<DATA FILE> ...]",
  "  The Friedman test and Cochran's Q-test.",
  "<OPTION>",
  "  -f=<FILE>:read options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -res=[-][{def|all|cc[<C-COEF>]|ft|k|m[{b|c}][{b|d|f|s|t}]|r",
  "    |tab[{fr|pr}]}][,...]:select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    cc[<C-COEF>]:confidence coefficient(0=<C-COEF<1 cc:cc0.95)",
  "    ft:Friedman test  k:matching coefficient kappa",
  "    m[{b|c}][{b|d|f|s|t}]:multiple comparison",
  "      mb:comparisons of first item vs other item",
  "      mc:all comparisons of every two item",
  "      b:Bonferroni type  f:Fisher type  d:Dunnett type  s:Scheffe type",
  "      t:Tukey type",
  "    r:Kendall's coefficient of concordance W",
  "    tab[{fr|pr}]:frequency table",
  "      fr:only frequency  pr:also output % for total",
  "    default:auto judgment according to data",
  "  -w=<NUMBER>:specify the number of output digits of data(minimum:1)"
};

const string Title = "=== Comparison of Correspondingly Ordinal Data ==="; // 表題
constexpr int Wdef = 4;         // 省略時出力桁数
constexpr int Wmax = LENcrt / 2; // 最大出力桁数

using pairdblint = pair<double, int>; // 実数と整数ペアの別名定義
using mapdv = map<double, vecint>;    // 度数データマップの別名定義
using mapdv_itr = mapdv::iterator;    // 反復子の別名定義
using mapdv_citr = mapdv::const_iterator; // const反復子の別名定義

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  mctype vmb;                   // 最初の項目との多重比較法
  mctype vmc;                   // 全項目間の多重比較法
  lstr vreslst;                 // 結果出力リスト
  map<string, bool> vresmap;    // 結果出力の有無マップ
  tabtype vtab;                 // 度数集計表の種類
  int vw;                       // データの出力桁数
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  mctype mb() const             // 最初の項目との多重比較法を返す
  { return vmb; }
  mctype mc() const             // 全項目間の多重比較法を返す
  { return vmc; }
  bool res(const string& code); // 解析結果出力フラッグを返す
  tabtype tab() const           // 度数集計表の種類を返す
  { return vtab; }
  int w() const                 // データの出力桁数を返す
  { return vw; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オプションのチェックとセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vmb(mcdun), vmc(mctuk), vreslst(), vresmap(), vtab(tabfprr), vw(Wdef)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "cc")           // 信頼係数の指定
  {
    vcc = opt[1].empty() ? CCdef : strtodf(opt[1]);
    if (vcc < 0.0 || 1.0 <= vcc) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "res")     // 解析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "w")       // データの出力桁数指定
  {
    char* ptr = nullptr;
    if (opt[1].empty()) vw = Wdef;
    else vw = int(strtol(opt[1].c_str(), &ptr, 10));
    if (vw < 1 || Wmax < vw) err = badopt(vpname, optstr, USAGE, 1, Wmax);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(size_t ni)    // オプションのセットとチェック
{
  if (ni < 2) return disperr(vpname, "At least 2 items are required !");
  if (set_res() < 0) return -2; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["ft"] = vresmap["tab"] = true;
  vresmap["k"] = vresmap["mb"] = vresmap["mc"] = vresmap["r"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["ft"] = vresmap["tab"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      { vresmap["def"] = vresmap["cc"] = vresmap["ft"] = vresmap["tab"] = true; }
      else if (res == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["ft"] = vresmap["k"] = vresmap["mb"] = vresmap["mc"] = vresmap["r"] = vresmap["tab"] = true;
        vmb = mcdun;  vmc = mctuk;  vtab = tabfprr;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (res == "m")      // -res=mと指定された時
      { vresmap["mb"] = vresmap["mc"] = true; }
      else if (res[0] == 'm' && res[1] == 'b' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vmb = mcbon;
          break;
        case 'f' : vmb = mcfis;
          break;
        case 's' : vmb = mcsch;
          break;
        case 't' : vmb = mctuk;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " option !");
        }
        vresmap["mb"] = true;
      }
      else if (res[0] == 'm' && res[1] == 'c' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vmc = mcbon;
          break;
        case 'd' : vmc = mcdun;
          break;
        case 'f' : vmc = mcfis;
          break;
        case 's' : vmc = mcsch;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " option !");
        }
        vresmap["mc"] = true;
      }
      else if (res == "tabfr") // 列合計に対する度数だけ出力
      { vresmap["tab"] = true;  vtab = tabfreq; }
      else if (res == "tabpr") // 列合計に対する％も出力
      { vresmap["tab"] = true;  vtab = tabfprr; }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class varclass                  // 変数クラス
{
private:
  int vni;                      // 項目数
  int vnc;                      // 集計したケース数
  int vnk;                      // 完全一致例数(一致係数κ用)
  mapdv vfriedmap;              // 度数データマップ
  vecdbl trvec;                 // 順位合計配列
  double vkt;                   // 多重比較における同位補正
  double vcsqr;                 // χ2乗値
  int vdf;                      // 自由度
  int sum_rank(vecdbl& trivec, double& sigt, vecdbl& fdatavec, bool fmap);
  // ブロックごとの順位を集計する
  void sort_rank(vector<pairdblint>& blockvec) const;
  // 順位とインデックスのソート(挿入法)
public:
  varclass(size_t ni) : vni(ni), vnc(0), vnk(0), vfriedmap(), trvec(ni, 0.0), vkt(0.0), vcsqr(0.0), vdf(0) // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  int ni() const                // 項目数を返す
  { return vni; }
  int nc() const                // ケース数を返す
  { return vnc; }
  int nk() const                // 完全一致例数を返す
  { return vnk; }
  double tr(size_t idx) const   // 順位和を返す
  { return (idx < trvec.size()) ? trvec[idx] : 0.0; }
  double kt() const             // 多重比較における同位補正を返す
  { return vkt; }
  double csqr() const           // χ2乗値を返す
  { return vcsqr; }
  int df() const                // 自由度を返す
  { return vdf; }
  const mapdv& friedmap() const // 度数データマップを返す
  { return vfriedmap; }
  int cal(const option& opt, ldata& datalst);
  // データを集計して各種統計量を計算する
  double witem(int& ni, const idd& id, ldata& datalst);
  // 項目間の一致係数Wを計算して返す
};

int varclass::sum_rank(vecdbl& trivec, double& sigti, vecdbl& fdatavec, bool fmap)
// ブロックごとの順位を集計する
{
  vector<pairdblint> blockvec(fdatavec.size());
  for (size_t j = 0; j < fdatavec.size(); ++j)
  {
    blockvec[j].first = fdatavec[j];  blockvec[j].second = j;
    if (fmap)                   // 度数データマップ作成
    {
      mapdv_itr fitr = vfriedmap.find(blockvec[j].first);
      if (fitr == vfriedmap.end()) // 新規データ
      {
        vecint nvec(fdatavec.size(), 0);
        nvec[j] = 1;
        vfriedmap.insert(map<double, vecint>::value_type(blockvec[j].first, nvec));
      }
      else                        // 既存データ
      { ++(fitr->second[j]); }
    }
  }
  sort_rank(blockvec);
  int nkok = 0;
  size_t i = 0;
  while (i < fdatavec.size())
  {
    size_t j = i + 1;
    while (j < fdatavec.size() && blockvec[i].first == blockvec[j].first) ++j;
    double ti = (double)(j - i);
    double r = (double)i + (ti + 1.0) * 0.5;
    sigti += (ti - 1.0) * ti * (ti + 1.0);
    for (size_t k = i; k < j; ++k) trivec[blockvec[k].second] += r;
    if (i == 0 && j >= fdatavec.size()) nkok = 1; // 完全一致例
    i = j;
  }
  return nkok;                  // 完全一致例は1を返す
}

void varclass::sort_rank(vector<pairdblint>& blockvec) const
// 順位とインデックスのソート(挿入法)
{
  pairdblint blocktmp;
  for (size_t i = 1; i < blockvec.size(); ++i)
  {
    blocktmp = blockvec[i];
    int j = (int)i - 1;
    while (j >= 0 && blockvec[j].first > blocktmp.first)
    {
      blockvec[j + 1] = blockvec[j];
      --j;
    }
    blockvec[j + 1] = blocktmp;
  }
}

int varclass::cal(const option& opt, ldata& datalst)
// データを集計して各種統計量を計算する
{
  vecdbl fdatavec(vni, 0.0);
  vnk = vnc = 0;
  double sigt = 0.0;
  for (auto& cdata : datalst)
  {
    bool complete = true;
    for (int j = 0; complete && j < vni; ++j)
    {
      if (cdata.fdata(j) == BLANKf) complete = false;
      else fdatavec[j] = cdata.fdata(j);
    }
    if (complete)               // 完全症例のみ集計
    { vnk += sum_rank(trvec, sigt, fdatavec, true);  ++vnc; }
  }
  vdf = vni - 1;
  double tg2 = (double)vni * (double)vnc * (double)(vni + 1);
  double tg = 0.5 * tg2;
  sigt = 1.0 - sigt / tg2 / (double)vdf;
  double eps = DBL_EPSILON * fmax(5.0, tg);
  if (sigt < eps) sigt = 0.0;
  double sigt2 = 0.0;
  for (int j = 0; j < vni; ++j) sigt2 += trvec[j] * trvec[j];
  vcsqr = (sigt > 0.0) ? 6.0  / sigt * (sigt2 / tg - tg / (double)vni) : 0.0;
  vkt = sigt * tg;
  return (int)(vfriedmap.size()); // 度数データマップ数を返す
}

double varclass::witem(int& ni, const idd& id, ldata& datalst)
// 項目間の一致係数Wを計算して返す
{
  vecdbl fdatavec(id.nc(), 0.0);
  vecdbl trivec(id.nc(), 0.0);
  double sigti = 0.0;
  ni = 0;
  for (size_t j = 0; j < id.ni(); ++j)
  {
    bool complete = true;
    int i = 0;
    for (ldata_itr ditr = datalst.begin(); complete && ditr != datalst.end(); ++ditr, ++i)
    {
      if (ditr->fdata(j) == BLANKf) complete = false;
      else fdatavec[i] = ditr->fdata(j);
    }
    if (complete)               // 完全症例のみ集計
    { sum_rank(trivec, sigti, fdatavec, false);  ++ni; }
  }
  if (ni < 2) return 0.0;       // 一致係数計算できず
  double tg2 = (double)(id.nc()) * (double)ni * (double)(id.nc() + 1);
  double tg = 0.5 * tg2;
  int df = id.nc() - 1;
  sigti = 1.0 - sigti / tg2 / (double)df;
  double eps = DBL_EPSILON * fmax(5.0, tg);
  if (sigti < eps) sigti = 0.0;
  double sigt2 = 0.0;
  for (int i = 0; i < vnc; ++i) sigt2 += trivec[i] * trivec[i];
  double csqr = (sigti > 0.0) ? 6.0  / sigti * (sigt2 / tg - tg / (double)(id.nc())) : 0.0;
  return csqr / (double)ni / (double)df; // 一致係数wを返す
}

class friedclass                // friedメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, idd& id) const;
  // タイトルとデータ識別情報を出力する
  void put_ftable(ostream& fout, const option& opt, const idd& id, const varclass& var) const;
  // 順序データ集計表を出力する
  void put_kappa(ostream& fout, const idd& id, const varclass& var) const;
  // 一致係数κを出力する
  void put_r(ostream& fout, const idd& id, ldata& datalst, varclass& var) const;
  // Kendallの一致係数Wを出力する
  void put_mbcon(ostream& fout, const option& opt, const idd& id, const varclass& var) const;
  // 最初の項目との多重比較を出力する
  void put_mbrdci(ostream& fout, const option& opt, const idd& id, const varclass& var) const;
  // 最初の項目との順位差の信頼区間を出力する
  void put_mccon(ostream& fout, const option& opt, const idd& id, const varclass& var) const;
  // 全項目間の多重比較を出力する
  void put_mcrdci(ostream& fout, const option& opt, const idd& id, const varclass& var) const;
  // 全項目間の順位差の信頼区間を出力する
public:
  friedclass() : fname()        // コンストラクタ
  {}
  ~friedclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込み累積する
  void put(option& opt, idd& id, ldata& datalst) const;
  // 集計結果を出力する
};

void friedclass::put_id(ostream& fout, idd& id) const
// タイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put(fout);
}

void friedclass::put_ftable(ostream& fout, const option& opt, const idd& id, const varclass& var) const
// 順序データ集計表を出力する
{
  constexpr int Wpr = 5;        // %の出力桁数
  bool putpr = (opt.tab() == tabfprr) ? true : false;
  int prlen = putpr ? Wpr + 2 : 0; // %の出力幅
  size_t ngrade = (var.friedmap()).size();
  fout << "\nDATA  ";
  for (auto& fivec : var.friedmap())
  {
    fout << setw(opt.w()) << fivec.first;
    if (putpr)                  // %出力あり
    {
      fout << "(" << setw(Wpr) << "%  ";
      fout << ")";
    }
  }
  int wtn = max(opt.w(), 5), wtr = max(opt.w(), 8);
  string tstr("TOTAL");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  fout << setw(wtn + npad) << tstr;
  if (putpr)                    // %出力あり
  {
    fout << "(" << setw(Wpr) << "%  " << ")";
  }
  if ((var.friedmap()).size() < 3)
  { wtr = 0; }
  else
  {
    tstr = "RNK MEAN";
    npad = tstr.size() - strclen(tstr, CHARwide);
    fout << setw(wtr + npad) << tstr;
  }
  fout << "\n";
  int w = 6 + (opt.w() + prlen) * ngrade + wtn + prlen + wtr;
  putline(fout, w);
  vecint tnvec(ngrade, 0);
  double ttr = 0.0;
  for (size_t j = 0; j < id.ni(); ++j)
  {
    fout << "ITEM" << setw(2) << j + 1;
    int i = 0;
    for (auto& fivec : var.friedmap())
    {
      fout << setw(opt.w()) << fivec.second[j];
      tnvec[i] += fivec.second[j];
      if (putpr)                // %出力
      {
        double pr = (double)(fivec.second[j]) / (double)(var.nc());
        int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
        fout.setf(ios::fixed);
        fout << "(" << setw(Wpr) << setprecision(1) << pr * 100.0;
        fout << setprecision(prec0) << ")";
        fout.unsetf(ios::fixed);
      }
      ++i;
    }
    fout << setw(wtn) << var.nc();
    if (putpr)                  // %出力
    {
      int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
      fout.setf(ios::fixed);
      fout << "(" << setw(Wpr) << setprecision(1) << 100.0;
      fout << setprecision(prec0) << ")";
      fout.unsetf(ios::fixed);
    }
    if ((var.friedmap()).size() > 2) fout << setw(wtr) << var.tr(j) / (double)(var.nc());
    fout << "\n";
    ttr += var.tr(j);
  }
  putline(fout, w);
  int ttn = var.nc() * id.ni();
  fout << "TOTAL ";
  for (size_t i = 0; i < ngrade; ++i)
  {
    fout << setw(opt.w()) << tnvec[i];
    if (putpr)                  // 行合計の%出力
    {
      double pr = (ttn > 0) ? (double)(tnvec[i]) / (double)ttn : 0.0;
      int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
      fout.setf(ios::fixed);
      fout << "(" << setw(Wpr) << setprecision(1) << pr * 100.0;
      fout << setprecision(prec0) << ")";
      fout.unsetf(ios::fixed);
    }
  }
  fout << setw(wtn) << ttn;
  if (putpr)                    // %出力
  {
    int prec0 = fout.precision(); // 現在の小数点以下の精度を保存する
    fout.setf(ios::fixed);
    fout << "(" << setw(Wpr) << setprecision(1) << 100.0;
    fout << setprecision(prec0) << ")";
    fout.unsetf(ios::fixed);
  }
  if ((var.friedmap()).size() > 2) fout << setw(wtr) << ttr / (double)ttn;
  fout << endl;
}

void friedclass::put_kappa(ostream& fout, const idd& id, const varclass& var) const
// 一致係数κを出力する
{
  double pc = 0.0, v = 0.0, g = (double)(var.nc());
  for (auto& fivec : var.friedmap())
  {
    double paipi = 1.0, sigpi = 0.0;
    for (size_t j = 0; j < id.ni(); ++j)
    {
      double pi = (double)(fivec.second[j]) / g;
      paipi *= pi;  sigpi += pi;
    }
    pc += paipi;  v += paipi * sigpi;
  }
  double po = (double)(var.nk()) / g;
  double k = 0.0;
  if (sgn(1.0 - pc) == 0)
  { v = 0.0; }
  else
  {
    k = (po - pc) / (1.0 - pc);
    v = (pc + pc * pc - v) / g / (1.0 - pc) / (1.0 - pc);
    if (fabs(k) < DBL_EPSILON) k = 0.0;
    if (fabs(v) < DBL_EPSILON) v = 0.0;
  }
  fout << "\nMatching Coefficient Kappa=" << k << "\n";
  if (id.ni() != 2) return;
  double z = (v > 0.0) ? k / sqrt(v) : 0.0;
  ldbl pr = ztop(z);
  fout << "  Test of Kappa:z=" << z << "  p=" << pr << pind(pr) << endl;
}

void friedclass::put_r(ostream& fout, const idd& id, ldata& datalst, varclass& var) const
// Kendallの一致係数Wを出力する
{
  double wc = (var.nc() < 1 || var.df() < 1) ? 0.0 : var.csqr() / (double)(var.nc()) / (double)(var.df());
  double rs = (var.nc() < 2) ? 0.0 : ((double)(var.nc()) * wc - 1.0) / (double)(var.nc() - 1);
  double f = 1.0 + (double)(var.nc() - 1) * rs;
  double raa = (sgn(f) == 0) ? 0.0 : (double)(var.nc()) * rs / f;
  fout << "\nKendall's Coefficient of Concordance\n";
  fout << "  Inter-case:W=" << wc;
  fout << "  Mean of rs:mrs=" << rs << "  Reliability-c:r" << var.nc() << var.nc() << "=" << raa << "\n";
  int ni = 0;
  double wi = var.witem(ni, id, datalst);
  rs = (ni < 2) ? 0.0 : ((double)ni * wi - 1.0) / (double)(ni - 1);
  f = 1.0 + (double)(ni - 1) * rs;
  raa = (sgn(f) == 0) ? 0.0 : (double)ni * rs / f;
  fout << "  Inter-item:W=" << wi;
  fout << "  Mean of rs:mrs=" << rs << "  Reliability-c:r" << ni << ni << "=" << raa << endl;
}

void friedclass::put_mbcon(ostream& fout, const option& opt, const idd& id, const varclass& var) const
// 最初の項目との多重比較を出力する
{
  int df = 1, dc = 1;
  string method = "DUNNETT";
  switch (opt.mb())
  {
  case mcbon : method = "BONFERRONI";  dc = id.ni() - 1;
    break;
  case mcfis : method = "FISHER";
    break;
  case mcsch : method = "SCHEFFE";  df = id.ni() - 1;
    break;
  }
  fout << "\n<" << method << " TYPE MULTIPLE COMPARISON OF FIRST ITEM VS OTHER ITEM>\n";
  int llen = 67;                // 区切り線の長さ
  if ((var.friedmap()).size() < 3) fout << "ITEM: Proport - ITEM: Proport";
  else fout << "ITEM:RankMean - ITEM:RankMean";
  if (opt.mb() == mcdun) { fout << "      d-vlue   Items";  llen += 8; }
  else fout << "  Chi-squared";
  fout << "      DF       p-value\n";
  putline(fout, llen);
  double tr0 = var.tr(0), ncd = (double)(var.nc());
  mapdv_citr fitr = (var.friedmap()).begin();
  ++fitr;
  double tt0 = (double)(fitr->second[0]);
  for (size_t i = 1; i < id.ni(); ++i)
  {
    double tri = var.tr(i);
    double csqr = (var.kt() > 0.0) ? 3.0 * (tr0 - tri) * (tr0 - tri) / var.kt() : 0.0;
    double pr = calmcpr(opt.mb(), csqr, id.ni(), df, 0, dc);
    double ti = ((var.friedmap()).size() < 3) ? (double)(fitr->second[i]) : tri;
    fout << right << setw(4) << i + 1 << ":" << setw(8) << ti / ncd;
    double t0 = ((var.friedmap()).size() < 3) ? tt0 : tr0;
    fout << " - " << setw(4) << 1 << ":" << setw(8) << t0 / ncd;
    if (opt.mb() == mcdun) fout << setw(12) << csqr << setw(8) << id.ni() << setw(8) << "inf";
    else fout << setw(12) << csqr << setw(8) << df;
    fout << setw(14) << pr << pind(pr) << "\n";
  }
  putline(fout, llen);
}

void friedclass::put_mbrdci(ostream& fout, const option& opt, const idd& id, const varclass& var) const
// 最初の項目との順位差の信頼区間出力
{
  int df = 1, dc = 1;
  string method = "DUNNETT";
  switch (opt.mb())
  {
  case mcbon : method = "BONFERRONI";  dc = id.ni() - 1;
    break;
  case mcfis : method = "FISHER";
    break;
  case mcsch : method = "SCHEFFE";  df = id.ni() - 1;
  }
  fout << "\n<" << method << " TYPE " << opt.cc() * 100.0 << "% SIMULTANEOUS CI OF FIRST ITEM VS OTHER ITEM>\n";
  fout << "ITEM    - ITEM    ";
  if ((var.friedmap()).size() < 3) fout << " Diff of Pro";
  else fout << "Diff of Mrnk";
  fout << "   Precision   Lower-Lim   Upper-Lim\n";
  const int LENline = 66;       // 区切り線の長さ
  putline(fout, LENline);
  double tr0 = var.tr(0), ncd = (double)(var.nc()), b = (double)(id.ni());
  double eps = DBL_EPSILON * fmax(5.0, ncd);
  for (size_t i = 1; i < id.ni(); ++i)
  {
    double tri = var.tr(i);
    double dm = (tri - tr0) / ncd;
    if (fabs(dm) < eps) dm = 0.0;
    double vd = var.kt() / 3.0 / ncd / ncd;
    double ci = calmcci(opt.mb(), opt.cc(), id.ni(), df, 0, dc, vd);
    if ((var.friedmap()).size() < 3) { dm = dm * 2.0 / b;  ci = ci * 2.0 / b; }
    fout << left << setw(8) << i + 1 << "- " << setw(8) << 1;
    fout << right << setw(12) << dm;
    fout << setw(12) << ci << setw(12) << dm - ci << setw(12) << dm + ci;
    fout << "\n";
  }
  putline(fout, LENline);
}

void friedclass::put_mccon(ostream& fout, const option& opt, const idd& id, const varclass& var) const
// 全項目間の多重比較を出力する
{
  int df = 1, dc = 1;
  string method = "TUKEY";
  switch (opt.mc())
  {
  case mcbon : method = "BONFERRONI";  dc = id.ni() * (id.ni() - 1) / 2;
    break;
  case mcfis : method = "FISHER";
    break;
  case mcsch : method = "SCHEFFE";  df = id.ni() - 1;
  }
  fout << "\n<" << method << " TYPE MULTIPLE COMPARISON OF EVERY TOW ITEM>\n";
  int llen = 67;                // 区切り線の長さ
  if ((var.friedmap()).size() < 3) fout << "ITEM: Proport - ITEM: Proport";
  else fout << "ITEM:RankMean - ITEM:RankMean";
  if (opt.mb() == mcdun) { fout << "      d-vlue   Items";  llen += 8; }
  else fout << "  Chi-squared";
  fout << "      DF       p-value\n";
  putline(fout, llen);
  double ncd = (double)(var.nc());
  mapdv_citr fitr = (var.friedmap()).begin();
  ++fitr;
  for (size_t i = 0; i < id.ni() - 1; ++i)
  {
    double tri = var.tr(i);
    double tti = (double)(fitr->second[i]);
    double mi = (((var.friedmap()).size() < 3) ? tti : tri) / ncd;
    for (size_t j = i + 1; j < id.ni(); ++j)
    {
      double trj = var.tr(j);
      double ttj = (double)(fitr->second[j]);
      double csqr = (var.kt() > 0.0) ? 3.0 * (tri - trj) * (tri - trj) / var.kt() : 0.0;
      double pr = calmcpr(opt.mc(), csqr, id.ni(), df, 0, dc);
      double mj = (((var.friedmap()).size() < 3) ? double(fitr->second[j]) : trj) / ncd;
      fout << right << setw(4) << j + 1 << ":" << setw(8) << mj;
      fout << " - " << setw(4) << i + 1 << ":" << setw(8) << mi;
      if (opt.mc() == mctuk) fout << setw(12) << csqr << setw(8) << id.ni() << setw(8) << "inf";
      else fout << setw(12) << csqr << setw(8) << df;
      fout << setw(14) << pr << pind(pr) << "\n";
    }
  }
  putline(fout, llen);
}

void friedclass::put_mcrdci(ostream& fout, const option& opt, const idd& id, const varclass& var) const
// 全項目間の順位差の信頼区間出力
{
  int df = 1, dc = 1;
  string method = "TUKEY";
  switch (opt.mc())
  {
  case mcbon : method = "BONFERRONI";  dc = id.ni() * (id.ni() - 1) / 2;
    break;
  case mcfis : method = "FISHER";
    break;
  case mcsch : method = "SCHEFFE";  df = id.ni() - 1;
  }
  fout << "\n<" << method << " TYPE " << opt.cc() * 100.0 << "% SIMULTANEOUS CI OF  EVERY TOW ITEM>\n";
  fout << "ITEM    - ITEM    ";
  if ((var.friedmap()).size() < 3) fout << " Diff of Pro";
  else fout << "Diff of Mrnk";
  fout << "   Precision   Lower-Lim   Upper-Lim\n";
  const int LENline = 66;       // 区切り線の長さ
  putline(fout, LENline);
  double ncd = (double)(var.nc()), b = (double)(id.ni());
  double eps = DBL_EPSILON * fmax(5.0, ncd);
  for (size_t i = 0; i < id.ni() - 1; ++i)
  {
    double tri = var.tr(i);
    for (size_t j = i + 1; j < id.ni(); ++j)
    {
      double trj = var.tr(j);
      double dm = (trj - tri) / ncd;
      if (fabs(dm) < eps) dm = 0.0;
      double vd = var.kt() / 3.0 / ncd / ncd;
      double ci = calmcci(opt.mc(), opt.cc(), id.ni(), df, 0, dc, vd);
      if ((var.friedmap()).size() < 3) { dm = dm * 2.0 / b;  ci = ci * 2.0 / b; }
      fout << left << setw(8) << j + 1 << "- " << setw(8) << i + 1;
      fout << right << setw(12) << dm;
      fout << setw(12) << ci << setw(12) << dm - ci << setw(12) << dm + ci;
      fout << "\n";
    }
  }
  putline(fout, LENline);
}

int friedclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込み累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time option error !");
  return nc;                    // 読み込んだケース数を返す
}

void friedclass::put(option& opt, idd& id, ldata& datalst) const
// 集計結果を出力する
{
  if (opt.res("res") == false) return;
  put_id(cout, id);
  varclass var(id.ni());        // 変数オブジェクト
  if (var.cal(opt, datalst) < 2 || id.ni() < 2 || id.nc() < 2) // 計算できず
  {
    cout << "\n" << endl;
    return;
  }
  if (opt.res("tab")) put_ftable(cout, opt, id, var);
  if (opt.res("k")) put_kappa(cout, id, var);
  if (opt.res("r")) put_r(cout, id, datalst, var);
  if (opt.res("ft"))
  {
    if ((var.friedmap()).size() < 3) cout << "\nCochran's Q-test:Q=";
    else cout << "\nFriedman Test:";
    ldbl pr = ctop(var.csqr(), var.df());
    cout << "chi-squared=" << var.csqr() << "  DF=" << var.df() << "  p=" << pr << pind(pr) << "\n";
  }
  if (id.ni() > 2)
  {
    if (opt.res("mb"))
    {
      put_mbcon(cout, opt, id, var);
      put_mbrdci(cout, opt, id, var);
    }
    if (opt.res("mc"))
    {
      put_mccon(cout, opt, id, var);
      put_mcrdci(cout, opt, id, var);
    }
  }
  cout << "\n" << endl;
}

int main(int argc, char* argv[]) // メインルーチン
{
  friedclass fried;             // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) fried.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { fried.get(argv[i], opt, id, datalst);  fok = false; }
    }
    fried.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
