// hteste.C Version 7.42e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-26 16:46:09 tac>
//
// クロス集計表を入力し、Kruskal-Wallisの検定とU検定を行う。
// クロス集計表は計数値解析コマンド用の標準入力形式に従っている必要がある。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W hteste hteste.C
//                -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "hteste  Version 7.42e        Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:hteste  [<OPTION> ...]  [<TAB FILE> ...]",
  "  The H-test or U-test with the cross tabulation table.",
  "<OPTION>",
  "  -f=<FILE>:read options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -res=[-][{def|all|bt[{b|f}]|cc[<C-COEF>]|ht|m[{a|d}][{b|d|f|s|t}]|r",
  "    tab[{fr|pc|pr|pt}]|ut[{a|c|o}]}][,...]:select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    bt[{b|f}]:multiple comparison with 2 x 2 contingency table",
  "      b:Bonferroni type  f:Fisher type",
  "    cc[<C-COEF>]:confidence coefficient(0=<C-COEF<1 cc:cc0.95)",
  "    ht:Kruskal-Wallis's H-test",
  "    m[{a|d}][{b|d|f|s|t}]:multiple comparison",
  "      ma:all comparisons of every two group",
  "      md:comparisons of first group vs other group",
  "      b:Bonferroni type  d:Dunnett type  f:Fisher type  s:Scheffe type",
  "      t:Tukey type",
  "    r:rank correlatiion ratio",
  "    tab[={fr|pc|pr|pt}]:cross tabulation table",
  "      fr:only frequency  pc:also output % for column total",
  "      pr:also output % for row total  pt:also output % for all total",
  "    ut[{a|c|o}]:Mann-Whitney's U-test",
  "      a:auto judgement Yates's correction  c:with Yates's correction",
  "      o:no Yates's correction",
  "    default:auto judgment according to data"
};

const string Title = "=== Group Comparison of Ordinal Data ==="; // 表題

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  mctype vbt;                   // 2×2分割表の多重比較法
  double vcc;                   // 信頼係数
  mctype vma;                   // 全ての2群ごとの多重比較法
  mctype vmd;                   // 対照群と他の群の多重比較法
  lstr vreslst;                 // 結果出力リスト
  map<string, bool> vresmap;    // 結果出力の有無マップ
  tabtype vtab;                 // 度数集計表の種類
  testtype vut;                 // U検定の種類
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  mctype bt() const             // 2×2分割表の多重比較法を返す
  { return vbt; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  mctype ma() const             // 全ての2群ごとの多重比較法を返す
  { return vma; }
  mctype md() const             // 対照群と他の群の多重比較法を返す
  { return vmd; }
  bool res(const string& code); // 解析結果出力フラッグを返す
  tabtype tab() const           // 度数集計表出力フラッグを返す
  { return vtab; }
  testtype ut() const           // U検定の種類を返す
  { return vut; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(int nfile);         // オプションのチェックとセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vbt(mcbon), vcc(CCdef), vma(mctuk), vmd(mcdun), vreslst(), vresmap(), vtab(tabfprr), vut(testaut)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "f")            // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "res")     // 解析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check(int nftab)    // オプションのセットとチェック
{
  if (nftab < 1) return disperr(vpname, "no cross tabulation table !");
  if (set_res() < 0) return -2; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["tab"] = true;
  vresmap["bt"] = vresmap["ht"] = vresmap["ma"] = vresmap["md"] = vresmap["r"] = vresmap["ut"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["tab"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      { vresmap["def"] = vresmap["cc"] = vresmap["tab"] = true; }
      else if (res == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["bt"] = vresmap["cc"] = vresmap["ht"] = vresmap["ma"] = vresmap["md"] = vresmap["r"] = vresmap["tab"] = vresmap["ut"] = true;
        vbt = mcbon;;  vma = mctuk;  vmd = mcdun;  vtab = tabfprr;  vut = testaut;
      }
      else if (res == "btb") // Bonferroni型
      { vresmap["bt"] = true;  vbt = mcbon; }
      else if (res == "btf") // Fisher型
      { vresmap["bt"] = true;  vbt= mcfis; }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (res == "m")      // -res=mと指定された時
      { vresmap["ma"] = vresmap["md"] = true; }
      else if (res[0] == 'm' && res[1] == 'a' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vma = mcbon;
          break;
        case 'f' : vma = mcfis;
          break;
        case 's' : vma = mcsch;
          break;
        case 't' : vma = mctuk;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " optiion !");
        }
        vresmap["ma"] = true;
      }
      else if (res[0] == 'm' && res[1] == 'd' && res.size() > 2)
      {
        switch (res[2])
        {
        case 'b' : vmd = mcbon;
          break;
        case 'd' : vmd = mcdun;
          break;
        case 'f' : vmd = mcfis;
          break;
        case 's' : vmd = mcsch;
          break;
        default : return disperr(vpname, "Wrong -res=" + res + " optiion !");
        }
        vresmap["md"] = true;
      }
      else if (res == "tabfr") // 度数だけ出力
      { vresmap["tab"] = true;  vtab = tabfreq; }
      else if (res == "tabpc") // 列合計に対する％出力
      { vresmap["tab"] = true;  vtab = tabfprc; }
      else if (res == "tabpr") // 行合計に対する％出力
      { vresmap["tab"] = true;  vtab = tabfprr; }
      else if (res == "tabpt") // 全合計に対する％出力
      { vresmap["tab"] = true;  vtab = tabfprt; }
      else if (res == "uta")   // U検定(修正自動判定)
      { vresmap["ut"] = true;  vut= testaut; }
      else if (res == "utc")   // U検定(修正有り)
      { vresmap["ut"] = true;  vut= testcor; }
      else if (res == "uto")   // U検定(修正無し)
      { vresmap["ut"] = true;  vut = testorg; }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class htestclass                // htestメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, idd& id) const;
  // タイトルとデータ識別情報を出力する
  void put_rr(ostream& fout, const option& opt, int df, double csqr, const ftable& ftab) const;
  // 順位相関比を出力する
  void put_utest(ostream& fout, const option& opt, vecdbl& tvec, const ftable& ftab) const;
  // U検定の結果を出力する
  void put_bun(ostream& fout, const option& opt, const ftable& ftab) const;
  // 2×2分割表に関する多重比較結果を出力する
  void put_hmcon(ostream& fout, mctype mc, bool ma, double kc, const vecdbl& tvec, const ftable& ftab) const;
  // 順位平均の多重比較を出力する
  void put_hmrdci(ostream& fout, const option& opt, mctype mc, bool ma, double kc, const vecdbl& tvec, const ftable& ftab) const;
  // 2群間の順位平均の差の信頼区間出力
public:
  htestclass() : fname()        // コンストラクタ
  {}
  ~htestclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst);
  // 度数集計表を読み込む
  void put(option& opt, listidd& idlst, listftab& ftablst) const;
  // 結果を出力する
};

void htestclass::put_id(ostream& fout, idd& id) const
// タイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  for (size_t i = 0; i < id.ni(); ++i)
  {
    switch (id.eitem(i).aval())
    {
    case avalrow : fout << "GROUP ITEM(ROW)  :" << id.item(i) << "\n";
      break;
    case avalcol : fout << "ORDINAL ITEM(COL):" << id.item(i) << "\n";
      break;
    default : id.put_item(fout, i);
    }
  }
}

void htestclass::put_rr(ostream& fout, const option& opt, int df, double csqr, const ftable& ftab) const
// 順位相関比を出力する
{
  int dft = ftab.ftotal() - 1;
  double r2 = csqr / (double)dft;
  fout << "\nRank Correlation Ratio:rr=" << sqrt(r2)<< "  rr-squared=" << r2 << "\n";
  double csqrl = 0.0, csqru = 0.0;
  calcsqrci(csqrl, csqru, opt.cc(), df, csqr);
  double rl = sqrt(csqrl / (double)dft);
  double ru = sqrt(csqru / (double)dft);
  if (ru > 1.0) ru = 1.0;
  fout << "  " << opt.cc() * 100.0 << "% CI of rr:Lower=" << rl << "  Upper=" << ru << "\n";
}

void htestclass::put_utest(ostream& fout, const option& opt, vecdbl& tvec, const ftable& ftab) const
// U検定の結果を出力する
{
  double u = 0.0, vu = 0.0, z = 0.0;
  testtype utype = utest(u, vu, z, ftab, opt.ut());
  ldbl pr = ztop(z);
  fout << "\nMann-Whitney's U-test(Wilcoxon's Signed Rank Test";
  if (utype == testcor) fout << ", with Yates's correction";
  fout << ")\n";
  fout << "  z=" << z << "  p=" << pr << pind(pr) << "\n";
  double n1 = (double)(ftab.frtotal(1));
  double r1 = tvec[1] / n1;
  string code1(ftab.rcode(1));
  double n2 = (double)(ftab.frtotal(2));
  double r2 = tvec[2] / n2;
  string code2(ftab.rcode(2));
  fout << "  Rank Mean:[" << code1 << "]=" << r1 << " [" << code2 << "]=" << r2 << "\n";
  double drm = r2 - r1;
  if (sgn(drm) == 0) drm = 0.0;
  double vd = vu / n1 / n2 * (n1 + n2) * (1.0 / n1 + 1.0 / n2);
  double ci = calmcci(mcfis, opt.cc(), ftab.nr(), 1, 0, 1, vd);
  double dl = drm - ci, du = drm + ci;
  fout << "  " << opt.cc() * 100.0 << "%CI of diff of Rank Mean=" << drm << " +/- " << ci << "(" << dl << " - " << du << ")\n";
  double pro = 100.0 / n1 / n2, umax = n1 * n2;
  double u1 = umax - u;
  fout << "  U-value(%):[" << code1 << "]=" << u1 << "(" << u1 * pro << "%)";
  fout << "  [" << code2 << "]=" << u << "(" << u * pro << "%)\n";
  ci = calmcci(mcfis, opt.cc(), ftab.nr(), 1, 0, 1, vu);
  dl = u - ci;  du = u + ci;
  if (dl < 0.0) dl = 0.0;
  if (du > n1 * n2) du = n1 * n2;
  fout << "  " << opt.cc() * 100.0 << "% CI of U[" << code2 << "]=" << u << " +/- " << ci << "(" << dl << "(" << dl * pro << "%) - " << du << "(" << du * pro << "%))\n";
}

void htestclass::put_bun(ostream& fout, const option& opt, const ftable& ftab) const
// 2×2分割表に関する多重比較結果を出力する
{
  int dc = 1;
  string method = "Fisher's Exact Test";
  if (opt.bt() == mcbon)
  {
    method = "Bonferroni Type Multiple Comparison(Fisher's Exact Test)";
    dc = ftab.nc() - 1;
  }
  fout << "\n" << method << "\n";
  int n1 = ftab.frtotal(1), n2 = ftab.frtotal(2);
  int fcum11 = 0, fcum12 = 0, fcum21 = 0, fcum22 = 0;
  string ccodee(ftab.ccode(ftab.nc()));
  strtrm(ccodee);
  for (size_t j = 1; j < ftab.nc(); ++j)
  {
    fcum11 += ftab.f(1, j), fcum21 += ftab.f(2, j);
    fcum12 = n1 - fcum11;  fcum22 = n2 - fcum21;
    ldbl pr1 = fisher(fcum11, fcum12, fcum21, fcum22) * dc;
    if (pr1 > 1.0L) pr1 = 1.0L;
    ldbl pr = (pr1 < 0.5L) ? 2.0L * pr1 : 1.0L;
    string ccode(ftab.ccode(j + 1));
    strtrm(ccode);
    double p1 = (double)fcum12 / (double)n1, p2 = (double)fcum22 / (double)n2;
    fout << "Pr of [" << ccode << "]-[" << ccodee << "]:G" << ftab.rcode(2) << "=" << p2;
    fout << " - G" << ftab.rcode(1) << "=" << p1;
    fout << "  1-tailed p=" << pr1 << pind(pr1) << "  2-tailed p=" << pr << pind(pr) << "\n";
    double d = 0.0;
    double ci = caldpci(d, opt.bt(), opt.cc(), ftab.nc(), 1, dc, fcum11, fcum12, fcum21, fcum22, true); // 連続補正あり
    fout << "  " << opt.cc() * 100.0 << "% CI of Diff of Pr=" << d << " +/- " << ci << "(" << d - ci << " - " << d + ci << ")\n";
  }
}

void htestclass::put_hmcon(ostream& fout, mctype mc, bool ma, double kc, const vecdbl& tvec, const ftable& ftab) const
// 順位平均値の多重比較を出力する
{
  int df = 1, dc = 1;
  string method("TUKEY");
  switch (mc)
  {
  case mcbon : method = "BONFERRONI";
    dc = ma ? ftab.nr() * (ftab.nr() - 1) / 2 : ftab.nr() - 1;
    break;
  case mcdun : method = "DUNNETT";
    break;
  case mcfis : method = "FISHER";
    break;
  case mcsch : method = "SCHEFFE";
    df = ftab.nr() - 1;
  }
  fout << "\n<" << method << " TYPE MULTIPLE COMPARISON OF GROUP>\n";
  int llen = 72;                // 区切り線の長さ
  fout << "GROUP    :Rnk Mean - GROUP    :Rnk Mean";
  if (mc == mctuk) { fout << "   q-value Groups";  llen += 5; }
  else if (mc == mcdun) { fout << "   d-value Groups";  llen += 5; }
  else fout << " Chi-squared";
  fout << "   DF     p-value\n";
  putline(fout, llen);
  size_t iend = ma ? ftab.nr() : 2;
  for (size_t i = 1; i < iend; ++i)
  {
    double ni = (double)(ftab.frtotal(i));
    double ri = tvec[i] / ni;
    string codei(ftab.rcode(i));
    int npadi = codei.size() - strclen(codei, CHARwide);
    for (size_t j = i + 1; j <= ftab.nr(); ++j)
    {
      double nj = (double)(ftab.frtotal(j));
      double rj = tvec[j] / nj;
      double drm = rj - ri;
      double csqr = 6.0 * drm * drm / (1.0 / ni + 1.0 / nj) / tvec[0] / kc;
      double r = (ni == nj) ? 0.5 : sqrt(ni / (ni + nj) * nj / (ni + nj));
      ldbl pr = calmcpr(mc, csqr, ftab.nr(), df, 0, dc, r);
      int npad = ftab.rcode(j).size() - strclen(ftab.rcode(j), CHARwide);
      fout << left << setw(9 + npad) << ftab.rcode(j) << ":";
      fout << right << setw(8) << rj;
      fout << left << " - " << setw(9 + npadi) << codei << ":";
      fout << right << setw(8) << ri;
      if (mc == mctuk || mc == mcdun) fout << setw(10) << csqr << setw(5) << ftab.nr() << setw(7) << "inf";
      else fout << setw(12) << csqr << setw(5) << df;
      fout << setw(12) << pr << pind(pr) << "\n";
    }
  }
  putline(fout, llen);
}

void htestclass::put_hmrdci(ostream& fout, const option& opt, mctype mc, bool ma, double kc, const vecdbl& tvec, const ftable& ftab) const
// 2群間の順位平均値の差の信頼区間出力
{
  int df = 1, dc = 1;
  string method("TUKEY");
  switch (mc)
  {
  case mcbon : method = "BONFERRONI";
    dc = ma ? ftab.nr() * (ftab.nr() - 1) / 2 : ftab.nr() - 1;
    break;
  case mcdun : method = "DUNNETT";
    break;
  case mcfis : method = "FISHER";
    break;
  case mcsch : method = "SCHEFFE";
    df = ftab.nr() - 1;
  }
  fout << "\n<" << method << " TYPE " << opt.cc() * 100.0 << "% SIMULTANEOUS CI OF DIFF OF RANK MEAN>\n";
  fout << "GROUP    - GROUP   Diff of Mrank   Precision   Lower-Lim   Upper-Lim\n";
  const int LENline = 68;       // 区切り線の長さ
  putline(fout, LENline);
  size_t iend = ma ? ftab.nr() : 2;
  double deps = DBL_EPSILON * (double)(ftab.ftotal());
  for (size_t i = 1; i < iend; ++i)
  {
    double ni = (double)(ftab.frtotal(i));
    double ri = tvec[i] / ni;
    string codei(ftab.rcode(i));
    int npadi = codei.size() - strclen(codei, CHARwide);
    for (size_t j = i + 1; j <= ftab.nr(); ++j)
    {
      double nj = (double)(ftab.frtotal(j));
      double rj = tvec[j] / nj;
      double drm = rj - ri;
      if (fabs(drm) < deps) drm = 0.0;
      double vd = kc * tvec[0] * (1.0 / ni + 1.0 / nj) / 6.0;
      double r = (ni == nj) ? 0.5 : sqrt(ni / (ni + nj) * nj / (ni + nj));
      double ci = calmcci(mc, opt.cc(), ftab.nr(), df, 0, dc, vd, r);
      int npad = (ftab.rcode(j)).size() - strclen(ftab.rcode(j), CHARwide);
      fout << left << setw(9 + npad) << ftab.rcode(j);
      fout << "- ";
      fout << setw(9 + npadi) << codei << right << setw(12) << drm;
      fout << setw(12) << ci << setw(12) << drm - ci << setw(12) << drm + ci;
      fout << "\n";
    }
  }
  putline(fout, LENline);
}

int htestclass::get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst)
// 度数集計表を読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nftab = fgetftable(fname, idlst, ftablst);
  if (nftab < 0) throw runtime_error("Run-time error !");
  if (opt.check(nftab) != 0) throw option_error("Run-time option error !");
  return nftab;                 // 読み込んだ度数集計表の数を返す
}

void htestclass::put(option& opt, listidd& idlst, listftab& ftablst) const
// 結果を出力する
{
  if (opt.res("res") == false) return;
  listidd::iterator iitr = idlst.begin();
  for (listftab::iterator fitr = ftablst.begin(); fitr != ftablst.end(); ++iitr, ++fitr)
  {
    put_id(cout, *iitr);
    fitr->trm();                // 行・列の合計が0のものを削除する
    if (fitr->nr() > 0)
    {
      if (opt.res("tab")) putftable(cout, *fitr, opt.tab());
      if (fitr->nr() > 1 && fitr->nc() > 1)
      {
        double csqr = 0.0, kc = 0.0;
        vecdbl tvec(fitr->nr() + 1);
        int df = htest(csqr, kc, tvec, *fitr);
        if (opt.res("r") || (opt.res("def") && fitr->nr() > 2)) put_rr(cout, opt, df, csqr, *fitr); // 順位相関比出力
        if (opt.res("ht") || (opt.res("def") && fitr->nr() > 2))
        {
          ldbl pr = ctop(csqr, df);
          cout << "\nKruskal-Wallis's H-test\n";
          cout << "  H=chi-squared=" << csqr << "  DF=" << df << "  p=" << pr << pind(pr) << "\n";
        }
        if (fitr->nr() == 2)
        {
          if (opt.res("ut") || opt.res("def")) put_utest(cout, opt, tvec, *fitr);
          if (opt.res("bt")) put_bun(cout, opt, *fitr);
        }
        else
        {
          if (opt.res("ma"))
          {
            put_hmcon(cout, opt.ma(), true, kc, tvec, *fitr);
            put_hmrdci(cout, opt, opt.ma(), true, kc, tvec, *fitr);
          }
          if (opt.res("md"))
          {
            put_hmcon(cout, opt.md(), false, kc, tvec, *fitr);
            put_hmrdci(cout, opt, opt.md(), false, kc, tvec, *fitr);
          }
        }
      }
      cout << "\n" << endl;
    }
  }
}

int main(int argc, char* argv[])
// メインルーチン
{
  htestclass htest;             // メインクラス変数
  option opt(argv[0]);          // オプション
  listidd idlst;                // データ識別情報リスト
  listftab ftablst;             // 度数表リスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok)                // 標準入力から読み込む
        {
          htest.get(nullptr, opt, idlst, ftablst);
          htest.put(opt, idlst, ftablst);
        }
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      {
        htest.get(argv[i], opt, idlst, ftablst);  fok = false;
        htest.put(opt, idlst, ftablst);
      }
    }
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
