// mcrosse.C Version 7.41e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-27 11:28:30 tac>
//
// 層別項目の分類別に縦と横のクロス集計を行い、層別クロス集計表を出力する。
// クロス集計表は、計数値解析コマンド用の標準入力形式に従っている。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法  :g++ -O -s -W mcrosse mcrosse.C
//               -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "mcrosse  Version 7.41e       Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:mcrosse  [<OPTION> ...]  [<DATA FILE> ...]",
  "  The stratified cross-tabulation of items.",
  "<OPTION>",
  "  -c=<NUMBER>:specify the item number of the column item(default:-c=3)",
  "  -f=<FILE>:read options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -m=<NUMBER>:specify the number of the stratifed item(default:-m=1)",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=mcrosse_ngp.dat",
  "  -r=<NUMBER>:specify the item number of the row item(default:-r=2)",
  "  -s[={-|+}]:select the data aggregation method(-s:-s=+)",
  "  -tab[={-|+|data|fr|pc|pr|pt}]:select the type of output table",
  "    data:standard input format for numerical analysis commands(default)",
  "    fr:only frequency  pc:also output % for column total",
  "    pr:also output % for row total  pt:also output % for all total",
  "  -tt[={-|+}]:specify whether to output the total table(-tt:-tt=+)",
  "  -wc=<NUMBER>:specify the number of outout digits for column data",
  "    default:-wc=6",
  "  -wr=<NUMBER>:specify the number of outout digits for row data",
  "    default:-wr=6(minimum number)"
};

const string Title = "=== Stratified Cross-tabulation ==="; // 表題
const string NGPNAMEdef = "mcrosse_ngp.dat"; // 省略時Ngraphファイル名
constexpr int Wdef = 6;         // 省略時出力桁数
constexpr int Wmax = LENcrt / 2; // 最大出力桁数

using paircode = pair<string, double>; // コードペアの別名定義

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  int vm;                       // 層別項目の項目番号
  int vr;                       // 行の項目番号
  int vc;                       // 列の項目番号
  string vngpname;              // Ngraphファイル名
  bool vstr;                    // 文字列集計フラッグ
  tabtype vtab;                 // 度数表の種類
  bool vtt;                     // 合計集計結果の出力フラッグ
  int vwr;                      // 行データの出力桁数
  int vwc;                      // 列データの出力桁数
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  int m() const                 // 層別項目の項目番号を返す
  { return vm; }
  int r() const                 // 行の項目番号を返す
  { return vr; }
  int c() const                 // 列の項目番号を返す
  { return vc; }
  const string& ngpname()       // Ngraphデータファイル名を返す
  { return vngpname; }
  bool str() const              // 文字列集計フラッグを返す
  { return vstr; }
  tabtype tab() const           // 度数表の種類を返す
  { return vtab; }
  bool tt() const          // 合計集計結果の出力フラッグフラッグを返す
  { return vtt; }
  int wr() const                // 行データの出力桁数を返す
  { return vwr; }
  int wc() const                // 列データの出力桁数を返す
  { return vwc; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(size_t ni);         // オブションのチェックとセット
};

option::option(const char* c_pname) : vpath(), vpname(), vm(1), vr(2), vc(3), vngpname(), vstr(false), vtab(tabdata), vtt(false), vwr(Wdef), vwc(Wdef)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;

  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0, dum = 0;
  if (opt[0] == "c")            // 列項目の指定
  {
    vc = 1;  dum = INT_MAX;
    if (getrange(vc, dum, opt[1]) < 1) err = badopt(vpname, optstr, USAGE);
    else if (vc < 1 || vc != dum) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "m")       // 層別項目の指定
  {
    vm = 1;  dum = INT_MAX;
    if (getrange(vm, dum, opt[1]) < 1) err = badopt(vpname, optstr, USAGE);
    else if (vm < 1 || vm != dum) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
  }
  else if (opt[0] == "r")       // 行項目の指定
  {
    vr = 1;  dum = INT_MAX;
    if (getrange(vr, dum, opt[1]) < 1) err = badopt(vpname, optstr, USAGE);
    else if (vr < 1 || vr != dum) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "s")       // 文字列として集計指定
  {
    if (opt[1].empty() || opt[1] == "+") vstr = true;
    else if (opt[1] == "-") vstr = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "tab")     // 度数表の種類指定
  {
    if (opt[1].empty() || opt[1] == "+" || opt[1] == "data") vtab = tabdata;
    else if (opt[1] == "-") vtab = tabnon;
    else if (opt[1] == "fr") vtab = tabfreq;
    else if (opt[1] == "p" || opt[1] == "pr") vtab = tabfprr;
    else if (opt[1] == "pc") vtab = tabfprc;
    else if (opt[1] == "pt") vtab = tabfprt;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "tt")      // 合計集計結果の出力指定
  {
    if (opt[1].empty() || opt[1] == "+") vtt = true;
    else if (opt[1] == "-") vtt = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "wc")      // 列データの出力桁数指定
  {
    char* ptr = nullptr;
    vwc = strtol(opt[1].c_str(), &ptr, 0);
    if (vwc < 2 || Wmax < vwc) err = badopt(vpname, optstr, USAGE, 2, Wmax);
  }
  else if (opt[0] == "wr")      // 行データの出力桁数指定
  {
    char* ptr = nullptr;
    vwr = int(strtol(opt[1].c_str(), &ptr, 10));
    if (vwr < 6 || Wmax < vwr) err = badopt(vpname, optstr, USAGE, 6, Wmax);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error !") + optstr);
}

int option::check(size_t ni)    // オプションのチェックとセット
{
  if ((int)ni < vm) return disperr(vpname, "Wrong number of -m option !");
  if ((int)ni < vr) return disperr(vpname, "Wrong number of -r option !");
  if ((int)ni < vc) return disperr(vpname, "Wrong number of -c option !");
  return 0;
}

class varclass                  // 変数クラス
{
private:
  vector<ftable> ftabvec;       // 度数表ベクトル
  ftable dumftab;               // ダミー度数表
  set<paircode > vmset;         // 層別コードセット
  set<paircode > vrset;         // 行コードセット
  set<paircode > vcset;         // 列コードセット
  void set_code(const option& opt, ftable& ftab) const;
  // 度数表のコードをセットする
  void set_freq(const option& opt, ldata& datalst);
  // 度数表の度数をクロス集計する
public:
  varclass() : ftabvec(), dumftab(), vmset(), vrset(), vcset() // コンストラクタ
  {}
  ~varclass()                   // デストラクタ
  {}
  const ftable& ftab(size_t idx) const // 度数表を返す
  { return (idx < ftabvec.size()) ? ftabvec[idx] : dumftab; }
  const set<paircode>& mset() const // 層別コードセットを返す
  { return vmset; }
  const set<paircode>& rset() const // 行コードセットを返す
  { return vrset; }
  const set<paircode>& cset() const // 列コードセットを返す
  { return vcset; }
  int setmcross(const option& opt, ldata& datalst);
  // データを集計して層別クロス集計表を作成する
};

void varclass::set_code(const option& opt, ftable& ftab) const
// 度数表のコードをセットする
{
  ostringstream ostr;
  string tstr("CODE");
  int npad = tstr.size() - strclen(tstr, CHARwide);
  ostr << left << setw(opt.wr() + npad) << tstr << right;
  ftab.hed(ostr.str());
  int i = 1;
  for (set<paircode>::const_iterator sitr = vcset.begin(); sitr != vcset.end(); ++sitr, ++i)
  {
    ostr.str("");
    if (opt.str()) ostr << setw(opt.wc()) << sitr->first;
    else ostr << setw(opt.wc()) << sitr->second;
    ftab.ccode(i, ostr.str());
  }
  i = 1;
  for (set<paircode>::const_iterator sitr = vrset.begin(); sitr != vrset.end(); ++sitr, ++i)
  {
    ostr.str("");
    if (opt.str()) ostr << sitr->first;
    else ostr << sitr->second;
    ftab.rcode(i, ostr.str());
  }
}

void varclass::set_freq(const option& opt, ldata& datalst)
// 度数表の度数をクロス集計する
{
  for (auto& cdata : datalst)
  {
    int midx = (opt.str() ? search_data(vmset, cdata.sdata(opt.m() - 1)) : search_data(vmset, cdata.fdata(opt.m() - 1))) - 1;
    if (midx >= 0)
    {
      int rno = opt.str() ? search_data(vrset, cdata.sdata(opt.r() - 1)) : search_data(vrset, cdata.fdata(opt.r() - 1));
      if (rno > 0)
      {
        int cno = opt.str() ? search_data(vcset, cdata.sdata(opt.c() - 1)) : search_data(vcset, cdata.fdata(opt.c() - 1));
        if (cno > 0)
        { ftabvec[midx].f(rno, cno, ftabvec[midx].f(rno, cno) + 1); }
      }
    }
  }
  int tidx = vmset.size();
  for (size_t k = 0; k < vmset.size(); ++k)
  {
    for (size_t i = 1; i <= ftabvec[tidx].nr(); ++i)
    {
      for (size_t j = 1; j <= ftabvec[tidx].nc(); ++j)
      { ftabvec[tidx].f(i, j, ftabvec[tidx].f(i, j) + ftabvec[k].f(i, j)); }
    }
    ftabvec[k].set();
  }
  ftabvec[tidx].set();
}

int varclass::setmcross(const option& opt, ldata& datalst)
// データを集計して層別クロス集計表を作成する
{
  int midx = opt.m() - 1, ridx = opt.r() - 1, cidx = opt.c() - 1;
  for (auto& cdata : datalst)
  {
    if (opt.str())              // 文字として集計
    {
      if (cdata.sdata(midx) != BLANKs && cdata.sdata(ridx) != BLANKs && cdata.sdata(cidx) != BLANKs)
      {
        paircode tdata(cdata.sdata(midx), cdata.fdata(midx));
        vmset.insert(tdata);    // 新規データだけ挿入する
        tdata.first = cdata.sdata(ridx);
        tdata.second = cdata.fdata(ridx);
        vrset.insert(tdata);
        tdata.first = cdata.sdata(cidx);
        tdata.second = cdata.fdata(cidx);
        vcset.insert(tdata);
      }
    }
    else
    {
      double mfdata = cdata.fdata(midx);
      double rfdata = cdata.fdata(ridx), cfdata = cdata.fdata(cidx);
      if (mfdata != BLANKf && rfdata != BLANKf && cfdata != BLANKf)
      {
        paircode tdata(cdata.sdata(midx), mfdata);
        vmset.insert(tdata);    // 新規データだけ挿入する
        tdata.first = cdata.sdata(ridx);
        tdata.second = rfdata;
        vrset.insert(tdata);
        tdata.first = cdata.sdata(cidx);
        tdata.second = cfdata;
        vcset.insert(tdata);
      }
    }
  }
  ftabvec.resize(vmset.size() + 1);
  for (size_t k = 0; k <= vmset.size(); ++k)
  {
    ftabvec[k].resize(vrset.size(), vcset.size());
    set_code(opt, ftabvec[k]);
  }
  set_freq(opt, datalst);
  dumftab.resize(vrset.size(), vcset.size()); // ダミー度数表
  set_code(opt, dumftab);
  return (int)(vmset.size());   // 層別コード数を返す
}

class mcrossclass               // mcrossメインクラス
{
private:
  string fname;                 // ファイル名
  void put_id(ostream& fout, const option& opt, idd& id, const paircode& mcode, const varclass& var) const;
  // 度数表のタイトルとデータ識別情報を出力する
  void put_dtab(ostream& fout, const ftable& ftab) const;
  // 度数表を度数データ形式で出力する
  void plot_ngp(option& opt, idd& id, const varclass& var) const;
  // Ngraph用データを出力する
  void plot_ngpid(ofstream& fout, const option& opt, const idd& id, const paircode& mcode, const varclass& var) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
public:
  mcrossclass() : fname()       // コンストラクタ
  {}
  ~mcrossclass()                // デストラクタ
  {}
  int get(const char* c_fname, option& opt, idd& id, ldata& datalst);
  // データを読み込み累積する
  void put(option& opt, idd& id, ldata& datalst) const;
  // 集計結果を出力する
};

void mcrossclass::put_id(ostream& fout, const option& opt, idd& id, const paircode& mcode, const varclass& var) const
// 度数表のタイトルとデータ識別情報を出力する
{
  idd id1(id);                  // 出力用データ識別情報
  stra tstra;
  if (opt.str() || mcode.second == BLANKf)
  { tstra.get('s', opt.m(), mcode.first); }
  else
  {
    ostringstream ostr;
    ostr << mcode.second;
    tstra.get('s', opt.m(), ostr.str());
  }
  lstra tstrals;
  tstrals.push_back(tstra);
  id1.append_stra(tstrals);
  id1.assign_item(opt.r() - 1, id1.item(opt.r() - 1), atype, avalrow);
  id1.assign_item(opt.c() - 1, id1.item(opt.c() - 1), atype, avalcol);
  int ns = (opt.tt()) ? (var.mset()).size() + 1 : (var.mset()).size();
  switch (opt.tab())
  {
  case tabdata : id1.put_title(fout, formtag);
    id1.put_stra(fout, formtag);
    id1.put_item(fout, opt.r() - 1, formtag);
    id1.put_item(fout, opt.c() - 1, formtag);
    break;
  default : puttitle(fout, Title);
    fout << "\n";
    id1.put_title(fout, formnon);
    id1.put_stra(fout, formnon);
    fout << "ROW ITEM   :" << id1.item(opt.r() - 1) << "\n";
    fout << "COLUMN ITEM:" << id1.item(opt.c() - 1) << "\n";
  }
}

void mcrossclass::put_dtab(ostream& fout, const ftable& ftab) const
// 度数表を度数データ形式で出力する
{
  datac tdata(ftab.nc() + 1);
  tdata.assign(0, ftab.hed());
  for (size_t j = 1; j <= ftab.nc(); ++j) tdata.assign(j, ftab.ccode(j));
  tdata.put(fout);
  for (size_t i = 1; i <= ftab.nr(); ++i)
  {
    tdata.assign(0, ftab.rcode(i));
    for (size_t j = 1; j <= ftab.nc(); ++j) tdata.assign(j, double(ftab.f(i, j)));
    tdata.put(fout);
  }
}

void mcrossclass::plot_ngp(option& opt, idd& id, const varclass& var) const
// Ngraph用データを出力する
{
  ofstream fout;
  if (ofopen(fout, (opt.ngpname()).c_str()) == false) return; // ファイルオープン失敗
  set<paircode>::const_iterator mitr = (var.mset()).begin();
  plot_ngpid(fout, opt, id, *mitr, var);
  int tidx = (var.mset()).size();
  int ns = (opt.tt()) ? tidx + 1 : tidx;
  double x = 0.0, y = (double)ns * ((double)((var.ftab(0)).nr()) + 0.5) + 0.5;
  for (size_t k = 0; k < (var.mset()).size(); ++k, ++mitr)
  {
    if (k > 0)
    {
      fout << "%STRATIFIED CODE:";
      if (opt.str() || mitr->second == BLANKf) fout << mitr->first;
      else fout << mitr->second;
      fout << "\n";
    }
    int n = 0;
    for (size_t i = 1; i <= (var.ftab(k)).nr(); ++i)
    {
      if ((var.ftab(k)).frtotal(i) > 0)
      {
        fout << "%GROUP CODE:" << (var.ftab(k)).rcode(i) << "\n";
        x = 0.0;  y -= 1.0;  n = 0;
        for (size_t j = 1; j <= (var.ftab(k)).nc(); ++j)
        {
          if ((var.ftab(k)).f(i, j) > 0)
          {
            fout << x << " " << y - 0.3 << " ";
            n += (var.ftab(k)).f(i, j);
            x = (double)n / (double)((var.ftab(k)).frtotal(i)) * 100.0;
            fout << x << " " << y + 0.3 << endl;
          }
        }
      }
    }
    y -= 0.5;
  }
  if (opt.tt())                 // 合計集計結果を出力
  {
    fout << "%ALL\n";
    int n = 0;
    for (size_t i = 1; i <= (var.ftab(tidx)).nr(); ++i)
    {
      if ((var.ftab(tidx)).frtotal(i) > 0)
      {
        fout << "%GROUP CODE:" << (var.ftab(tidx)).rcode(i) << "\n";
        x = 0.0;  y -= 1.0;  n = 0;
        for (size_t j = 1; j <= (var.ftab(tidx)).nc(); ++j)
        {
          if ((var.ftab(tidx)).f(i, j) > 0)
          {
            fout << x << " " << y - 0.3 << " ";
            n += (var.ftab(tidx)).f(i, j);
            x = (double)n / (double)((var.ftab(tidx)).frtotal(i)) * 100.0;
            fout << x << " " << y + 0.3 << endl;
          }
        }
      }
    }
  }
}

void mcrossclass::plot_ngpid(ofstream& fout, const option& opt, const idd& id, const paircode& mcode, const varclass& var) const
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  idd id1(id);                  // 出力用データ識別情報
  stra tstra;
  if (opt.str() || mcode.second == BLANKf)
  { tstra.get('s', opt.m(), mcode.first); }
  else
  {
    ostringstream ostr;
    ostr << mcode.second;
    tstra.get('s', opt.m(), ostr.str());
  }
  lstra tstrals;
  tstrals.push_back(tstra);
  id1.append_stra(tstrals);
  id1.assign_item(opt.r() - 1, id1.item(opt.r() - 1), atype, avalrow);
  id1.assign_item(opt.c() - 1, id1.item(opt.c() - 1), atype, avalcol);
  int ns = (opt.tt()) ? (var.mset()).size() + 1 : (var.mset()).size();
  fout << "-s1 -drectangle_fill -x1 -y3 -zx0,100,20 -zy0,";
  fout << double(ns) * (double((var.ftab(0)).nr()) + 0.5) + 0.5 << ",1\n";
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id1.put_title(fout, formngp);
  id1.put_stra(fout, formngp);
  id1.put_item(fout, opt.r() - 1, formngp);
  id1.put_item(fout, opt.c() - 1, formngp);
}

int mcrossclass::get(const char* c_fname, option& opt, idd& id, ldata& datalst)
// データを読み込み累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  if (opt.check(id.ni()) != 0) throw option_error("Run-time opttion error !");
  return nc;                    // 読み込んだケース数を返す
}

void mcrossclass::put(option& opt, idd& id, ldata& datalst) const
// 集計結果を出力する
{
  varclass var;                     // 変数オブジェクト
  int nm = var.setmcross(opt, datalst); // 層別クロス集計表を作成する
  if (id.nc() < 1 || nm < 1)            // データなし
  {
    if (opt.tab() != tabnon && opt.tab() != tabdata)
    {
      paircode tdata("", BLANKf);
      put_id(cout, opt, id, tdata, var);
      cout << "\n" << endl;
    }
    return;
  }
  paircode tcode("ALL", BLANKf);
  set<paircode>::const_iterator mitr = (var.mset()).begin();
  if (opt.tab() != tabnon)
  {
    switch (opt.tab())
    {
    case tabdata :              // 度数データ形式で出力
      for (int k = 0; k < nm; ++k, ++mitr)
      {
        if (k > 0) cout << "\n";
        put_id(cout, opt, id, *mitr, var);
        put_dtab(cout, var.ftab(k));
      }
      if (opt.tt())               // 合計集計結果を出力
      {
        cout << "\n";
        put_id(cout, opt, id, tcode, var);
        put_dtab(cout, var.ftab(nm));
      }
      break;
    default :                     // 度数表形式で出力
      for (int k = 0; k < nm; ++k, ++mitr)
      {
        put_id(cout, opt, id, *mitr, var);
        putftable(cout, var.ftab(k), opt.tab());
        cout << "\n" << endl;
      }
      if (opt.tt())               // 合計集計結果を出力
      {
        put_id(cout, opt, id, tcode, var);
        putftable(cout, var.ftab(nm), opt.tab());
        cout << "\n" << endl;
      }
    }
  }
  if (opt.ngpname() != "") plot_ngp(opt, id, var); // Ngraph用データ出力
}

int main(int argc, char* argv[]) // メインルーチン
{
  mcrossclass mcross;           // メインクラス変数
  option opt(argv[0]);          // オプション
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) mcross.get(nullptr, opt, id, datalst); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { mcross.get(argv[i], opt, id, datalst);  fok = false; }
    }
    mcross.put(opt, id, datalst);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
