// mteste.C Version 7.42e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-07-04 11:43:37 tac>
//
// 層別クロス集計表を入力し、Mantel-Haenszelの検定(拡張法、一般化拡張法含む)と
// Breslow-Dayの検定を行う。
// 層別クロス集計表は計数値解析コマンド用の標準入力形式に従っている必要がある。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W mteste mteste.C
//              -I/usrlocal/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(mate.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "mteste  Version 7.42e        Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:mteste  [<OPTIONN> ...]  [<TAB FILE> ...]",
  "  The Mantel-Haenszel test with the Stratified cross-tabulation table.",
  "<OPTION>",
  "  -f=<FILE>:read options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ngp={mhor|mhrd|mhrr|or|rr}[,...]:select the data for Ngraph",
  "    mhor:adjusted odds ratio  mhrd:adjusted risk difference",
  "    mhrr:adjusted risk ratio  or:odds ratio  rr:risk ratio  default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output  -ngpname:-ngpname=mteste_ngp<2-digit number>.dat",
  "  -res=[-][{def|all|bt|cc[<C-COEF>]|emt|gemt|lgt|m|mhor|mhrr|mt[{c|o}]|r",
  "      |rrt|sts|stt|tab[{fr|pc|pr|pt}]}][,...]:select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    bt:Breslow-Day test",
  "    cc[<C-COEF>]:confidence coefficient(0=<C-COEF<1 cc:cc0.95)",
  "    emt:Mantel-extension test  gemt:generalized Mantel-extension test",
  "    lgt:logit test  m:Bonferroni type multiple comparison",
  "    mhor:odds ratio of MH-test  mhrr:risk ratio of MH-test",
  "    mt[{c|o}]:Mantel-Haenszel test",
  "      c:with Yates's correction  o:no Yates's correction",
  "    r:rank correlation ratio",
  "    rrt:risk ratio test  sts:stratified results  stt:overall results",
  "    tab[{fr|pc|pr|pt}]:cross tabulation table",
  "      fr:only frequency  pc:also output % for column total",
  "      pr:also output % for row total  pt:also output % for all total",
  "    default:auto judgment according to data",
  "  -skip[={-|+}]:specify whether to skip the inappropriate tables",
  "    -skip:-skip=+  default:-skip=-"
};

const string Title = "=== Stratified Group Compariosn of Ordinal Data ==="; // 表題
const string NGPNAMEdef = "mteste_ngp"; // 省略時Ngraphファイル基本名

class rci                       // リスク比と信頼区間クラス
{
private:
  int vn;                       // 例数
  double vr;                    // リスク比等
  double vrl;                   // リスク比下限
  double vru;                   // リスク比上限
public:
  rci() : vn(0), vr(0.0), vrl(0.0), vru(0.0) // デフォールトコンストラクタ
  {}
  rci(int n0, double r0, double rl0, double ru0) : vn(n0), vr(r0), vrl(rl0), vru(ru0)
  // コンストラクタ
  {}
  rci(const rci& rci0) : vn(rci0.vn), vr(rci0.vr), vrl(rci0.vrl), vru(rci0.vru)
  // コピーコンストラクタ
  {}
  virtual ~rci()                // デストラクタ
  {}
  const rci& operator=(const rci& rci0); // コピー代入演算子
  int n() const                          // 例数を返す
  { return vn; }
  int n(int n0)                 // 例数を設定する
  { return (vn = n0); }
  double r() const              // リスク比を返す
  { return vr; }
  double r(double r0)           // リスク比を設定する
  { return (vr = r0); }
  double rl() const             // リスク比下限を返す
  { return vrl; }
  double rl(double rl0)         // リスク比下限を設定する
  { return (vrl = rl0); }
  double ru() const             // リスク比上限を返す
  { return vru; }
  double ru(double ru0)         // リスク比上限を設定する
  { return (vru = ru0); }
};

const rci& rci::operator=(const rci& rci0) // コピー代入演算子
{
  if (&rci0 == this) return *this; // 自分自身への代入を防ぐ
  vn = rci0.vn;  vr = rci0.vr;  vrl = rci0.vrl;  vru = rci0.vru;
  return *this;                 // 自分自身への参照を返す
}

using vecrci = vector<rci>;    // リスク比と信頼区間ベクトルの短名定義

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  testtype vmt;                 // Mantel-Haenszel検定の種類
  string vngpname;              // Ngraphファイル基本名
  string vngpnamen;             // Ngraphファイル名(番号付き)
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  lstr vreslst;                 // 分析結果出力リスト
  map<string, bool> vresmap;    // 分析結果出力の有無マップ
  bool vskip;                   // 不適集計表スキップフラッグ
  tabtype vtab;                 // 度数集計費用の種類
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  testtype mt() const           // Mantel-Haenszel検定の種類を返す
  { return vmt; }
  const string& ngpname(int fno); // Ngraphデータファイル名を返す
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 解析結果出力フラッグを返す
  bool skip() const             // 不適集計表スキップフラッグを返す
  { return vskip; }
  tabtype tab() const           // 度数集計表の種類を返す
  { return vtab; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check();                  // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vngpname(), vngpnamen(), vngplst(), vngpmap(), vmt(testcor), vreslst(), vresmap(), vskip(false), vtab(tabfprr)
// コンストラクタ
{
  if (c_pname == 0) return;     // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

const string& option::ngpname(int fno) // Ngraphデータファイル名を返す
{
  if (vngpname == "") return vngpname; // -ngpname指定なし
  ostringstream ostr;
  if (fno < 10) ostr << vngpname << "0" << fno << ".dat";
  else ostr << vngpname << fno << ".dat";
  vngpnamen = (string)(ostr.str());
  return vngpnamen;             // 番号付きファイル名を返す
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngpitr = vngpmap.find(code);
  if (ngpitr == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "f")            // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル基本名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
  }
  else if (opt[0] == "res")     // 解析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "skip")   // 不適集計表スキップ指定
  {
    if (opt[1].empty() || opt[1] == "+") vskip = true;
    else if (opt[1] == "-") vskip = false;
    else err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error !") + optstr);
}

int option::check()             // オプションのチェックとセット
{
  if (set_ngp() < 0) return -2; // -ngpオプションのセット
  if (set_res() < 0) return -3; // -resオプションのセット
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["mhor"] = vngpmap["mhrd"] = vngpmap["mhrr"] = vngpmap["or"] = vngpmap["rr"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["mhor"] = vngpmap["mhrd"] = vngpmap["mhrr"] = vngpmap["or"] = vngpmap["rr"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngpitr = vngpmap.find(ngp);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["mt"] = vresmap["sts"] = vresmap["stt"] = vresmap["tab"] = true;
  vresmap["bt"] = vresmap["emt"] = vresmap["gemt"] = vresmap["lgt"] = vresmap["m"] = vresmap["mhor"] = vresmap["mhrr"] = vresmap["r"] = vresmap["rrt"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["mt"] = vresmap["sts"] = vresmap["stt"] = vresmap["tab"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["mt"] = vresmap["sts"] = vresmap["stt"] = vresmap["tab"] = true;
      }
      else if (res == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["mt"] = vresmap["sts"] = vresmap["stt"] = vresmap["tab"] = vresmap["bt"] = vresmap["emt"] = vresmap["gemt"] = vresmap["lgt"] = vresmap["m"] = vresmap["mhor"] = vresmap["mhrr"] = vresmap["r"] = vresmap["rrt"] = true;
        vmt = testcor;  vtab = tabfprr;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (res == "mtc")    // 連続修正有
      { vresmap["mt"] = true;  vmt = testcor; }
      else if (res == "mto")    // 連続修正無
      { vresmap["mt"] = true;  vmt = testorg; }
      else if (res == "tabfr")  // 度数だけ出力
      { vresmap["tab"] = true;  vtab = tabfreq; }
      else if (res == "tabpc")  // 列合計に対する％も出力
      { vresmap["tab"] = true;  vtab = tabfprc; }
      else if (res == "tabpr")  // 行合計に対する％も出力
      { vresmap["tab"] = true;  vtab = tabfprr; }
      else if (res == "tabpt")  // 全体合計に対する％も出力
      { vresmap["tab"] = true;  vtab = tabfprt; }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class mtestclass                // mtestメインクラス
{
private:
  string fname;                 // ファイル名
  int check_ftable(const listftab& ftablst) const;
  // 有効な度数表をチェックして、その数を返す
  int cal_mhstat(const option& opt, listidd& idlst, const listftab& ftablst, vecrci& mhorvec, vecrci& mhrdvec, vecrci& mhrrvec, vecrci& orvec, vecrci& rrvec) const;
  // Mantel-Haenzelの各種統計量を計算して返す
  void cal_mtest(const option& opt, vecrci& mhorvec, vecrci& mhrdvec, vecrci& mhrrvec, int fidx, const listftab& ftablst0, int nt) const;
  // Mantel-Haenszelの検定を行い各種統計量を返す
  void cal_rrtest(const option& opt, vecrci& rrvec, int fidx, const listftab& ftablst0, int nt) const;
  // リスク比検定を行い各種統計量を返す
  void cal_lgtest(const option& opt, vecrci& orvec, int fidx, const listftab& ftablst0, int nt) const;
  // ロジット検定(オッズ比検定)を行い各種統計量を返す
  void put_ftable(ostream& fout, option& opt, listidd& idlst, const listftab& ftablst, int nftab) const;
  // 度数表と解析結果を出力する
  void put_resid(ostream& fout, int idno, idd& id) const;
  // 解析結果のタイトルとデータ識別情報を出力する
  void put_ftabi(ostream& fout, option& opt, int idno, const ftable& ftabi, int nftab, bool ftabt) const;
  // 個々の度数集表と検定結果を出力する
  void put_overall(ostream& fout, option& opt, const listftab& ftablst, int nftab, int nt) const;
  // 全体の解析結果を出力する
  void put_mtest(ostream& fout, option& opt, int fidx, const listftab& ftablst0, int nftab, int nt, bool ftabt) const;
  // Mantel-Haenzelの検定結果を出力する
  void put_mr(ostream& fout, const listftab& ftablst0, int nt) const;
  // Mantel-Haenszelの順位相関比を出力する
  void put_lgtest(ostream& fout, option& opt, int fidx, const listftab& ftablst0, int nftab, int nt, bool ftabt) const;
  // ロジット検定(オッズ比検定)の結果を出力する
  void put_rrtest(ostream& fout, option& opt, int fidx, const listftab& ftablst0, int nftab, int nt, bool ftabt) const;
  // リスク比検定の結果を出力する
  void put_emtest(ostream& fout, option& opt, const listftab& ftablst0, int nftab, int nt, bool ftabt) const;
  // 拡張Mantel検定の結果を出力する
  void put_gemtest(ostream& fout, option& opt, const listftab& ftablst0, int nftab, int nt, bool ftabt) const;
  // 一般化拡張Mantel検定の結果を出力する
  void plot_ngp(ofstream& fout, option& opt, listidd& idlst, int nftab, int nt, vecrci& mhorvec, vecrci& mhrdvec, vecrci& mhrrvec, vecrci& orvec, vecrci& rrvec) const;
  // Ngraph用データを出力する
  void plot_ngpid(ofstream& fout, listidd& idlst) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_rcivec(ofstream& fout, const string& rname, int nftab, int nt, vecrci& rcivec) const;
  // 個々のリスク比と信頼区間をNgraph用に出力する
public:
  mtestclass() : fname()        // コンストラクタ
  {}
  ~mtestclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst);
  // 度数集計表を読み込む
  void put(option& opt, listidd& idlst, listftab& ftablst, int fno) const;
  // 集計結果を出力する
};

int mtestclass::check_ftable(const listftab& ftablst) const
// 有効な度数表をチェックして、その数を返す
{
  if (ftablst.empty()) return 0; // 度数表なし
  listftab::const_iterator fitr = ftablst.begin();
  size_t nr = fitr->nr(), nc = fitr->nc();
  int nftabok = 0;
  while (fitr != ftablst.end())
  {
    ftable ftabtrm(*fitr);
    ftabtrm.trm();
    if (fitr->nr() == nr && fitr->nc() == nc && ftabtrm.nr() > 1 && ftabtrm.nc() > 1) ++nftabok;
    ++fitr;
  }
  return nftabok;               // 有効な度数表数を返す
}

int mtestclass::cal_mhstat(const option& opt, listidd& idlst, const listftab& ftablst, vecrci& mhorvec, vecrci& mhrdvec, vecrci& mhrrvec, vecrci& orvec, vecrci& rrvec) const
// Mantel-Haenzelの各種統計量を計算して返す
{
  listidd::iterator iitr = idlst.begin();
  listftab::const_iterator fitr = ftablst.begin();
  size_t nr = fitr->nr(), nc = fitr->nc();
  ftable ftabt(*fitr);          // 合計度数表用
  ftabt.fclear();
  int nt = 0, fidx = 0;
  while (fitr != ftablst.end())
  {
    ftable ftabtrm(*fitr);
    ftabtrm.trm();
    listftab ftablsti;
    ftablsti.push_back(ftabtrm);
    if (fitr->nr() == nr && fitr->nc() == nc && ftabtrm.nr() > 1 && ftabtrm.nc() > 1)
    {
      cal_mtest(opt, mhorvec, mhrdvec, mhrrvec, fidx, ftablsti, ftabtrm.ftotal());
      cal_rrtest(opt, rrvec, fidx, ftablsti, ftabtrm.ftotal());
      cal_lgtest(opt, orvec, fidx, ftablsti, ftabtrm.ftotal());
      nt += fitr->ftotal();
      for (size_t i = 1; i <= nr; ++i)
      {
        for (size_t j = 1; j <= nc; ++j) ftabt.f(i, j, ftabt.f(i, j) + fitr->f(i, j));
      }
      ++fidx;
    }
    ++fitr;  ++iitr;
  }
  ftabt.set();
  listftab ftablstt;
  ftablstt.push_back(ftabt);
  cal_mtest(opt, mhorvec, mhrdvec, mhrrvec, fidx, ftablstt, nt);
  cal_rrtest(opt, rrvec, fidx, ftablstt, nt);
  cal_lgtest(opt, orvec, fidx, ftablstt, nt);
  cal_mtest(opt, mhorvec, mhrdvec, mhrrvec, fidx + 1, ftablst, nt);
  cal_rrtest(opt, rrvec, fidx + 1, ftablst, nt);
  cal_lgtest(opt, orvec, fidx + 1, ftablst, nt);
  return nt;                    // 全体の例数を返す
}

void mtestclass::cal_mtest(const option& opt, vecrci& mhorvec, vecrci& mhrdvec, vecrci& mhrrvec, int fidx, const listftab& ftablst0, int nt) const
// Mantel-Haenszelの検定を行い各種統計量を返す
{
  double rdm = 0.0, vrdm = 0.0, rrm = 0.0, rra = 0.0, rrb = 0.0, rrc = 0.0;
  double orm = 0.0, ora = 0.0, orb = 0.0, orc = 0.0, cmm = 0.0, cmr = 0.0;
  int dfm = 1;
  int dfr = mtest(rdm, vrdm, rrm, rra, rrb, rrc, orm, ora, orb, orc, cmm, cmr, ftablst0, opt.mt());
  double ln_rrm = (rrm == 0.0) ? 0.0 : log(rrm);
  double ln_orm = (orm == 0.0) ? 0.0 : log(orm);
  double z = ptoz(1.0 - opt.cc());
  double ci = z * sqrt(vrdm);
  double cil = rdm - ci, ciu = rdm + ci;
  double eps = DBL_EPSILON * fmax(5.0, vrdm);
  mhrdvec[fidx].n(nt);  mhrdvec[fidx].r(rdm);
  mhrdvec[fidx].rl(cil);  mhrdvec[fidx].ru(ciu);
  if (0.0 < rrm)
  {
    double rrt = 2.0 * rra * rrb + z * z * rrc;
    double vrrt = (4.0 * rra * rrb + z * z * rrc) * rrc;
    if (rrb < eps)
    { cil = ciu = rrt; }
    else
    {
      cil = (rrt - z * sqrt(vrrt)) / 2.0 / rrb / rrb;
      ciu = (rrt + z * sqrt(vrrt)) / 2.0 / rrb / rrb;
    }
  }
  mhrrvec[fidx].n(nt);  mhrrvec[fidx].r(rrm);
  mhrrvec[fidx].rl(cil);  mhrrvec[fidx].ru(ciu);
  if (0.0 < orm)
  {
    double ort = 2.0 * ora * orb;
    double vort = (4.0 * ora * orb + z * z * orc) * orc;
    if (orb < eps)
    { cil = ciu = orb; }
    else
    {
      cil = (ort + z * z * orc - z * sqrt(vort)) / 2.0 / orb / orb;
      ciu = (ort - z * z * orc + z * sqrt(vort)) / 2.0 / orb / orb;
    }
  }
  mhorvec[fidx].n(nt);  mhorvec[fidx].r(orm);
  mhorvec[fidx].rl(cil);  mhorvec[fidx].ru(ciu);
}

void mtestclass::cal_rrtest(const option& opt, vecrci& rrvec, int fidx, const listftab& ftablst0, int nt) const
// リスク比検定を行い各種統計量を返す
{
  double rr = 0.0, vrr = 0.0, ca = 0.0, ch = 0.0;
  int nc = rrtest(rr, vrr, ca, ch, ftablst0);
  int dfa = 1, dfh = nc - 1;
  if (nc < 1) return;
  ldbl pr = ctop(ca, dfa);
  double ln_rr = (rr == 0.0) ? 0.0 : log(rr);
  double z = ptoz(1.0 - opt.cc());
  double ci = z * sqrt(vrr);
  double ln_rrl = ln_rr - ci, ln_rru = ln_rr + ci;
  double rrl = exp(ln_rrl), rru = exp(ln_rru);
  rrvec[fidx].n(nt);  rrvec[fidx].r(rr);
  rrvec[fidx].rl(rrl);  rrvec[fidx].ru(rru);
}

void mtestclass::cal_lgtest(const option& opt, vecrci& orvec, int fidx, const listftab& ftablst0, int nt) const
// ロジット検定(オッズ比検定)を行い各種統計量を返す
{
  double odr = 0.0, vodr = 0.0, ca = 0.0, ch = 0.0;
  int dfa = 1, dfh = lgtest(odr, vodr, ca, ch, ftablst0);
  ldbl pr = ctop(ca, dfa);
  double ln_odr = (odr == 0.0) ? 0.0 : log(odr);
  double z = ptoz(1.0 - opt.cc());
  double ci = z * sqrt(vodr);
  double ln_odrl = ln_odr - ci, ln_odru = ln_odr + ci;
  double odrl = exp(ln_odrl), odru = exp(ln_odru);
  orvec[fidx].n(nt);  orvec[fidx].r(odr);
  orvec[fidx].rl(odrl);  orvec[fidx].ru(odru);
}

void mtestclass::put_ftable(ostream& fout, option& opt, listidd& idlst, const listftab& ftablst, int nftab) const
// 度数表と解析結果を出力する
{
  listidd::iterator iitr = idlst.begin();
  listftab::const_iterator fitr = ftablst.begin();
  size_t nr = fitr->nr(), nc = fitr->nc();
  idd idt(*iitr);               // 合計度数表用
  ftable ftabt(*fitr);          // 合計度数表用
  ftabt.fclear();
  int nt = 0, idno = 1;
  bool puttab = false;
  while (fitr != ftablst.end())
  {
    if (opt.res("sts"))         // 層ごとの結果出力
    {
      if (puttab) fout << "\n\n";
      put_resid(fout, idno, *iitr);
      put_ftabi(fout, opt, idno, *fitr, nftab, false);
      puttab = true;
    }
    ftable ftabtrm(*fitr);
    ftabtrm.trm();
    if (fitr->nr() != nr || fitr->nc() != nc || ftabtrm.nr() < 2 || ftabtrm.nc() < 2)
    {
      if (opt.res("sts")) fout << "\n[EXCLUDED FROM THE OVERALL ANALYSIS DUE TO INAPPROPRIATE TABLE !]\n";
    }
    else
    {
      nt += fitr->ftotal();
      for (size_t i = 1; i <= nr; ++i)
      {
        for (size_t j = 1; j <= nc; ++j) ftabt.f(i, j, ftabt.f(i, j) + fitr->f(i, j));
      }
    }
    ++fitr;  ++iitr;  ++idno;
  }
  if (opt.res("stt") == false) return; // 層だけ出力
  ftabt.set();
  if (puttab) fout << "\n\n";
  put_resid(fout, 0, idt);
  put_ftabi(fout, opt, idno, ftabt, nftab, true);
}

void mtestclass::put_resid(ostream& fout, int idno, idd& id) const
// 解析結果のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  if (idno > 0) id.put_stra(fout);
  else fout << "STRATIFY:ALL\n\n";
  for (size_t i = 0; i < id.ni(); ++i)
  {
    switch (id.eitem(i).aval())
    {
    case avalrow : fout << "GROUP ITEM(ROW)     :" << id.item(i);
      if (idno == 0) fout << "-TOTAL";
      fout << "\n";
      break;
    case avalcol : fout << "ORDINAL ITEM(COLUMN):" << id.item(i);
      if (idno == 0) fout << "-TOTAL";
      fout << "\n";
      break;
    default : id.put_item(fout, i);
    }
  }
}

void mtestclass::put_ftabi(ostream& fout, option& opt, int idno, const ftable& ftabi, int nftab, bool ftabt) const
// 個々の度数表と解析結果を出力する
{
  bool puttab = true;           // put_resid()を出力した後で実行するため
  if (opt.res("tab")) putftable(fout, ftabi, opt.tab());
  ftable ftabtrm(ftabi);
  ftabtrm.trm();
  if (ftabtrm.nr() < 2 || ftabtrm.nc() < 2) return; // 検定不可能
  listftab ftablsti;
  ftablsti.push_back(ftabtrm);
  if (ftabtrm.nr() == 2)        // 2群
  {
    if (ftabtrm.nc() == 2)      // 2順位
    {
      if (opt.res("mt"))
      {
        if (puttab) fout << "\n";
        put_mtest(fout, opt, idno - 1, ftablsti, nftab, ftabtrm.ftotal(), ftabt);
        puttab = true;
      }
      if (opt.res("rrt"))
      {
        if (puttab) fout << "\n";
        put_rrtest(fout, opt, idno - 1, ftablsti, nftab, ftabtrm.ftotal(), ftabt);
        puttab = true;
      }
      if (opt.res("lgt"))
      {
        if (puttab) fout << "\n";
        put_lgtest(fout, opt, idno - 1, ftablsti, nftab, ftabtrm.ftotal(), ftabt);
        puttab = true;
      }
      if (opt.res("emt"))
      {
        if (puttab) fout << "\n";
        put_emtest(fout, opt, ftablsti, nftab, ftabtrm.ftotal(), ftabt);
        puttab = true;
      }
    }
    else                        // 3順位以上(省略時は拡張Mantel検定出力)
    {
      if (opt.res("def") || opt.res("emt"))
      {
        if (puttab) fout << "\n";
        put_emtest(fout, opt, ftablsti, nftab, ftabtrm.ftotal(), ftabt);
        puttab = true;
      }
    }
  }
  if ((ftabi.nr() > 2 && opt.res("def")) || opt.res("gemt"))
  {                       // 3群以上(省略時は一般化拡張Mantel検定出力)
    if (puttab) fout << "\n";
    put_gemtest(fout, opt, ftablsti, nftab, ftabtrm.ftotal(), ftabt);
  }
}

void mtestclass::put_overall(ostream& fout, option& opt, const listftab& ftablst, int nftab, int nt) const
// 全体の解析結果を出力する
{
  fout << "\n<OVERALL>\n";
  bool puttab = false;
  listftab::const_iterator fitr = ftablst.begin();
  if (fitr->nr() == 2)          // 2群
  {
    if (fitr->nc() == 2)        // 2順位
    {
      if (opt.res("mt"))
      {
        put_mtest(fout, opt, nftab + 1, ftablst, nftab, nt, false);
        puttab = true;
      }
      if (opt.res("rrt"))
      {
        if (puttab) fout << "\n";
        put_rrtest(fout, opt, nftab + 1, ftablst, nftab, nt, false);
        puttab = true;
      }
      if (opt.res("lgt"))
      {
        if (puttab) fout << "\n";
        put_lgtest(fout, opt, nftab + 1, ftablst, nftab, nt, false);
        puttab = true;
      }
      if (opt.res("emt"))
      {
        if (puttab) fout << "\n";
        puttab = true;
        put_emtest(fout, opt, ftablst, nftab, nt, false);
      }
    }
    else                        // 3順位以上(省略時は拡張Mantel検定出力)
    {
      if (opt.res("def") || opt.res("emt"))
      {
        put_emtest(fout, opt, ftablst, nftab, nt, false);
        puttab = true;
      }
    }
  }
  if ((fitr->nr() > 2 && opt.res("def")) || opt.res("gemt"))
  {                       // 3群以上(省略時は一般化拡張Mantel検定出力)
    if (puttab) fout << "\n";
    put_gemtest(fout, opt, ftablst, nftab, nt, false);
  }
}

void mtestclass::put_mtest(ostream& fout, option& opt, int fidx, const listftab& ftablst0, int nftab, int nt, bool ftabt) const
// Mantel-Haenszelの検定結果を出力する
{
  double rdm = 0.0, vrdm = 0.0, rrm = 0.0, rra = 0.0, rrb = 0.0, rrc = 0.0;
  double orm = 0.0, ora = 0.0, orb = 0.0, orc = 0.0, cmm = 0.0, cmr = 0.0;
  int dfm = 1;
  int dfr = mtest(rdm, vrdm, rrm, rra, rrb, rrc, orm, ora, orb, orc, cmm, cmr, ftablst0, opt.mt());
  double ln_rrm = (rrm == 0.0) ? 0.0 : log(rrm);
  double ln_orm = (orm == 0.0) ? 0.0 : log(orm);
  double z = ptoz(1.0 - opt.cc());
  double ci = z * sqrt(vrdm);
  double cil = rdm - ci, ciu = rdm + ci;
  ldbl pr = 0.0L;
  double eps = DBL_EPSILON * fmax(5.0, vrdm);
  fout << "Mantel-Haenszel Test\n";
  if (opt.res("r")) put_mr(fout, ftablst0, nt);
  fout << "  ";
  if (dfr > 0) fout << "Adjusted ";
  fout << "Risk Difference:RD=" << rdm << "\n";
  if (opt.res("cc")) fout << "    " << opt.cc() * 100.0 << "% CI of RD:Lower=" << cil << "  Upper=" << ciu << "\n";
  if (0.0 < rrm)
  {
    double rrt = 2.0 * rra * rrb + z * z * rrc;
    double vrrt = (4.0 * rra * rrb + z * z * rrc) * rrc;
    if (rrb < eps)
    { cil = ciu = rrt; }
    else
    {
      cil = (rrt - z * sqrt(vrrt)) / 2.0 / rrb / rrb;
      ciu = (rrt + z * sqrt(vrrt)) / 2.0 / rrb / rrb;
    }
  }
  if (opt.res("mhrr"))          // リスク比出力
  {
    fout << "  ";
    if (dfr > 0) fout << "adjusted ";
    fout << "Risk Ratio:RR=" << rrm << "  ln(RR)=" << ln_rrm << "\n";
    if (0.0 < rrm && opt.res("cc")) fout << "    " << opt.cc() * 100.0 << "% CI of RR:Lower=" << cil << "  Upper=" << ciu << "\n";
  }
  if (0.0 < orm)
  {
    double ort = 2.0 * ora * orb;
    double vort = (4.0 * ora * orb + z * z * orc) * orc;
    if (orb < eps)
    { cil = ciu = orb; }
    else
    {
      cil = (ort + z * z * orc - z * sqrt(vort)) / 2.0 / orb / orb;
      ciu = (ort - z * z * orc + z * sqrt(vort)) / 2.0 / orb / orb;
    }
  }
  if (opt.res("mhor"))          // オッズ比出力
  {
    fout << "  ";
    if (dfr > 0) fout << "Adjusted ";
    fout << "Odds Ratio:OR=" << orm << "  ln(OR)=" << ln_orm << "\n";
    if (0.0 < orm && opt.res("cc")) fout << "    " << opt.cc() * 100.0 << "% CIof OR:Lower=" << cil << "  Upper=" << ciu << "\n";
  }
  pr = ctop(cmm, dfm);
  fout << "  test of RD";
  if (opt.mt() == testorg) fout << "(no Yates's cor)";
  else fout << "(with Yates's cor)";
  fout << ":chi-squared=" << cmm << "  DF=" << dfm << "  p=" << pr << pind(pr) << "\n";
  if (ftabt) return;            // 全度数表合計
  if (dfr < 1)
  {
    if (opt.res("m") == false) return;
    pr *= (ldbl)nftab;
    if (pr > 1.0L) pr = 1.0L;
    fout << "  Bonferroni Type Multiple Comparison:p=" << pr << pind(pr) << "\n";
    if (opt.res("cc"))
    {
      z = ptoz((1.0 - opt.cc()) / (double)nftab);
      ci = z * sqrt(vrdm);
      fout << "    " << opt.cc() * 100.0 << "% Simultaneous CI of RD:Lower=" << rdm - ci << "  Upper=" << rdm + ci << endl;
    }
    return;
  }
  pr = ctop(cmr, dfr);
  fout << "  Test of Heterogeneity:chi-squared=" << cmr << "  DF=" << dfr << "  p=" << pr << pind(pr) << "\n";
  if (opt.res("bt"))
  {
    dfr = mbtest(orm, cmm, cmr, ftablst0, opt.mt());
    pr = ctop(cmr, dfr);
    fout << "  Breslow-Day Test(heterogeneity):chi-squared=" << cmr << "  DF=" << dfr << "  p=" << pr << pind(pr) << "\n";
  }
}

void mtestclass::put_mr(ostream& fout, const listftab& ftablst0, int nt) const
// Mantel-Haenszelの順位相関比を出力する
{
  double rdm = 0.0, vrdm = 0.0, rrm = 0.0, rra = 0.0, rrb = 0.0, rrc = 0.0;
  double orm = 0.0, ora = 0.0, orb = 0.0, orc = 0.0, cmm = 0.0, cmr = 0.0;
  mtest(rdm, vrdm, rrm, rra, rrb, rrc, orm, ora, orb, orc, cmm, cmr, ftablst0, testorg);
  double r2 = cmm / (double)(nt - 1);
  int dfm = 1;
  ldbl pr = ctop(cmm, dfm);
  fout << "  Rank Correlation Ratio:rr=" << sqrt(r2) << "  rr-squared=" << r2 << "\n";
  fout << "  Test of rr:chi-squared=" << cmm << "  DF=" << dfm << "  p=" << pr << pind(pr) << endl;
}

void mtestclass::put_rrtest(ostream& fout, option& opt, int fidx, const listftab& ftablst0, int nftab, int nt, bool ftabt) const
// リスク比検定の検定結果を出力する
{
  double rr = 0.0, vrr = 0.0, ca = 0.0, ch = 0.0;
  int nc = rrtest(rr, vrr, ca, ch, ftablst0);
  int dfa = 1, dfh = nc - 1;
  if (nc < 1)
  {
    fout << "\nRisk Ratio Test:IMPOSSIBLE\n";
    return;
  }
  ldbl pr = ctop(ca, dfa);
  double ln_rr = (rr == 0.0) ? 0.0 : log(rr);
  double z = ptoz(1.0 - opt.cc());
  double ci = z * sqrt(vrr);
  double ln_rrl = ln_rr - ci, ln_rru = ln_rr + ci;
  double rrl = exp(ln_rrl), rru = exp(ln_rru);
  fout << "Risk Ratio Test\n";
  fout << "  ";
  if (dfh > 0) fout << "Adjusted ";
  fout << "Risk Ratio:RR=" << rr << "  ln(RR)=" << ln_rr << "\n";
  if (opt.res("cc")) fout << "    " << opt.cc() * 100.0 << "% CI of RR:Lower=" << rrl << "  Upper=" << rru << "\n";
  fout << "  Test of ";
  if (dfh > 0) fout << "Adjusted ";
  fout << "Risk Ratio:chi-squared=" << ca << "  DF=" << dfa << "  p=" << pr << pind(pr) << "\n";
  if (ftabt) return;            // 全度数表合計
  if (nc > 1)
  {
    pr = ctop(ch, dfh);
    fout << "  Test of Heterogeneity:chi-squared=" << ch << "  DF=" << dfh << "  p=" << pr << pind(pr) << endl;
  }
}

void mtestclass::put_lgtest(ostream& fout, option& opt, int fidx, const listftab& ftablst0, int nftab, int nt, bool ftabt) const
// ロジット検定(オッズ比検定)の検定結果を出力する
{
  double odr = 0.0, vodr = 0.0, ca = 0.0, ch = 0.0;
  int dfa = 1, dfh = lgtest(odr, vodr, ca, ch, ftablst0);
  ldbl pr = ctop(ca, dfa);
  double ln_odr = (odr == 0.0) ? 0.0 : log(odr);
  double z = ptoz(1.0 - opt.cc());
  double ci = z * sqrt(vodr);
  double ln_odrl = ln_odr - ci, ln_odru = ln_odr + ci;
  double odrl = exp(ln_odrl), odru = exp(ln_odru);
  fout << "Logit Test\n";
  fout << "  ";
  if (dfh > 0) fout << "Adjusted ";
  fout << "Odds Ratio:OR=" << odr << "  ln(OR)=" << ln_odr << "\n";
  if (opt.res("cc")) fout << "    " << opt.cc() * 100.0 << "% CI of OR:Lower=" << odrl << "  Upper=" << odru << "\n";
  fout << "  Test of ";
  if (dfh > 0) fout << "Adjusted ";
  fout << "OR:chi-squared=" << ca << "  DF=" << dfa << "  p=" << pr << pind(pr) << "\n";
  if (ftabt) return;            // 全度数表合計
  if (dfh < 1)
  {
    if (opt.res("m") == false) return;
    pr *= (double)nftab;
    if (pr > 1.0) pr = 1.0;
    fout << "  Bonferroni Type Multiple Comparison:p=" << pr << pind(pr) << "\n";
    if (opt.res("cc"))
    {
      ci = ptoz((1.0 - opt.cc()) / (double)nftab) * sqrt(vodr);
      ln_odrl = ln_odr - ci;  ln_odru = ln_odr + ci;
      fout << "    " << opt.cc() * 100.0 << "% Simultaneous CI of OR:Lower=" << exp(ln_odrl) << "  Upper=" << exp(ln_odru) << endl;
    }
    return;
  }
  pr = ctop(ch, dfh);
  fout << "  Test of Heterogeneity:chi-squared=" << ch << "  DF=" << dfh << "  p=" << pr << pind(pr) << endl;
}

void mtestclass::put_emtest(ostream& fout, option& opt, const listftab& ftablst0, int nftab, int nt, bool ftabt) const
// 拡張Mantel検定の結果を出力する
{
  listftab::const_iterator fitr = ftablst0.begin();
  double cmm = 0.0, cmr = 0.0, rd = 0.0, vrd = 0.0;
  int dfm = (int)(fitr->nr()) - 1, dfr = emtest(cmm, cmr, rd, vrd, ftablst0);
  ldbl pr = ctop(cmm, dfm);
  double r2 = cmm / (double)(nt - 1);
  fout << "Mantel-extension Test\n";
  if (opt.res("r")) fout << "  Rank Correlation Ratio:rr=" << sqrt(r2) << "  rr-squared=" << r2 << "\n";
  double pro = 100.0 / (double)nt;
  double ci = 0.0, rdl = 0.0, rdu = 0.0;
  fout << "  ";
  if (dfr > 0) fout << "Adjusted ";
  fout << "Difference of Rank Mean(%)=" << rd << "(" << rd * pro << "%)\n";
  if (opt.res("cc"))
  {
    ci = ptoz(1.0 - opt.cc()) * sqrt(vrd);
    rdl = rd - ci;  rdu = rd + ci;
    fout << "    " << opt.cc() * 100.0 << "%CI of Drm=" << rdl  << "-" << rdu << "(" << rdl * pro << "%-" << rdu * pro << "%)\n";
  }
  fout << "  Test of ";
  if (dfr > 0) fout << "Adjusted ";
  fout << "Drm:chi-squared=" << cmm << "  DF=" << dfm << "  p=" << pr << pind(pr) << "\n";
  if (ftabt) return;            // 全度数表合計
  if (dfr < 1)
  {
    if (opt.res("m") == false) return;
    pr *= (ldbl)nftab;
    if (pr > 1.0L) pr = 1.0L;
    fout << "  Bonferroni Type Multiple Comparison:p=" << pr << pind(pr) << "\n";
    if (opt.res("cc"))
    {
      ci = ptoz((1.0 - opt.cc()) / (double)nftab) * sqrt(vrd);
      rdl = rd - ci;  rdu = rd + ci;
      fout << "    " << opt.cc() * 100.0 << "% Simultaneous CI of Drm=" << rdl  << "-" << rdu << "(" << rdl * pro << "%-" << rdu * pro << "%)\n";
    }
    return;
  }
  pr = ctop(cmr, dfr);
  fout << "  Test of Heterogeneity:chi-squared=" << cmr << "  DF=" << dfr << "  p=" << pr << pind(pr) << endl;
}

void mtestclass::put_gemtest(ostream& fout, option& opt, const listftab& ftablst0, int nftab, int nt, bool ftabt) const
// 一般化拡張Mantel検定の結果を出力する
{
  listftab::const_iterator fitr = ftablst0.begin();
  double cmm = 0.0, cmr = 0.0;
  int dfm = (int)(fitr->nr()) - 1, dfr = gemtest(cmm, cmr, ftablst0);
  ldbl pr = ctop(cmm, dfm);
  double r2 = cmm / (double)(nt - 1);
  fout << "Generalized Mantel-extension Test\n";
  if (opt.res("r")) fout << "  Rank Correlation Ratio:rr=" << sqrt(r2) << "  rr-squared=" << r2 << "\n";
  fout << "  Test of rr:chi-squared=" << cmm << "  DF=" << dfm << "  p=" << pr << pind(pr) << "\n";
  if (ftabt) return;            // 全度数表合計
  if (dfr < 1)
  {
    if (opt.res("m") == false) return;
    pr *= (ldbl)nftab;
    if (pr > 1.0L) pr = 1.0L;
    fout << "  Bonferroni Type Multiple Comparison:p=" << pr << pind(pr) << "\n";
    return;
  }
  pr = ctop(cmr, dfr);
  fout << "  Test of Heterogeneity:chi-squared=" << cmr << "  DF=" << dfr << "  p=" << pr << pind(pr) << endl;
}

void mtestclass::plot_ngp(ofstream& fout, option& opt, listidd& idlst, int nftab, int nt, vecrci& mhorvec, vecrci& mhrdvec, vecrci& mhrrvec, vecrci& orvec, vecrci& rrvec) const
// Ngraph用データを出力する
{
  bool delml = false;
  if (opt.ngp("mhrd"))          // 調整リスク差
  {
    fout << "-s1 -dline\n";
    plot_ngpid(fout, idlst);
    plot_rcivec(fout, "M-H ADJUSTED RISK DIFFERENCE", nftab, nt, mhrdvec);
    delml = true;
  }
  if (opt.ngp("mhrr"))          // 調整リスク比
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80\n";
      delml = true;
    }
    plot_ngpid(fout, idlst);
    plot_rcivec(fout, "M-H ADJUSTED RISK RATIO", nftab, nt, mhrrvec);
  }
  if (opt.ngp("mhor"))          // 調整オッズ比
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80\n";
      delml = true;
    }
    plot_ngpid(fout, idlst);
    plot_rcivec(fout, "M-H ADJUSTED ODDS RATIO", nftab, nt, mhorvec);
  }
  if (opt.ngp("rr"))            // リスク比
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80\n";
      delml = true;
    }
    plot_ngpid(fout, idlst);
    plot_rcivec(fout, "RISK RATIO", nftab, nt, rrvec);
  }
  if (opt.ngp("or"))            // オッズ比
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dline -w80\n";
    plot_ngpid(fout, idlst);
    plot_rcivec(fout, "ODDS RATIO", nftab, nt, orvec);
  }
}

void mtestclass::plot_ngpid(ofstream& fout, listidd& idlst) const
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  listidd::iterator iitr = idlst.begin();
  iitr->put_title(fout, formngp);
  iitr->put_stra(fout, formngp);
  for (size_t i = 0; i < iitr->ni(); ++i)
  {
    fout << "%";
    switch (iitr->eitem(i).aval())
    {
    case avalrow : fout << "GROUP ITEM(ROW)  :" << iitr->item(i) << "\n";
      break;
    case avalcol : fout << "ORDINAL ITEM(COL):" << iitr->item(i) << "\n";
      break;
    default : iitr->put_item(fout, i);
    }
  }
}

void mtestclass::plot_rcivec(ofstream& fout, const string& rname, int nftab, int nt, vecrci& rcivec) const
// 個々のリスク比と信頼区間をNgraph用に出力する
{
  int ymax = nftab + 1;
  double xmin = FLT_MAX, xmax = 0.0;
  fout << "%<" << rname << " CI(line)>\n";
  for (int i = 0; i < nftab; ++i)
  {
    fout << rcivec[i].rl() << " " << ymax - i << "\n";
    fout << rcivec[i].ru() << " " << ymax - i << "\n";
    fout << "= =\n";
    if (rcivec[i].rl() < xmin) xmin = rcivec[i].rl();
    if (xmax < rcivec[i].ru()) xmax = rcivec[i].ru();
  }
  double xw = (xmax - xmin) * 0.1, x = 0.0, y = 0.0;
  double yw = (nftab > 8) ? 0.9 : (double)ymax * 0.1;
  fout << NGPdelm << "\n";
  fout << "%<" << rname << " POINT ESTIMATE(fill rectangle)>\n";
  for (int i = 0; i < nftab; ++i)
  {
    double w = (double)(rcivec[i].n()) / (double)nt;
    x = rcivec[i].r() - xw * 0.5 * w;
    y = (double)(ymax - i) - yw * 0.5 * w;
    fout << x << " " << y << " ";
    x = rcivec[i].r() + xw * 0.5 * w;
    y = (double)(ymax - i) + yw * 0.5 * w;
    fout << x << " " << y << "\n";
  }
  fout << NGPdelm << "\n";
  fout << "%<" << rname << " OVERALL CI(solid polygon)>\n";
  fout << rcivec[nftab + 1].rl() << " " << 1 << "\n";
  y = 1.0 + yw * 0.5;
  fout << rcivec[nftab + 1].r() << " " << y << "\n";
  fout << rcivec[nftab + 1].ru() << " " << 1 << "\n";
  y = 1.0 - yw * 0.5;
  fout << rcivec[nftab + 1].r() << " " << y << endl;
}

int mtestclass::get(const char* c_fname, option& opt, listidd& idlst, listftab& ftablst)
// 度数集計表を読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nftab = fgetftable(fname, idlst, ftablst);
  if (nftab < 0) throw runtime_error("Run-time error !");
  if (opt.check() != 0) throw option_error("Run-time option error !");
  return nftab;                 // 読み込んだ度数表数を返す
}

void mtestclass::put(option& opt, listidd& idlst, listftab& ftablst, int fno) const
// 集計結果を出力する
{
  int nftabok = check_ftable(ftablst);
  if (opt.skip() == false && nftabok < (int)(ftablst.size())) throw runtime_error("cross-tabulatiion table is uneven !");
  vecrci mhorvec(nftabok + 2);  // Mantel-Haenzelの調整オッズ比
  vecrci mhrdvec(nftabok + 2);  // Mantel-Haenzelの調整リスク差
  vecrci mhrrvec(nftabok + 2);  // Mantel-Haenzelの調整リスク比
  vecrci orvec(nftabok + 2);    // オッズ比
  vecrci rrvec(nftabok + 2);    // リスク比
  int nt = cal_mhstat(opt, idlst, ftablst, mhorvec, mhrdvec, mhrrvec, orvec, rrvec);
  if (opt.res("res"))
  {
    put_ftable(cout, opt, idlst, ftablst, nftabok);
    if (nftabok < 2)
    {
      cout << "\n" << endl;
      return;
    }
    if (opt.res("stt") && (opt.res("mt") || opt.res("rrt") || opt.res("lgt") || opt.res("emt") || opt.res("gemt"))) put_overall(cout, opt, ftablst, nftabok, nt);
    cout << "\n" << endl;
  }
  if (opt.ngpname(fno) != "")
  {
    ofstream foutngp;
    if (ofopen(foutngp, (opt.ngpname(fno)).c_str()) == false) return;
    plot_ngp(foutngp, opt, idlst, nftabok, nt, mhorvec, mhrdvec, mhrrvec, orvec, rrvec);
    foutngp.close();
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  mtestclass mtest;             // メインクラス変数
  option opt(argv[0]);          // オプション
  listidd idlst;                // データ識別情報リスト
  listftab ftablst;             // 度数表リスト
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    int fno = 1;                // 有効な度数表の数とファイル番号
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok)                // 標準入力から読み込む
        {
          mtest.get(nullptr, opt, idlst, ftablst);
          mtest.put(opt, idlst, ftablst, fno);
          ++fno;
        }
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      {
        mtest.get(argv[i], opt, idlst, ftablst);  fok = false;
        mtest.put(opt, idlst, ftablst, fno);
        ++fno;
      }
    }
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
