// trende.C Version 7.42e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-28 14:00:55 tac>
//
// クロス集計表を入力し、Cochran-Armitageの傾向分析とロジット分析を行う。
// クロス集計表は計数値解析コマンド用の標準入力形式に従っている必要がある。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W trende trende.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "trende  Version 7.42e        Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:trende  [<OPTION> ...]  [<TAB FILE> ...]",
  "  The trend test with cross-tabulation table.",
  "<OPTION>",
  "  -f=<FILE>:read options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -ngp={log|pl|pro}[,...]:select the data for Ngraph",
  "    log:logistic curve  pl:proportion line  pro:proportion",
  "    default:all",
  "  -ngpname[=<FILE>]:specify the file to output Ngraph data",
  "    default:no output",
  "    -ngpname:-ngpname=trende_ngp<2-digit number>.dat",
  "  -res=[-][{def|all|ca|cc[<C-COEF>]|log[{c|w}]|tab[{fr|pc|pr|pt}]}]",
  "      [,...]:select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    ca:Cochran-Armitage's trend analysis",
  "    cc[<C-COEF>]:confidence coefficient(0=<C-COEF<1 cc:cc0.95)",
  "    log[{c|w}]:logit analysis",
  "      c:repetitive completion method  w:working logit method",
  "    tab[{fr|p|pc|pr|pt}]:cross-tabulation table",
  "      fr:only frequency  pc:also output % for column total",
  "      pr:also output % for row total  pt:also output % for all total",
  "    default:auto judgment according to data"
};

const string Title = "=== Trend Analysis of Proportion ==="; // 表題
const string NGPNAMEdef = "trende_ngp"; // 省略時Ngraphファイル基本名

enum logtype                    // ロジット分析の種類
{
  lognon,                       // 実行せず
  logcomp,                      // 繰り返し補完法
  logwork,                      // 実用ロジット法
  logetc                        // その他
};

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  logtype vlog;                 // ロジット分析の種類
  string vngpname;              // Ngraphファイル基本名
  string vngpnamen;             // Ngraphファイル名(番号付き)
  lstr vngplst;                 // Ngraph出力リスト
  map<string, bool> vngpmap;    // Ngraph出力の有無マップ
  lstr vreslst;                 // 解析結果出力リスト
  map<string, bool> vresmap;    // 解析結果出力の有無マップ
  tabtype vtab;                 // 度数集計表の種類
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  logtype log() const           // ロジット分析の種類を返す
  { return vlog; }
  const string& ngpname(int fno); // Ngraphデータファイル名を返す
  bool ngp(const string& code); // Ngraph出力フラッグを返す
  bool res(const string& code); // 解析結果出力フラッグを返す
  tabtype tab() const           // 度数集計表の種類を返す
  { return vtab; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check();                  // オプションのチェックとセット
  int set_ngp();                // -ngpオプションのセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vlog(logcomp), vngpname(), vngpnamen(), vngplst(), vngpmap(), vreslst(), vresmap(), vtab(tabfprr)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

const string& option::ngpname(int fno) // Ngraphデータファイル名を返す
{
  if (vngpname == "") return vngpname; // -ngpname指定なし
  ostringstream ostr;
  if (fno < 10) ostr << vngpname << "0" << fno << ".dat";
  else ostr << vngpname << fno << ".dat";
  vngpnamen = (string)(ostr.str());
  return vngpnamen;             // 番号付きファイル名を返す
}

bool option::ngp(const string& code) // Ngraph出力フラッグを返す
{
  map<string, bool>::iterator ngpitr = vngpmap.find(code);
  if (ngpitr == vngpmap.end()) return false; // code無し！
  return vngpmap[code];                     // true/falseを返す
}

bool option::res(const string& code) // 分析結果出力フラッグを返す
{
  map<string, bool>::iterator resitr = vresmap.find(code);
  if (resitr == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "cc")           // 信頼係数の指定
  {
    vcc = opt[1].empty() ? CCdef : strtodf(opt[1]);
    if (vcc < 0.0 || 1.0 <= vcc) err = badopt(vpname, optstr, USAGE, 0.0, 1.0);
  }
  else if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "ngp")     // Ngraph出力項目指定
  {
    if (strsplit(vngplst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else if (opt[0] == "ngpname")     // Ngraphファイル基本名指定
  {
    if (opt[1].empty()) vngpname = NGPNAMEdef;
    else vngpname = opt[1];
  }
  else if (opt[0] == "res")     // 分析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this message");
  if (err < 0) throw option_error(string("Option specification error->") + optstr);
}

int option::check()             // オプションのチェックとセット
{
  if (set_ngp() < 0) return -2; // -ngpオプションのエラー
  if (set_res() < 0) return -3; // -resオプションのエラー
  return 0;                     // エラーなし
}

int option::set_ngp()           // -ngpオプションのセット
{
  vngpmap["log"] = vngpmap["pl"] = vngpmap["pro"] = true;
  if (vngplst.size() > 0)
  {
    if (vngpname == "") return disperr(vpname, "Requires -ngpname option !");
    vngpmap["log"] = vngpmap["pl"] = vngpmap["pro"] = false;
    for (string ngp : vngplst)
    {
      map<string, bool>::iterator ngpitr = vngpmap.find(ngp);
      if (ngpitr == vngpmap.end())
      { return disperr(vpname, "Wrong -ngp=" + ngp + " option !"); }
      else
      { vngpmap[ngp] = true; }
    }
  }
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["ca"] = vresmap["cc"] = vresmap["tab"] = true;
  vresmap["log"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["ca"] = vresmap["cc"] = vresmap["tab"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator resitr = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        resitr = vresmap.find(res.substr(1));
        if (resitr == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      {
        vresmap["def"] = vresmap["ca"] = vresmap["cc"] = vresmap["tab"] = true;
      }
      else if (res == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["ca"] = vresmap["cc"] = vresmap["tab"] = vresmap["log"] = true;
        vlog = logcomp;  vtab = tabfprr;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (res == "logc")
      { vresmap["log"] = true;  vlog = logcomp; }
      else if (res == "logw")
      { vresmap["log"] = true;  vlog = logwork; }
      else if (res == "tabfr")
      { vresmap["tab"] = true;  vtab = tabfreq; }
      else if (res == "tabpr")
      { vresmap["tab"] = true;  vtab = tabfprr; }
      else if (res == "tabpc")
      { vresmap["tab"] = true;  vtab = tabfprc; }
      else if (res == "tabpt")
      { vresmap["tab"] = true;  vtab = tabfprt; }
      else if (resitr == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class trendclass                // trendメインクラス
{
private:
  string fname;                 // ファイル名
  listidd idlst;                // データ識別情報リスト
  listftab ftablst;             // 度数表リスト
  void put_id(ostream& fout, idd& id) const;
  // 度数表のタイトルとデータ識別情報を出力する
  void put_trend(ostream& fout, option& opt, const ftable& ftab) const;
  // 回帰係数の信頼区間を出力する
  void put_ccoef(ostream& fout, const option& opt, double b, double sxxw) const;
  // 回帰係数の信頼区間を出力する
  void put_cor(ostream& fout, const option& opt, double b, double sxxw) const;
  void put_catab(ostream& fout, const string& head, double syyw, double sb, int dft) const;
  // Cochran-Armitageの傾向分析の分散分析表を出力する
  int cal_caline(double& a, double& b, double& sxxw, double& syyw, double &sb, int& dft, const ftable& ftab) const;
  // Cochran-Armitageの傾向分析
  int cal_lpline(const option& opt, double& a, double& b, double& sxxw, double& syyw, double &sb, int& dft, const ftable& ftab) const;
  // ロジット分析の繰り返し計算
  int cal_logit(double& a, double& b, double& sxxw, double& syyw, double &sb, int& dft, const ftable& ftab, bool comp) const;
  // ロジット分析
  void plot_ngp(ofstream& fout, option& opt, idd& id, const ftable& ftab) const;
  // Ngraph用データを出力する
  void plot_ngpid(ofstream& fout, idd& id) const;
  // Ngraph用データのタイトルとデータ識別情報を出力する
  void plot_ngplog(ofstream& fout, double a, double b, const ftable& ftab) const;
  // Ngraph用にロジスティック曲線を出力する
  void plot_ngppl(ofstream& fout, double a, double b, const ftable& ftab) const;
  // Ngraph用に出現率直線を出力する
  void plot_ngppro(ofstream& fout, const ftable& ftab) const;
  // Ngraph用に出現率を出力する
public:
  trendclass() : fname(), idlst(), ftablst() // コンストラクタ
  {}
  ~trendclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // 度数集計表を読み込む
  void put(option& opt, int fno);            // 結果を出力する
};

void trendclass::put_id(ostream& fout, idd& id) const
// 度数表のタイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  for (size_t i = 0; i < id.ni(); ++i)
  {
    switch (id.eitem(i).aval())
    {
    case avalrow : fout << "EXPLANATORY VARIAVLE(ROW) :" << id.item(i) << "\n";
      break;
    case avalcol : fout << "CRITERION VARIABLE(COLUMN):" << id.item(i) << "\n";
      break;
    default : id.put_item(fout, i);
    }
  }
}

void trendclass::put_trend(ostream& fout, option& opt, const ftable& ftab) const
// 回帰係数の信頼区間を出力する
{
  string ccode = ftab.ccode(2);
  strtrm(ccode);
  double a = 0.0, b = 0.0, sxxw = 0.0, syyw = 0.0, sb = 0.0;
  int nr = 0, dft = 0;
  bool puttab = false;
  if (opt.res("ca"))            // Cochran-Armitageの傾向分析出力
  {
    nr = cal_caline(a, b, sxxw, syyw, sb, dft, ftab);
    fout << "Proportion Line of Column Code=" << ccode << ":p=" << a << showpos << b << noshowpos << "x\n";
    if (opt.res("cc")) put_ccoef(fout, opt, b, sxxw);
    double r2 = sb / syyw;
    fout << "  r-squared=" << r2 << "  r=" << sqrt(r2) << "\n";
    put_catab(fout, "Cochran-Armitage's Trend Analysis", syyw, sb, dft);
    puttab = true;
  }
  if (opt.res("log"))           // ロジット分析出力
  {
    nr = cal_lpline(opt, a, b, sxxw, syyw, sb, dft, ftab);
    if (puttab) fout << "\n";
    fout << "Proportion Curve of Column Code=" << ccode << ":p=1/{1+exp(" << -a << showpos << -b << noshowpos << "x)}\n";
    fout << "  p0=1/{1+exp(" << -a << ")}=" << 1.0 / (1.0 + exp(-a));
    if (nr > 0) fout << "  (number of repeated completions=" << nr << ")";
    fout << "\n";
    if (opt.res("cc"))
    {
      put_ccoef(fout, opt, b, sxxw);
      put_cor(fout, opt, b, sxxw);
    }
    double r2 = sb / syyw;
    fout << "  r-squared=" << r2 << "  r=" << sqrt(r2) << "\n";
    put_catab(fout, "Logit Analysis", syyw, sb, dft);
  }
}

void trendclass::put_ccoef(ostream& fout, const option& opt, double b, double sxxw) const
// 回帰係数の信頼区間を出力する
{
  double z = ptoz(1.0 - opt.cc());
  double ccpr = opt.cc() * 100.0;
  double ci = (sxxw > 0.0) ? z * sqrt(1.0 / sxxw) : 0.0;
  fout << "  " << ccpr << "% CI of b=";
  fout << b << " +/- " << ci << "(" << b - ci << " - " << b + ci << ")\n";
}

void trendclass::put_cor(ostream& fout, const option& opt, double b, double sxxw) const
// オッズ比の信頼区間を出力する
{
  double z = ptoz(1.0 - opt.cc());
  double ccpr = opt.cc() * 100.0;
  double ci = (sxxw > 0.0) ? z * sqrt(1.0 / sxxw) : 0.0;
  fout << "  Odds Ratio:OR=" << exp(b) << "  " << ccpr << "% CI of OR=";
  fout << exp(b - ci) << " - " << exp(b + ci) << endl;
}

void trendclass::put_catab(ostream& fout, const string& head, double syyw, double sb, int dft) const
// Cochran-Armitage分析またはロッジト分析の分散分析表を出力する
{
  int w = strclen(head, CHARwide);
  int npad = head.size() - w;
  fout << "\n" << setw((68 + w) / 2 + npad) << head;
  fout << "\nFACTOR      Sum of Squares        DF   Chi-squared       p-value\n";
  constexpr int LENline = 68;
  putline(fout, LENline);
  int dfb = 1;
  int dfl = dft - dfb;
  double sl = syyw - sb;
  double eps = DBL_EPSILON * fmax(5.0, syyw);
  if (dfl < 1 || sl < eps) sl = 0.0;
  ldbl pr = ctop(sb, dfb);
  if (head == "Cochran-Armitage's Trend Analysis") fout << "LINEARITY" << setw(17) << sb;
  else fout << "REGRESSION" << setw(16) << sb;
  fout << setw(10) << dfb;
  fout << setw(14) << sb << setw(14) << pr << pind(pr) << "\n";
  pr = ctop(sl, dfl);
  fout << "LOF" << setw(23) << sl << setw(10) << dfl;
  fout << setw(14) << sl << setw(14) << pr << pind(pr) << "\n";
  putline(fout, LENline);
  pr = ctop(syyw, dft);
  fout << "TOTAL" << setw(21) << syyw << setw(10) << dft;
  if (head == "Cochran-Armitage's Trend Analysis") fout << setw(14) << syyw << setw(14) << pr << pind(pr);
  fout << endl;
}

int trendclass::cal_caline(double& a, double& b, double& sxxw, double& syyw, double &sb, int& dft, const ftable& ftab) const
// Cochran-Armitageの傾向分析
{
  int sn = 0, sr = 0;
  double mx = 0.0, sxyw = 0.0;
  sxxw = syyw = sb = 0.0;  dft = 0;
  for (size_t i = 1; i <= ftab.nr(); ++i)
  {
    double x = strtodf(ftab.rcode(i));
    int n = ftab.frtotal(i), r = ftab.f(i, 2);
    if (x != BLANKf && n > 0)
    {
      ++dft;  sn += n;  sr += r;
      mx += (double)n * x;  sxxw += (double)n * x * x;
      syyw += (double)r * (double)r / (double)n;  sxyw += double(r) * x;
    }
  }
  --dft;
  if (dft < 1)
  {
    sxxw = syyw = sb = 0.0;
    return -1;                  // ロジット分析不可能
  }
  const double Seps = DBL_EPSILON * fmax(5.0, syyw);
  double pc = (double)sr / (double)sn;
  double w = 1.0 / pc / (1.0 - pc);
  mx /= (double)sn;
  sxxw = w * (sxxw - (double)sn * mx * mx);
  if (sxxw < Seps) sxxw = 0.0;
  syyw = w * (syyw - (double)sn * pc * pc);
  if (syyw < Seps) syyw = 0.0;
  sxyw = w * (sxyw - (double)sn * mx * pc);
  if (fabs(sxyw) < Seps) sxyw = 0.0;
  sb = sxyw * sxyw / sxxw;
  if (fabs(sb) < Seps) sb = 0.0;
  b = (sxxw > 0.0) ? sxyw / sxxw : 0.0;
  a = pc - b * mx;
  return dft;                   // 自由度を返す
}

int trendclass::cal_lpline(const option& opt, double& a, double& b, double& sxxw, double& syyw, double &sb, int& dft, const ftable& ftab) const
// ロジット分析の繰り返し計算
{
 constexpr double EPSdef = 0.00001; // 省略時収束誤差
 constexpr int Rmax = 100;          // 繰り返し数最大値
 int nr = 0;
 int nest = cal_logit(a, b, sxxw, syyw, sb, dft, ftab, false);
  const double Eps = syyw * EPSdef;
  if (nest > 0 && opt.log() == logcomp) // 繰り返し補完法
  {
    double syywpre = syyw;
    nest = 0;
    bool comp = true;
    while (comp)
    {
      nest = cal_logit(a, b, sxxw, syyw, sb, dft, ftab, true);
      ++nr;
      if (nest < 0 || fabs(syyw - syywpre) < Eps || nr >= Rmax) comp = false;
      else syywpre = syyw;
    }
    double at = a, bt = b;    // 収束後のaとbで平方和を求める
    nest = cal_logit(at, bt, sxxw, syyw, sb, dft, ftab, true);
  }
  if (nest < 0) return -1;      // 計算できず
  return nr;                    // 繰り返し数を返す
}

int trendclass::cal_logit(double& a, double& b, double& sxxw, double& syyw, double &sb, int& dft, const ftable& ftab, bool comp) const
// ロジット分析
{
  double snw = 0.0, mxw = 0.0, myw = 0.0, sxyw = 0.0;
  sxxw = syyw = sb = 0.0;  dft = 0;
  int nest = 0;                 // 推測値で補完した用量数
  double y = 0.0, nw = 0.0;
  for (size_t i = 1; i <= ftab.nr(); ++i)
  {
    double x = strtodf(ftab.rcode(i));
    int n = ftab.frtotal(i), r = ftab.f(i, 2);
    if (x != BLANKf && n > 0)
    {
      if (0 < r && r < n)
      {
        y = log((double)r / (double)(n - r));
        nw = (double)r / (double)n * (double)(n - r);
      }
      else
      {
        if (comp)               // 推測値で補完する
        {
          double p = 1.0 / (1.0 + exp(-a - b * x));
          y = log(p / (1.0 - p));  nw = (double)n * p * (1.0 - p);
        }
        else                    // 実用ロジットを用いる
        {
          y = log(((double)r + 0.5) / ((double)(n - r) + 0.5));
          nw = ((double)r + 0.5) / (double)(n + 1) * ((double)(n - r) + 0.5);
        }
        ++nest;
      }
      ++dft;  snw += nw;  mxw += nw * x;  sxxw += nw * x * x;
      myw += nw * y;  syyw += nw * y * y;  sxyw += nw * x * y;
    }
  }
  --dft;
  const double Seps = DBL_EPSILON * fmax(5.0, syyw);
  if (dft < 1 || snw < Seps)
  {
    sxxw = syyw = sb = 0.0;
    return -1;                  // ロジット分析不可能
  }
  sxxw = sxxw - mxw * mxw / snw;
  if (sxxw < Seps) sxxw = 0.0;
  syyw = syyw - myw * myw / snw;
  if (syyw < Seps) syyw = 0.0;
  sxyw = sxyw - mxw * myw / snw;
  if (fabs(sxyw) < Seps) sxyw = 0.0;
  sb = sxyw * sxyw / sxxw;
  if (fabs(sb) < Seps) sb = 0.0;
  mxw = mxw / snw;  myw = myw / snw;
  b = sxyw / sxxw;
  a = myw - b * mxw;
  return nest;                  // 推測値で補完した用量数を返す
}

void trendclass::plot_ngp(ofstream& fout, option& opt, idd& id, const ftable& ftab) const
// Ngraph用データを出力する
{
  double a = 0.0, b = 0.0, sxxw = 0.0, syyw = 0.0, sb = 0.0;
  int dft = 0, nr = 0;
  bool delml = false;
  if (opt.ngp("pro"))           // 出現率
  {
    fout << "-s1 -dmark,00 -o300 -zy0,1,0.2\n";
    plot_ngpid(fout, id);
    plot_ngppro(fout, ftab);
    delml = true;
  }
  if (opt.ngp("log"))           // ロジスティック曲線
  {
    if (delml)
    { fout << NGPdelm << "\n"; }
    else
    {
      fout << "-s1 -dline -w80 -zy0,1,0.2\n";
      delml = true;
    }
    plot_ngpid(fout, id);
    nr = cal_lpline(opt, a, b, sxxw, syyw, sb, dft, ftab);
    plot_ngplog(fout, a, b, ftab);
  }
  if (opt.ngp("pl"))            // 出現率直線
  {
    if (delml) fout << NGPdelm << "\n";
    else fout << "-s1 -dline -w80 -zy0,1,0.2\n";
    plot_ngpid(fout, id);
    nr = cal_caline(a, b, sxxw, syyw, sb, dft, ftab);
    plot_ngppl(fout, a, b, ftab);
  }
}

void trendclass::plot_ngpid(ofstream& fout, idd& id) const
// Ngraph用データのタイトルとデータ識別情報を出力する
{
  fout << "%";  puttitle(fout, Title);
  fout << "%\n";
  id.put_title(fout, formngp);
  id.put_stra(fout, formngp);
  for (size_t i = 0; i < id.ni(); ++i)
  {
    switch (id.eitem(i).aval())
    {
    case avalrow : fout << "EXPLANATORY VARIAVLE(ROW) :" << id.item(i) << "\n";
      break;
    case avalcol : fout << "CRITERION VARIABLE(COLUMN):" << id.item(i) << "\n";
      break;
    default : id.put_item(fout, i, formngp);
    }
  }
}

void trendclass::plot_ngplog(ofstream& fout, double a, double b, const ftable& ftab) const
// Ngraph用にロジスティック曲線を出力する
{
  fout << "%<LOGISTIC CURVE>\n";
  constexpr uint RDdef = 50;    // ロジスティック曲線の分解数
  double xmin = strtodf(ftab.rcode(1)), xmax = strtodf(ftab.rcode(ftab.nr()));
  double range = xmax - xmin;
  double xstart = xmin - range * 0.1, xend = xmax + range * 0.1;
  double dx = (xend - xstart) / (double)RDdef;
  double x = xstart;
  for (uint i = 0; i <= RDdef; ++i, x += dx)
  {
    double pr = 1.0 / (1.0 + exp(-a - b * x));
    fout << x << " " << pr << "\n";
  }
}

void trendclass::plot_ngppl(ofstream& fout, double a, double b, const ftable& ftab) const
// Ngraph用に出現率直線を出力する
{
  fout << "%<PROPORTION LINE>\n";
  double xmin = strtodf(ftab.rcode(1)), xmax = strtodf(ftab.rcode(ftab.nr()));
  double range = xmax - xmin;
  double x = xmin - range * 0.1;
  fout << x << " " << a + b * x << "\n";
  x = xmax + range * 0.1;
  fout << x << " " << a + b * x << endl;
}

void trendclass::plot_ngppro(ofstream& fout, const ftable& ftab) const
// Ngraph用に出現率を出力する
{
  fout << "%<PROPORTION>\n";
  for (size_t i = 1; i <= ftab.nr(); ++i)
  {
    double x = strtodf(ftab.rcode(i));
    double pro = (double)(ftab.f(i, 2)) / (double)(ftab.frtotal(i));
    fout << x << " " << pro << "\n";
  }
}

int trendclass::get(const char* c_fname, option& opt) // 度数集計表を読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nftab = fgetftable(fname, idlst, ftablst);
  if (nftab < 0) throw runtime_error("Run-time error !");
  if (opt.check() != 0) throw option_error("Run-time option error !");
  return nftab;                 // 読み込んだ度数集計表の数を返す
}

void trendclass::put(option& opt, int fno) // 結果を出力する
{
  listidd::iterator iitr = idlst.begin();
  bool puttab = false;
  for (listftab::iterator fitr = ftablst.begin(); fitr != ftablst.end(); ++iitr, ++fitr, ++fno)
  {
    fitr->trm();                // 行・列の合計が0のものを削除する
    if (fitr->nr() > 0)
    {
      if (opt.res("res"))
      {
        put_id(cout, *iitr);
        if (opt.res("tab")) putftable(cout, *fitr, opt.tab());
        puttab = true;
      }
      if (fitr->nr() > 1 && fitr->nc() == 2)
      {
        double a = 0.0, b = 0.0, sxxw = 0.0, syyw = 0.0, sb = 0.0;
        int dft = 0, nr = 0;
        if (opt.res("res"))
        {
          if (opt.res("ca") || opt.res("log"))
          {
            if (puttab) cout << "\n";
            put_trend(cout, opt, *fitr);
          }
          cout << "\n" << endl;
        }
        if (opt.ngpname(fno) != "") // Ngraph用データ出力
        {
          ofstream foutngp;
          if (ofopen(foutngp, (opt.ngpname(fno)).c_str()) == false) return;
          plot_ngp(foutngp, opt, *iitr, *fitr);
          foutngp.close();
        }
      }
    }
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  trendclass trend;             // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    int fno = 1;                // ファイル番号
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok)                // 標準入力から読み込む
        {
          int nftab = trend.get(nullptr, opt);
          trend.put(opt, fno);
          fno += nftab;
        }
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      {
        int nftab = trend.get(argv[i], opt);  fok = false;
        trend.put(opt, fno);
        fno += nftab;
      }
    }
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
