// wtabe.C Version 7.41e   Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-28 14:08:49 tac>
//
// 順位データの差を集計し、符号付き順位検定用度数データを出力する。
// 度数データはwtesteコマンドによって解析する。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W wtabe wtabe.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "wtabe  Version 7.41e         Copyleft(Vo\\oV) 2024e Sugimoto Norio",
  "USAGE:wtabe  [<OPTION> ...]  [<DATA FILE> ...]",
  "  The tabulate ordinal data differences for the Wilcoxon's signed rank test.",
  "<OPTION>",
  "  -f=<FILE>:read options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -w=<NUMBER>:specify the number of output digits(minimum:2)",
  "    default:-w=6"
};

constexpr int Wdef = 6;         // 省略時出力桁数
constexpr int Wmax = LENcrt / 2; // 最大出力桁数

using mapdp = map<double, pairint>; // グレードマップの別名定義
using mapdp_itr = map<double, pairint>::iterator; // 指定子の別名定義
using mapdp_citr = map<double, pairint>::const_iterator; // const指定子の別名定義

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  int vw;                       // データの出力桁数
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  int w() const                 // データの出力桁数を返す
  { return vw; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
};

option::option(const char* c_pname) : vpath(), vpname(), vw(Wdef)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "f")       // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "w")       // データの出力桁数指定
  {
    char* ptr = 0;
    vw = strtol(opt[1].c_str(), &ptr, 0);
    if (vw < 2 || Wmax < vw) err = badopt(vpname, optstr, USAGE, 2, Wmax);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error !") + optstr);
}

class wtabclass                 // wtabメインクラス
{
private:
  string fname;                 // ファイル名
  idd id;                       // データ識別情報
  ldata datalst;                // データリスト
  int sum_wtab(vector<mapdp>& wmapvec, const option& opt);
  // グレードの度数集計
  void put_id(ostream& fout, int idx); // タイトルとデータ識別情報を出力する
public:
  wtabclass() : fname(), id(), datalst() // コンストラクタ
  {}
  ~wtabclass()                  // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // データを読み込み累積する
  int get(const char* c_fname); // データを読み込み累積する
  void put(option& opt);        // 集計結果を出力する
};

int wtabclass::sum_wtab(vector<mapdp>& wmapvec, const option& opt)
// グレードの度数集計
{
  int ngrade = 0;
  for (auto& cdata : datalst)
  {
    double gkey = cdata.fdata(0);
    if (gkey != BLANKf)
    {
      for (size_t j = 0; j < cdata.n() - 1; ++j)
      {
        double g = cdata.fdata(j + 1);
        double dg = g - gkey;
        if (g != BLANKf && sgn(dg) != 0)
        {
          double abs_dg = fabs(dg); // グレードの差の絶対値
          mapdp_itr witr = wmapvec[j].find(abs_dg);
          if (witr == wmapvec[j].end()) // 新規データ
          {
            pairint tf(0, 0);
            if (dg > 0.0) tf.first = 1; // 差が正の度数
            else tf.second = 1;     // 差が負の度数
            wmapvec[j].insert(mapdp::value_type(abs_dg, tf));
          }
          else                      // 既存データ
          {
            if (dg > 0.0) ++(witr->second.first);
            else ++(witr->second.second);
          }
          ++ngrade;
        }
      }
    }
  }
  return ngrade;                // 集計したグレード数を返す
}

void wtabclass::put_id(ostream& fout, int idx)
// タイトルとデータ識別情報を出力する
{
  id.put_title(fout, formtag);
  id.put_stra(fout, formtag);
  id.put_item(fout, 0, formtag);
  id.put_item(fout, idx, formtag);
}

int wtabclass::get(const char* c_fname, option& opt) // データを読み込み累積する
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nc = fgetdata(fname, id, datalst);
  if (nc < 0) throw runtime_error("Run-time error !");
  return nc;                    // 読み込んだケース数を返す
}

void wtabclass::put(option& opt) // 集計結果を出力する
{
  vector<mapdp> wmapvec(id.ni() - 1); // グレードマップベクトル
  if (id.nc() < 1 || sum_wtab(wmapvec, opt) < 1) return; // データなし
  datac tdata;
  ostringstream ostr;
  for (size_t j = 0; j < id.ni() - 1; ++j)
  {
    if (wmapvec[j].size() > 0)
    {
      put_id(cout, j + 1);
      tdata.resize(wmapvec[j].size() + 1);
      tdata.assign(0, "|DIFF| ");
      int idx = 1;
      for (mapdp_citr witr = wmapvec[j].begin(); witr != wmapvec[j].end(); ++witr, ++idx)
      {
        ostr.str("");
        ostr << setw(opt.w()) << witr->first;
        tdata.assign(idx, ostr.str());
      }
      tdata.put(cout);
      tdata.assign(0, "DIFF:+");
      idx = 1;
      for (mapdp_citr witr = wmapvec[j].begin(); witr != wmapvec[j].end(); ++witr, ++idx)
      {
        ostr.str("");
        ostr << witr->second.first;
        tdata.assign(idx, ostr.str());
      }
      tdata.put(cout);
      tdata.assign(0, "DIFF:-");
      idx = 1;
      for (mapdp_citr witr = wmapvec[j].begin(); witr != wmapvec[j].end(); ++witr, ++idx)
      {
        ostr.str("");
        ostr << witr->second.second;
        tdata.assign(idx, ostr.str());
      }
      tdata.put(cout);
      cout << endl;
    }
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  wtabclass wtab;               // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok) wtab.get(nullptr, opt); // 標準入力から読み込む
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      { wtab.get(argv[i], opt);  fok = false; }
    }
    wtab.put(opt);
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
