// wteste.C Version 7.42e  Copyleft (c) 杉本典夫
// Time-stamp: <2024-05-28 14:31:27 tac>
//
// グレード差の集計表を入力し、Wilcoxonの1標本検定と符号検定を行う。
// 通常はwtabeコマンドとパイプで接続して用いる。
// 3時期でも計算するが、その場合は本来はFriedmanの検定を用いるべきである。
//
// 使用ライブラリ:libdanse.so, libfune.so
// コンパイル法 :g++ -O -s -W wteste wteste.C
//              -I/usr/local/include -L/usr/local/lib -ldanse -lfune

#include <danse.H>              // dans用ヘッダファイル(fune.H含む)

using namespace dans;           // dans名前空間使用(名前空間matvec含む)

const vecstr USAGE =            // プログラム使用法表示用
{
  "wteste  Version 7.42e        Copyleft(Vo\\oV) 2024 Sugimoto Norio",
  "USAGE:wteste [<OPTION> ...] [<TAB FILE> ...]",
  "  The Wilcoxon's signed rank test with the tablation table of grade",
  "  differences.",
  "<OPTION>",
  "  -f=<FILE>:read options from the specified file",
  "  -h[=<MAN COMMAND>]:show online manual(-h:-h=man)",
  "  -res=[-][{def|all|cc[<C-COEF>]|m[<NUMVER>]|mt[{c|o}]|r|st",
  "    |tab[{fr|p|pc|pr|pt}]|wt[{a|c|o}]}][,...]:select output results",
  "    -:with -, no output(-res=-:no output all)",
  "    def:same output as default  all:output all",
  "    cc[<C-COEF>]:confidence coefficient(0=<C-COEF<1 cc:cc0.95)",
  "    m[<NUMBER>]:multiple test(m:auto seting 1<NUMBER)",
  "    mt[{c|o}]:McNemar's test",
  "      c:with Yates's correction  o:no Yates's correction",
  "    r:coefficient of determination  st:sign test",
  "    tab[={fr|p|pc|pr|pt}]:tablation table of grade differences",
  "      fr:only frequency  p:also output % for column or row total(default)",
  "      pc:also output % for column total  pr:also output % for row total",
  "      pt:also output % for all total",
  "    wt[{a|c|o}]:Wilcoxon's signed rank test",
  "      a:auto judgment Yates's correction  c:with Yates's correction",
  "      o:no Yates's correction",
  "    default:auto judgment according to data"
};

const string Title = "=== Intra Group Comparison of Ordinal Data ==="; // 表題

class option                    // オプション集合体
{
private:
  string vpath;                 // プログラムのパス
  string vpname;                // プログラム名
  double vcc;                   // 信頼係数
  int vm0;                      // 多重検定時の指定検定数
  int vm;                       // 多重検定時の検定数
  testtype vmt;                 // McNemar検定結果の種類
  lstr vreslst;                 // 結果出力リスト
  map<string, bool> vresmap;    // 結果出力の有無マップ
  tabtype vtab;                 // 度数集計表の種類
  testtype vwt;                 // Wilcoxonの1標本検定結果の種類
public:
  option(const char* c_pname = nullptr); // コンストラクタ
  ~option()                              // デストラクタ
  {}
  const string& path() const    // プログラムのパスを返す
  { return vpath; }
  const string& pname() const   // プログラム名を返す
  { return vpname; }
  double cc() const             // 信頼係数を返す
  { return vcc; }
  int m0() const                // 多重検定時の指定検定数を返す
  { return vm0; }
  int m() const                 // 多重検定時の検定数を返す
  { return vm; }
  int m(int m0)                 // 多重検定時の検定数を設定する
  { return (vm = m0); }
  testtype mt() const           // McNemar検定結果の種類を返す
  { return vmt; }
  bool res(const string& code); // 解析結果出力フラッグを返す
  tabtype tab() const           // 度数集計表出力フラッグを返す
  { return vtab; }
  testtype wt() const           // Wilcoxonの1標本検定結果の種類を返す
  { return vwt; }
  void setdef()                 // 省略時オプション指定を設定する
  { getoptdef(*this); }
  void set(const string& optstr, char devc = OPTdev);
  // オプション指定によってオプションを設定する
  int check(int nfile);         // オプションのチェックとセット
  int set_res();                // -resオプションのセット
};

option::option(const char* c_pname) : vpath(), vpname(), vcc(CCdef), vm0(0), vm(0), vmt(testcor), vtab(tabetc), vwt(testaut)
// コンストラクタ
{
  if (c_pname == nullptr) return; // 初期設定せず
  vpname = getname(c_pname);    // パス名なしのプログラム名取得
  vpath = c_pname;
  strsize_t idx = vpath.find(vpname);
  if (idx != string::npos) vpath.erase(idx); // プログラム名削除
}

bool option::res(const string& code) // 結果出力フラッグを返す
{
  map<string, bool>::iterator respos = vresmap.find(code);
  if (respos == vresmap.end()) return false; // code無し！
  return vresmap[code];                     // true/falseを返す
}

void option::set(const string& optstr, char devc)
// オプション指定によってオプションを設定する
{
  string opt[2];                // オプションのopt[0]:項目,opt[1]:内容
  if (getopt(opt, optstr, devc) < 1) return; // オプション指定なし
  int err = 0;
  if (opt[0] == "f")            // オプション指定ファイル
  { err = fgetopt(*this, opt[1], true); }
  else if (opt[0] == "h")       // オンラインヘルプ表示
  { err = help(vpname, USAGE, opt[1]); }
  else if (opt[0] == "res")     // 解析結果の出力指定
  {
    if (strsplit(vreslst, opt[1], CHRdelm) < 1) err = badopt(vpname, optstr, USAGE);
  }
  else                          // 間違ったオプション指定
  { err = nonopt(vpname, optstr, USAGE); }
  if (err > 0) throw help_except("Show this help message");
  if (err < 0) throw option_error(string("Option specification error !") + optstr);
}

int option::check(int nftab)    // オプションのセットとチェック
{
  if (set_res() < 0) return -2; // -resオプションのセット
  if (vm0 == 0) vm = nftab;     // 多重検定数自動設定
  else vm = vm0;                // 多重検定数を指定値に設定する
  return 0;                     // エラーなし
}

int option::set_res()           // -resオプションのセット
{
  vresmap["res"] = vresmap["def"] = vresmap["cc"] = vresmap["tab"] = vresmap["wt"] = true;
  vresmap["m"] = vresmap["mt"] = vresmap["r"] = vresmap["st"] = false;
  if (vreslst.size() > 0)
  {
    vresmap["def"] = vresmap["cc"] = vresmap["tab"] = vresmap["wt"] = false;
    for (string res : vreslst)
    {
      map<string, bool>::iterator respos = vresmap.find(res);
      if (res == "-")           // 全て出力せず
      { vresmap["res"] = false; }
      else if (res[0] == '-' && res.size() > 1) // 特定の項目を出力せず
      {
        respos = vresmap.find(res.substr(1));
        if (respos == vresmap.end())
        { return disperr(vpname, "Wrong -res=" + res + " option !"); }
        else
        { vresmap[res.substr(1)] = false; }
      }
      else if (res == "def")    // 省略時出力
      { vresmap["def"] = vresmap["cc"] = vresmap["tab"] = vresmap["wt"] = true; }
      else if (res == "all")    // 全て出力
      {
        vresmap["def"] = vresmap["cc"] = vresmap["m"] = vresmap["mt"] = vresmap["r"] = vresmap["st"] = vresmap["tab"] = vresmap["wt"] = true;
        vmt = testcor;  vtab = tabetc;  vwt = testaut;
      }
      else if (res[0] == 'c' && res[1] == 'c' && res.size() > 2)
      {
        vcc = strtodf(res.substr(2)); // 信頼係数
        if (vcc < 0.0 || 1.0 <= vcc) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["cc"] = true;
      }
      else if (res[0] == 'm' && res.size() > 1)
      {
        char* ptr = nullptr;
        vm0 = strtol(res.substr(1).c_str(), &ptr, 0); // 検定数
        if (vm0 < 2) return disperr(vpname, "Wrong -res=" + res + " option !");
        vresmap["m"] = true;
      }
      else if (res == "mtc") // 連続修正有り
      { vresmap["mt"] = true;  vmt = testcor; }
      else if (res == "mto") // 連続修正無し
      { vresmap["mt"] = true;  vmt= testorg; }
      else if (res == "tabfr") // 度数だけ出力
      { vresmap["tab"] = true;  vtab = tabfreq; }
      else if (res == "tabp")  // 列または行合計に対する％出力
      { vresmap["tab"] = true;  vtab = tabetc; }
      else if (res == "tabpc") // 列合計に対する％出力
      { vresmap["tab"] = true;  vtab = tabfprc; }
      else if (res == "tabpr") // 行合計に対する％出力
      { vresmap["tab"] = true;  vtab = tabfprr; }
      else if (res == "tabpt") // 全合計に対する％出力
      { vresmap["tab"] = true;  vtab = tabfprt; }
      else if (res == "wta")   // 連続修正自動判定
      { vresmap["wt"] = true;  vwt= testaut; }
      else if (res == "wtc")   // 連続修正有り
      { vresmap["wt"] = true;  vwt= testcor; }
      else if (res == "wto")   // 連続修正無し
      { vresmap["wt"] = true;  vwt = testorg; }
      else if (respos == vresmap.end())
      { return disperr(vpname, "Wrong -res=" + res + " option !"); }
      else
      { vresmap[res] = true; }
    }
  }
  return 0;                     // エラーなし
}

class wtestclass                // wtestメインクラス
{
private:
  string fname;                 // ファイル名
  listidd idlst;                // データ識別情報リスト
  listftab ftablst;             // 度数表リスト
  void put_id(ostream& fout, idd& id) const;
  // タイトルとデータ識別情報を出力する
  void put_wtest(ostream& fout, option& opt, const ftable& ftab) const;
  // Wilcoxonの1標本検定の結果を出力する
  void put_tpci(ostream& fout, const option& opt, double tp, double vt, mctype mc, int dc, const ftable& ftab) const;
  // Wilcoxonの1標本検定のTpの信頼区間出力
  void put_stest(ostream& fout, option& opt, const ftable& ftab) const;
  // 符号検定の結果を出力する
  void put_stpci(ostream& fout, const option& opt, mctype mc, int dc, const ftable& ftab) const;
  // 符号検定のpp信頼区間出力
  void put_mtest(ostream& fout, option& opt, const ftable& ftab) const;
  // Mcnemar検定の結果を出力する
  void put_mtpci(ostream& fout, const option& opt, mctype mc, int dc, const ftable& ftab) const;
  // Mcnemar検定のpp信頼区間出力
public:
  wtestclass() : fname(), idlst(), ftablst() // コンストラクタ
  {}
  ~wtestclass()                 // デストラクタ
  {}
  int get(const char* c_fname, option& opt); // 度数集計表を読み込む
  void put(option& opt);                     // 結果を出力する
};

void wtestclass::put_id(ostream& fout, idd& id) const
// タイトルとデータ識別情報を出力する
{
  puttitle(fout, Title);
  fout << "\n";
  id.put_title(fout);
  id.put_stra(fout);
  fout << "ANALYSIS ITEM :" << id.item(1) << "\n";
  fout << "-BASELINE ITEM:" << id.item(0) << "\n";
}

void wtestclass::put_wtest(ostream& fout, option& opt, const ftable& ftab) const
// Wilcoxonの1標本検定の結果を出力する
{
  double tp = 0.0, vt = 0.0;
  double z = wtest(ftab, tp, vt, opt.wt());
  ldbl pr = ztop(z);
  fout << "\nWilcoxon's Signed Rank Test";
  if (opt.wt() == testcor || (opt.wt() == testaut && fabs(z) < 2.0)) fout << "(with Yates's correction)";
  fout << "\n";
  fout << "  Normal Approximation Test:z=" << z << "  p=" << pr << pind(pr) << "\n";
  if (opt.res("r"))             // 寄与率出力
  {
    z = wtest(ftab, tp, vt, testorg);
    fout << "  Coefficient of Determnation:r-squared=" << z * z / double(ftab.ftotal()) << "\n";
  }
  put_tpci(fout, opt, tp, vt, mcfis, 1, ftab);
  if (opt.res("m") == false || opt.m() == 1) return;
  pr = pr * (ldbl)(opt.m());
  if (pr > 1.0L) pr = 1.0L;
  fout << "\n  Bonferroni Type Multiple Test:p=" << pr << pind(pr) << "\n";
  put_tpci(fout, opt, tp, vt, mcbon, opt.m(), ftab);
}

void wtestclass::put_tpci(ostream& fout, const option& opt, double tp, double vt, mctype mc, int dc, const ftable& ftab) const
// Wilcoxonの1標本検定のTpの信頼区間出力
{
  fout << "  " << opt.cc() * 100.0 << "%";
  if (mc != mcfis) fout << " Simultaneous";
  fout << " CI of Signed Rank Sum\n";
  double nf = ftab.ftotal();
  double tn = nf * (nf + 1.0) * 0.5;
  double tm = tn - tp;
  double ts = tp - tm;
  double ci = calmcci(mc, opt.cc(), 1, 1, 0, dc, 4.0 * vt);
  double tsl = ts - ci, tsu = ts + ci;
  if (tsl < -tn) tsl = -tn;
  if (tsu > tn) tsu = tn;
  double rs = ts / nf, cin = ci  / nf;
  double rsl = rs - cin, rsu = rs + cin;
  if (rsl < -tn / nf) rsl = -tn / nf;
  if (rsu > tn / nf) rsu = tn / nf;
  double pro = 100.0 / nf;
  fout << "  " << tp << " - " << tm << "=" << ts << " +/- " << ci << "(" << tsl << " - " << tsu << ")\n";
  fout << "  signed rank mean=" << rs << " +/- " << cin << "(" << rsl << " - " << rsu << ")\n";
  fout << "  proportion(%)=" << rs * pro << " +/- " << cin * pro << "(" << rsl * pro << " - " << rsu * pro << ")\n";
}

void wtestclass::put_stest(ostream& fout, option& opt, const ftable& ftab) const
// 符号検定の結果を出力する
{
  ldbl pr1 = binom(ftab.frtotal(1), ftab.frtotal(2));
  ldbl pr = (pr1 < 0.5L) ? 2.0L * pr1 : 1.0L;
  fout << "\nSign Test(Binomial Test of Theoretical Probability=0.5)\n";
  fout << "  One-tailed p=" << pr1 << pind(pr1) << "  Two-tailed p=" << pr << pind(pr) << "\n";
  put_stpci(fout, opt, mcfis, 1, ftab);
  if (opt.res("m") == false || opt.m() == 1) return;
  pr1 = pr1 * (ldbl)(opt.m());
  if (pr1 > 1.0L) pr1 = 1.0L;
  pr = (pr1 < 0.5L) ? 2.0L * pr1 : 1.0L;
  fout << "\n  Bonferroni Type Multiple Test\n";
  fout << "    One-tailed p=" << pr1 << pind(pr1) << "  Two-tailed p=" << pr << pind(pr) << "\n";
  put_stpci(fout, opt, mcbon, opt.m(), ftab);
}

void wtestclass::put_stpci(ostream& fout, const option& opt, mctype mc, int dc, const ftable& ftab) const
// 符号検定のpp信頼区間出力
{
  fout << "  " << opt.cc() * 100.0 << "%";
  if (mc != mcfis) fout << " Simultaneous";
  fout << " CI of pr of [" << ftab.rcode(1) << "] - theoretical pr\n";
  int nt = ftab.ftotal();
  int n = ftab.frtotal(1);
  double p0 = (double)n / (double)nt;
  double dp = p0 - 0.5;
  double al = 1.0 - opt.cc();
  double cc = 1.0 - al / (double)dc;
  double plow = 0.0, pup = 1.0;
  double pci = calbexci(plow, pup, n, nt, cc);
  fout << "  " << p0 << "-0.5=" << dp << "  lower=" << plow - 0.5 << "  upper=" << pup - 0.5 << endl;
}

void wtestclass::put_mtest(ostream& fout, option& opt, const ftable& ftab) const
// McNemar検定の結果を出力する
{
  double tp = (double)(ftab.frtotal(1) - ftab.frtotal(2));
  double vt = (double)(ftab.ftotal());
  double z = (tp - sgn(tp)) / sqrt(vt);
  if (opt.mt() == testorg) z = tp / sqrt(vt);
  ldbl pr = ztop(z);
  fout << "\nMcNemar's Test";
  if (opt.mt() != testorg) fout << "(with Yates's correction)";
  fout << "\n  Normal Approximation Test:z=" << z << " p=" << pr << pind(pr) << "\n";
  put_mtpci(fout, opt, mcfis, 1, ftab);
  if (opt.res("m") == false || opt.m() == 1) return;
  pr = pr * (ldbl)(opt.m());
  if (pr > 1.0L) pr = 1.0L;
  fout << "\n  Bonferroni Type Multiple Test  p=" << pr << pind(pr) << "\n";
  put_mtpci(fout, opt, mcbon, opt.m(), ftab);
}

void wtestclass::put_mtpci(ostream& fout, const option& opt, mctype mc, int dc, const ftable& ftab) const
// McNemarの検定のpp信頼区間出力
{
  fout << "  " << opt.cc() * 100.0 << "%";
  if (mc != mcfis) fout << " Simultaneous";
  fout << " CI of pr of [" << ftab.rcode(1) << "] - theoretical pr\n";
  int nt = ftab.ftotal();
  double p0 = (double)(ftab.frtotal(1)) / (double)nt;
  double dp = p0 - 0.5;
  double al = 1.0 - opt.cc();
  double cc = 1.0 - al / (double)dc;
  bool cor = (opt.mt() == testorg) ? false : true;
  double plow = 0.0, pup = 1.0;
  double pci = calbnmci(plow, pup, p0, nt, cc, cor);
  fout << "  " << p0 << "-0.5=" << dp << " +/- " << pci << "(" << dp - pci << " - " << dp + pci << ")\n";
}

int wtestclass::get(const char* c_fname, option& opt) // 度数集計表を読み込む
{
  if (c_fname == nullptr) fname.erase();
  else fname = c_fname;
  int nftab = fgetftable(fname, idlst, ftablst);
  if (nftab < 0) throw runtime_error("Run-time error !");
  if (opt.check(nftab) != 0) throw option_error("Run-time option error !");
  return nftab;                 // 読み込んだ度数集計表の数を返す
}

void wtestclass::put(option& opt) // 結果を出力する
{
  if (opt.res("res") == false) return;
  listidd::iterator iitr = idlst.begin();
  for (listftab::iterator fitr = ftablst.begin(); fitr != ftablst.end(); ++iitr, ++fitr)
  {
    put_id(cout, *iitr);
    if (fitr->nr() > 0)
    {
      if (opt.res("tab"))
      {
        if (opt.tab() == tabetc)
        {
          if (fitr->nc() == 1) putftable(cout, *fitr, tabfprc);
          else putftable(cout, *fitr, tabfprr);
        }
        else
        { putftable(cout, *fitr, opt.tab()); }
      }
      if (fitr->nr() == 2)
      {
        if (opt.res("wt")) put_wtest(cout, opt, *fitr);
        if (opt.res("st") || (opt.res("def") && fitr->nc() == 1)) put_stest(cout, opt, *fitr);
        if (opt.res("mt")) put_mtest(cout, opt, *fitr);
        cout << "\n" << endl;
      }
    }
  }
}

int main(int argc, char* argv[]) // メインルーチン
{
  wtestclass wtest;             // メインクラス変数
  option opt(argv[0]);          // オプション
  try
  {
    opt.setdef();               // オプションの初期設定
    bool fok = true;            // ファイル処理フラッグ
    for (int i = 1; i <= argc; ++i)
    {
      if (i >= argc)
      {
        if (fok)                // 標準入力から読み込む
        {
          wtest.get(nullptr, opt);
          wtest.put(opt);
        }
      }
      else if (*(argv[i]) == CHRopt)
      { opt.set(argv[i] + 1);  fok = true; }
      else
      {
        wtest.get(argv[i], opt);  fok = false;
        wtest.put(opt);
      }
    }
  }
  catch (help_except& helpex)   // ヘルプファイル表示
  { return 0; }
  catch (option_error& opterr)  // オプション指定エラー
  { return 1; }
  catch (runtime_error& rterr)  // 実行時エラー
  {
    disperr(opt.pname(), rterr.what());
    return 2;
  }
  catch (...)                   // 予期せぬ例外
  {
    disperr(opt.pname(), "Unknown error occurred !");
    return 255;
  }
  return 0;
}
